/*
 * linux/tcpcp.h - TCP connection passing, data structures and prototypes
 *
 * Written 2002 by Werner Almesberger
 *
 * Copyright (C) 2005 NTT Corporation
 */

#ifndef _LINUX_TCPCP_H
#define _LINUX_TCPCP_H

#include <linux/unistd.h>

#ifdef __KERNEL__
#include <linux/types.h>
#include <linux/in.h>
#include <linux/in6.h>
#include <linux/tcp.h>

#ifdef CONFIG_TCPCP_DEBUG
#define TCPCP_DPRINTK(fmt, args...) printk(KERN_DEBUG "TCPCP3: %s: " fmt, __func__, ## args)
#else
#define TCPCP_DPRINTK(fmt, args...)
#endif


#define TCPCP_MODULE_NAME	"tcpcp"

#define	TCPCP_FLAG_MAKE_SI	0x01
#define	TCPCP_FLAG_STOP		0x02
#define	TCPCP_FLAG_CLOSE	0x04
#define	TCPCP_FLAG_SET_SI	0x08
#define	TCPCP_FLAG_START	0x10
#define TCPCP_FLAG_FLUSH        0x20
#define	TCPCP_FLAG_ALL		0x80

#define plugin_dummy() __asm__ __volatile__("nop\n\t" \
	"nop\n\t" \
	"nop\n\t" \
	"nop\n\t" \
	"nop\n\t" \
	"nop\n\t" \
	"nop\n\t" \
	"nop\n\t" \
	"nop\n\t" \
	"nop\n\t" \
	"nop\n\t" \
	"nop\n\t" \
	"nop\n\t" \
	"nop\n\t")

#endif /* __KERNEL__ */

/* ----- TCP Socket Informations (TCP-SIs) --------------------------------- */

/**
 * TCP Socket Information (TCP-SI) option: packet information
 **/
struct tcpcp_si_skb {
	uint8_t type;		/* TCP-SI option type */
				/*  1: TCPCP_SIE_BUF_SND */
				/*  2: TCPCP_SIE_BUF_RCV */
	uint8_t __pad1;
	uint16_t length;	/* segment data length */
	uint8_t flags;		/* TCP header flags */
	uint8_t sacked;		/* state flags for SACK/FACK */
	uint8_t __pad2[2];
	uint16_t gso_segs;
	uint16_t gso_size;
	uint32_t seq;		/* sequence number of first byte */
	uint8_t data[0];        /* data, padded to multiple of 4 bytes */
};

enum {
	TCPCP_SIE_BUF_SND = 1,	/* send buffer (only TCP segment, no IP) */
	TCPCP_SIE_BUF_RCV = 2,	/* receive buffer (only TCP segment, no IP) */
};

/**
 * TCP-SI element: globally unique TCP connection ID
 **/
struct tcpcp_sie_id4 {
	struct in_addr ip_src;	/* source IPv4 address */
	struct in_addr ip_dst;	/* destination IPv4 address */
};

struct tcpcp_sie_id6 {
	struct in6_addr ip_src;	/* source IPv6 address */
	struct in6_addr ip_dst;	/* destination IPv6 address */
	uint32_t scope_id;	/* scope id */
};

struct tcpcp_sie_id {
	uint8_t ip_version;	/* IP version */
				/*  4: IPv4 */
				/*  6: IPv6 */
	uint8_t __pad[3];
	union {
		struct tcpcp_sie_id4 v4;	/* IPv4 */
		struct tcpcp_sie_id6 v6;	/* IPv6 */
	} ip;
	uint16_t tcp_sport;	/* TCP source port */
	uint16_t tcp_dport; 	/* TCP destinatipon port */
};

/**
 * TCP-SI element: fixed general data
 **/
struct tcpcp_sie_fixgen {
	uint8_t tcp_flags;	/* TCP flags; from linux/tcp.h */
				/*  1: TCPI_OPT_TIMESTAMPS */
				/*  2: TCPI_OPT_SACK */
				/*  4: TCPI_OPT_WSCALE */
				/*  8: TCPI_OPT_ECN */
	uint8_t snd_wscale;	/* send window scale (0 if unused) */
	uint8_t rcv_wscale;	/* receive window scale (0 if unused) */
	uint8_t __pad;
	uint16_t snd_mss;	/* MSS sent */
	uint16_t rcv_mss;	/* MSS received */
};

/**
 * TCP-SI element: variable general data
 **/
struct tcpcp_sie_vargen {
	uint8_t state;		/* connection state; from linux/tcp.h */
				/*  1: TCP_ESTABLISHED */
				/*  2: TCP_SYN_SENT */
				/*  3: TCP_SYN_RECV */
				/*  4: TCP_FIN_WAIT1 */
				/*  5: TCP_FIN_WAIT2 */
				/*  6: TCP_TIME_WAIT */
				/*  7: TCP_CLOSE */
				/*  8: TCP_CLOSE_WAIT */
				/*  9: TCP_LAST_ACK */
				/* 10: TCP_LISTEN */
				/* 11: TCP_CLOSING */
				/* Note: TCP-SI may not ever use some of */
				/*       these values. */
	uint8_t __pad[3];
	uint32_t snd_nxt;	/* sequence number of next new byte to send */
	uint32_t rcv_nxt;	/* sequence number of next new byte expected */
				/* to receive */
	uint32_t snd_wnd;	/* window received from peer */
	uint32_t rcv_wnd;	/* window advertized to peer */
	uint32_t ts_recent;	/* cached timestamp from peer (0 if none) */
	uint32_t ts_gen;	/* current locally generated timestamp */
				/* (0 if not using timestamps) */
  uint32_t copied_seq;    /* sequence number of next unread byte */
};

/**
 * TCP-SI element: socket options (socket)
 **/
struct tcpcp_sie_sk_sockopt {
	uint8_t flags;		/* flags */
				/*  1: TCPCP_SIE_SOCK_REUSEADDR */
				/*  2: TCPCP_SIE_SOCK_KEEPOPEN */
				/*  4: TCPCP_SIE_SOCK_LINGER */
	uint8_t userlocks;	/* %SO_SNDBUF and %SO_RCVBUF settings */
	uint8_t __pad[2];
	uint32_t priority;	/* %SO_PRIORITY setting */
	uint32_t sndbuf;	/* size of send buffer in bytes */
	uint32_t rcvbuf;	/* size of receive buffer in bytes */
	uint32_t l_linger;	/* How long to linger for */
};

enum {
	TCPCP_SIE_SOCK_REUSEADDR	= 1,
	TCPCP_SIE_SOCK_KEEPOPEN		= 2,
	TCPCP_SIE_SOCK_LINGER		= 4,
};

/**
 * TCP-SI element: socket options (IPv4)
 **/
struct tcpcp_sie_ipv4_sockopt {
	uint8_t flags;		/* flags */
				/*  1: TCPCP_SIE_IP_DEFAULT_TTL */
	uint8_t uc_ttl;		/* unicast TTL */
	uint8_t pmtudisc;	/* IP_MTU_DISCOVER state */
	uint8_t tos;		/* TOS */
	uint8_t opt_len;
	uint8_t __pad[3];
	uint8_t opt_data[40];
};

enum {
	TCPCP_SIE_IP_DEFAULT_TTL = 1,
};

/**
 * TCP-SI element: socket options (IPv6)
 **/
struct tcpcp_sie_ipv6_sockopt {
	uint8_t flags;		/* flags */
				/*  1: TCPCP_SIE_IPV6_DEFAULT_HOPLIMIT */
	uint8_t hop_limit;	/* unicast hop limit */
	uint8_t pmtudisc;	/* IPV6_MTU_DISCOVER state */
	uint8_t __pad;
};

enum {
	TCPCP_SIE_IPV6_DEFAULT_HOPLIMIT = 1,
};

/**
 * TCP-SI element: socket options (TCP)
 **/
struct tcpcp_sie_tcp_sockopt {
	uint8_t nonagle;	/* Disable Nagle aliorithm ? */
	uint8_t keepcnt;	/* num of allowed keep alive probes */
	uint8_t __pad[2];
	uint16_t keepidle;	/* time before keep alive takes place */
	uint16_t keepintvl;	/* time interval between keep alive */
				/* probes */
};

/**
 * TCP-SI element: socket options
 **/
struct tcpcp_sie_sockopt {
	struct tcpcp_sie_sk_sockopt sk;			/* socket */
	union {
		struct tcpcp_sie_ipv4_sockopt v4;	/* IPv4 */
		struct tcpcp_sie_ipv6_sockopt v6;	/* IPv6 */
	} ip;
	struct tcpcp_sie_tcp_sockopt tcp;		/* TCP */
};

/**
 * TCP Socket Information (TCP-SI) header
 **/
struct tcpcp_si_hdr {
	uint32_t si_length;		/* total length of TCP-SI / 4byte */
	uint32_t fd;
	struct tcpcp_sie_id id;		/* globally unique TCP connection ID */
	struct tcpcp_sie_fixgen fixgen;	/* fixed general data */
	struct tcpcp_sie_vargen vargen;	/* variable general data */
	struct tcpcp_sie_sockopt sockopt;	/* socket option */
	struct tcpcp_si_skb si_skb[0];
};

/**
 * TCP Socket Informations (TCP-SIs) total header
 **/
struct tcpcp_total_si_hdr {
	uint32_t total_si_length;	/* total length of TCP-SIs / 4byte */
	uint8_t major;			/* incompatible structure revision */
					/*  0: current version */
	uint8_t minor;			/* compatible structure extension */
					/*  0: current version */
	uint16_t si_cnt;
	struct tcpcp_si_hdr si_hdr[0];
};


 /**
  * Convenience macros going from tcpcp_total_si_hdr (i.e., user
  * interface object) to specific members.
  */
#define tcpcp_hdr_ip6_src(__b) ((__b)->si_hdr->id.ip.v6.ip_src)
#define tcpcp_hdr_ip6_dst(__b) ((__b)->si_hdr->id.ip.v6.ip_dst)
#define tcpcp_hdr_ip4_src(__b) ((__b)->si_hdr->id.ip.v4.ip_src)
#define tcpcp_hdr_ip4_dst(__b) ((__b)->si_hdr->id.ip.v4.ip_dst)
#define tcpcp_hdr_tcp_sport(__b) ((__b)->si_hdr->id.tcp_sport)
#define tcpcp_hdr_tcp_dport(__b) ((__b)->si_hdr->id.tcp_dport)
#define tcpcp_hdr_var_snd_nxt(__b) ((__b)->si_hdr->vargen.snd_nxt)
#define tcpcp_hdr_var_rcv_nxt(__b) ((__b)->si_hdr->vargen.rcv_nxt)
#define tcpcp_hdr_var_snd_wnd(__b) ((__b)->si_hdr->vargen.snd_wnd)
#define tcpcp_hdr_var_ts_gen(__b) ((__b)->si_hdr->vargen.ts_gen)
#define tcpcp_hdr_var_ts_recent(__b) ((__b)->si_hdr->vargen.ts_recent)
#define tcpcp_hdr_isvalid(__b)                               \
  ( (sizeof(struct tcpcp_si_hdr) <=  (__b)->total_si_length) \
    && (1 <= (__b)->si_cnt) )


/* ----- FD Result (FDR) Information (only "ALL") -------------------------- */

struct fdr {
	int fd;
	int result;
};

struct result_list {
  unsigned int fd_all;    /* total of fd that succeeds in processing */
  unsigned int fd_ng;     /* Number of fd that fails in processing */
  struct fdr fdrs[0];     /* processing result pointer */
};

struct result {
	unsigned int fd_all;    /* total of fd that succeeds in processing */
	unsigned int fd_ng;     /* Number of fd that fails in processing */
	unsigned int buf_size;
	unsigned int fdrs_size;
};


/* ---- Process, Request, TCP-SIs and FDR Info. Entry ---------------------- */

#ifdef __KERNEL__

/**
 * TCP-SI buffer (It's used at setting TCP-SI)
 **/
struct tcpcp_si_skb_buf {
	unsigned int max_buf_len;
	unsigned int buf_len;
	unsigned int skb_cnt;
	char *si_skb;
};

struct tcpcp_si_buf {
	struct tcpcp_si_hdr si_hdr;	/* TCP-SI header */
	struct tcpcp_si_skb_buf snd_buf;
	struct tcpcp_si_skb_buf rcv_buf;
};

#define TCPCP_PAGE_NUM_SET_SI	24
#define TCPCP_SI_SKB_BUF_SIZE		PAGE_SIZE * TCPCP_PAGE_NUM_SET_SI

/**
 * common entry (It's used at making TCP-SIs or FDR Info.)
 **/
struct tcpcp_com_entry {
	struct tcpcp_com_entry *next;	/* next entry */
	unsigned int max_data_len;
	unsigned int data_len;		/* length of data */
	char data[0];
};

#define TCPCP_PAGE_NUM_MAKE_SI_ONE  3
#define TCPCP_COM_ENTRY_SIZE_FOR_ONE	PAGE_SIZE * TCPCP_PAGE_NUM_MAKE_SI_ONE
#define TCPCP_COM_ENTRY_SIZE_FOR_ALL	(PAGE_SIZE << 5) /* MAX_GFP_ORDER */

/**
 * Request Entry
 **/
struct tcpcp_req_entry {
	struct list_head node;
	struct list_head timer_node;
	uint32_t fd;
	struct sock *sk;
	pid_t pid;
	unsigned char flag;
	char dummy[3];
	unsigned int timeout;
	struct result res;
	struct tcpcp_com_entry *s_entry;	/* TCP-SIs Entry */
	struct tcpcp_com_entry *f_entry;	/* FDR Entry (only "ALL") */
	struct tcpcp_proc_entry *p_entry;	/* parent Process Entry */
};

#define TCPCP_WATCH_INTERVAL    	HZ
#define TCPCP_RETRY_WATCH_INTERVAL	TCPCP_WATCH_INTERVAL / 10
#define TCPCP_REQ_ENTRY_TIMEOUT		HZ * 10

/**
 * Process Entry
 **/
struct tcpcp_proc_entry {
	struct list_head node;
	gid_t tgid;
	unsigned char flag;
	char dummy[3];
	struct list_head r_list_head;	/* Request List */
};

/**
 * Process List or Timeout List
 **/
struct tcpcp_list {
	struct list_head head;
	rwlock_t lock;
};

/* ---- -------------------------------------------------------------------- */

#define TCPCP_SET_FLAG_OFF 0
#define TCPCP_SET_FLAG_ON  1

/*
 * Buffers are in sequence, first all send, then all out-of-order buffers.
 * Buffers must not overlap, and may not contain any extraneous data (e.g.
 * ack'ed bytes, or such). snd_nxt does not have to be at a buffer boundary.
 */

#if defined(CONFIG_TCPCP) || defined(CONFIG_TCPCP_MODULE)

#include <linux/tcp.h>
#include <linux/time.h>
#include <net/sock.h>
#include <net/tcp.h>

typedef struct tcpcp_config_st {
   /**
    * Hooks
    **/
  int (*tcpcp_com_ope_hook)(struct sock *sk, int optname,
                            struct result *user_res);
  int (*tcpcp_get_si_hook)(struct sock *sk,
                           struct tcpcp_total_si_hdr *user_si,
                           unsigned int *user_size);
  int (*tcpcp_get_res_hook)(struct sock *sk, struct fdr *user_fdrs,
                            unsigned int *user_size);
  int (*tcpcp_set_si_hook)(struct sock *sk, int optname,
                           const struct tcpcp_total_si_hdr *user_si,
                           unsigned int size);
  void (*tcpcp_close_hook)(struct sock *sk);

  int (*tcpcp_flush_hook)(struct sock *sk, struct timespec *user_timeout);

   /**
    * Socket operations
    **/
#ifdef CONFIG_TCP_MD5SIG
  struct tcp_sock_af_ops tcpcp_sock_ipv4_specific;
  const struct tcp_sock_af_ops *orig_sock_ipv4_specific_ptr;
#endif /* CONFIG_TCP_MD5SIG */
  struct inet_connection_sock_af_ops tcpcp_ipv4_specific;
  const struct inet_connection_sock_af_ops *orig_ipv4_specific_ptr;
#if defined(CONFIG_IPV6) || defined(CONFIG_IPV6_MODULE)
  struct inet_connection_sock_af_ops tcpcp_ipv6_specific;
  struct inet_connection_sock_af_ops tcpcp_ipv6_mapped;
  const struct inet_connection_sock_af_ops *orig_ipv6_specific_ptr;
  const struct inet_connection_sock_af_ops *orig_ipv6_mapped_ptr;
#ifdef CONFIG_TCP_MD5SIG
  struct tcp_sock_af_ops tcpcp_sock_ipv6_mapped_specific;
  const struct tcp_sock_af_ops *orig_sock_ipv6_mapped_specific_ptr;
  struct tcp_sock_af_ops tcpcp_sock_ipv6_specific;
  const struct tcp_sock_af_ops *orig_sock_ipv6_specific_ptr;
#endif /* CONFIG_TCP_MD5SIG */
#endif /* CONFIG_IPV6 || CONFIG_IPV6_MODULE */
} tcpcp_config_t;

/* timeout of polling for tcpcp_flush() */
#define TCPCP_FLUSH_POLL_TIMEOUT_USECS 100

 /**
  * Get access to global tcpcp config options
  */
tcpcp_config_t *tcpcp_get_config(void);

 /* Initialize IPv4 settings */
void tcpcp_v4_init(void);
#if defined(CONFIG_IPV6) || defined(CONFIG_IPV6_MODULE)
 /* Initialize IPv6 settings */
void tcpcp_v6_init(void);
#endif

extern struct sk_filter drop_all;

int tcpcp_discard_xmit(struct sk_buff *skb, int ipfragok);
void tcpcp_fix_seq(struct sock *sk, __u32 seq, __u32 ackseq);
void tcpcp_fix_ts(struct sock *sk, __u32 ack_tsecr);

/**
 * Hook call functions 
 **/
int tcpcp_call_com_ope_hook(struct sock *sk, int optname,
			    struct result *user_res);
int tcpcp_call_set_si_hook(struct sock *sk, int optname,
			   const struct tcpcp_total_si_hdr *user_si,
			   unsigned int size);
int tcpcp_call_get_si_hook(struct sock *sk, struct tcpcp_total_si_hdr *user_si,
			   unsigned int *user_size);
int tcpcp_call_get_res_hook(struct sock *sk, struct fdr *user_fdrs,
			    unsigned int *user_size);
void tcpcp_call_close_hook(struct sock *sk);
int tcpcp_call_flush_hook(struct sock *sk, struct timespec *user_timeout);

#else /* defined(CONFIG_TCPCP) || defined(CONFIG_TCPCP_MODULE) */

#define tcpcp_call_com_ope_hook(sk, optname, val) (-ENOPROTOOPT)
#define tcpcp_call_set_si_hook(sk, optname, val, size) (-ENOPROTOOPT)
#define tcpcp_call_get_si_hook(sk, val, size) (-ENOPROTOOPT)
#define tcpcp_call_get_res_hook(sk, val, size) (-ENOPROTOOPT)
#define tcpcp_call_close_hook(sk) (-ENOPROTOOPT)
#define tcpcp_call_flush_hook(sk, val) (-ENOPROTOOPT)

#endif /* !defined(CONFIG_TCPCP) && !defined(CONFIG_TCPCP_MODULE) */

#endif /* __KERNEL__ */

#endif /* _LINUX_TCPCP_H */
