
/*
 * tcpcp_hooks.c - Hooks when using TCP connection passing as a module
 *
 * Written 2002,2004 by Werner Almesberger
 *
 * Copyright (C) 2005 NTT Corporation
 */


#if defined(CONFIG_TCPCP) || defined(CONFIG_TCPCP_MODULE)

#include <linux/module.h>
#include <linux/rwsem.h>
#include <linux/kmod.h>
#include <linux/tcpcp.h>

#include <net/tcp.h>


/* ----- Hooks ------------------------------------------------------------- */

static tcpcp_config_t __tcpcp_config;

 /**
  *
  */
tcpcp_config_t *tcpcp_get_config() {
  return &__tcpcp_config;
}
EXPORT_SYMBOL(tcpcp_get_config);

/**
 * tcpcp_call_com_ope_hook
 *   - Call hook function of other operations of those other than
 *     GET_SI, SET_SI, SET_SI_ALL and GET_RES.
 *     The target operations are MAKE_SI, MAKE_SI_ALL, STOP, STOP_ALL,
 *     CLOSE_ALL, START, START_ALL and GET_RES_SIZE.
 *
 * [Args]
 *   (IN)  *sk       : socket (if ALL, socket is dummy fd)
 *   (IN)  optname   : option name
 *                     (TCPCP_MAKE_SI, TCPCP_MAKE_SI_ALL, TCPCP_STOP,
 *                      TCPCP_STOP_ALL, TCPCP_CLOSE_ALL, TCPCP_START,
 *                      TCPCP_START_ALL, TCPCP_GET_RES_SIZE)
 *   (OUT) *user_res : result info
 *
 * [Return]
 *   0      : OK.
 *  -ENOSYS : NG. (function not implemented)
 *   other  : (skip)
 **/
int tcpcp_call_com_ope_hook(struct sock *sk, int optname,
			    struct result *user_res)
{
	int error = -ENOSYS;

	/* plug-in. */
	plugin_dummy();

	/* start. */
	TCPCP_DPRINTK("start. sk=0x%p, optname=%d\n",
		sk, optname);

#ifdef CONFIG_KMOD
	/* hook is not activated ? */
	if (NULL == __tcpcp_config.tcpcp_com_ope_hook) {
		TCPCP_DPRINTK("request TCPCP module.\n");
     /* This is protected by a lock internally. */
		request_module(TCPCP_MODULE_NAME);
	}
#endif  

	/* hook is activated ? */
	if (NULL != __tcpcp_config.tcpcp_com_ope_hook) {
		error = __tcpcp_config.tcpcp_com_ope_hook(sk, optname, user_res);
  }
  
	/* end. */
	TCPCP_DPRINTK("end. error=%d\n", error);
	return error;
}

/**
 * tcpcp_call_get_si_hook - call hook function for getting TCP-SIs.
 *
 * [Args]
 *   (IN)  *sk        : socket (dummy)
 *   (OUT) *user_si   : TCP-SIs acquired from Kernel
 *   (OUT) *user_size : size of user_si
 *
 * [Return]
 *   0      : OK.
 *  -ENOSYS : NG. (function not implemented)
 *   other  : (skip)
 **/
int tcpcp_call_get_si_hook(struct sock *sk,
			   struct tcpcp_total_si_hdr *user_si,
			   unsigned int *user_size)
{
	int error = -ENOSYS;

	/* plug-in. */
	plugin_dummy();

	/* start. */
	TCPCP_DPRINTK("start. sk=0x%p\n", sk);

#ifdef CONFIG_KMOD
	/* hook is not activated ? */
	if (NULL == __tcpcp_config.tcpcp_get_si_hook) {
		TCPCP_DPRINTK("request TCPCP module.\n");
     /* This is protected by a lock internally. */
		request_module(TCPCP_MODULE_NAME);
	}
#endif

	/* hook is activated ? */
	if (NULL != __tcpcp_config.tcpcp_get_si_hook) {
		error = __tcpcp_config.tcpcp_get_si_hook(sk, user_si, user_size);
  }

	/* end. */
	TCPCP_DPRINTK("end. error=%d\n", error);
	return error;
}

/**
 * tcpcp_call_get_res_hook - call hook function for getting FDR Info.
 *
 * [Args]
 *   (IN)  *sk        : socket (dummy)
 *   (OUT) *user_fdrs : FDRs info acquired from Kernel
 *   (OUT) *user_size : size of user_fdrs
 *
 * [Return]
 *   0      : OK.
 *  -ENOSYS : NG. (function not implemented)
 *   other  : (skip)
 **/
int tcpcp_call_get_res_hook(struct sock *sk, struct fdr *user_fdrs,
			    unsigned int *user_size)
{
	int error = -ENOSYS;

	/* plug-in. */
	plugin_dummy();

	/* start. */
	TCPCP_DPRINTK("start. sk=0x%p\n", sk);

#ifdef CONFIG_KMOD  
	/* hook is not activated ? */
	if (NULL == __tcpcp_config.tcpcp_get_res_hook) {
		TCPCP_DPRINTK("request TCPCP module.\n");
     /* This is protected by a lock internally. */
		request_module(TCPCP_MODULE_NAME);
	}
#endif

	/* hook is activated ? */
	if (NULL != __tcpcp_config.tcpcp_get_res_hook) {
		error = __tcpcp_config.tcpcp_get_res_hook(sk, user_fdrs, user_size);
  }

	/* end. */
	TCPCP_DPRINTK("end. error=%d\n", error);
	return error;
}

/**
 * tcpcp_call_set_si_hook - call hook function for setting TCP-SI to Kernel.
 *
 * [Args]
 *   (IN) *sk      : socket
 *   (IN) optname  : option name (TCPCP_SET_SI or TCPCP_SET_SI_ALL)
 *   (IN) *user_si : TCP-SI to set to Kernel (const)
 *   (IN) size     : size of user_si
 *
 * [Return]
 *   0      : OK.
 *  -ENOSYS : NG. (function not implemented)
 *   other  : (skip)
 **/
int tcpcp_call_set_si_hook(struct sock *sk, int optname,
			   const struct tcpcp_total_si_hdr *user_si,
			   unsigned int size)
{
	int error = -ENOSYS;

	/* plug-in. */
	plugin_dummy();

	/* start. */
	TCPCP_DPRINTK("start. sk=0x%p, optname=%d\n",
		sk, optname);

#ifdef CONFIG_KMOD
	/* hook is not activated ? */
	if (NULL == __tcpcp_config.tcpcp_set_si_hook) {
		TCPCP_DPRINTK("request TCPCP module.\n");
     /* This is protected by a lock internally. */
		request_module(TCPCP_MODULE_NAME);
	}
#endif

	/* tcpcp_set_si_hook is activated ? */
	if (NULL != __tcpcp_config.tcpcp_set_si_hook) {
		error = __tcpcp_config.tcpcp_set_si_hook(sk, optname, user_si, size);
  }

	/* end. */
	TCPCP_DPRINTK("end. error=%d\n", error);
	return error;
}

/**
 * tcpcp_call_close_hook - call hook function for TCPCP_CLOSE (not "ALL").
 *
 * [Args]
 *   (IN) *sk : TCP socket
 **/
void tcpcp_call_close_hook(struct sock *sk)
{
	/* plug-in. */
	plugin_dummy();

	/* start. */
	TCPCP_DPRINTK("start. sk=0x%p\n", sk);

	/* hook is not activate ? */
	if (NULL == __tcpcp_config.tcpcp_close_hook) {
		/* no need to request module. */
		return;
  }

	/* hook is activated ? */
	if (NULL != __tcpcp_config.tcpcp_close_hook) {
		__tcpcp_config.tcpcp_close_hook(sk);
  }

	/* end. */
	TCPCP_DPRINTK("end.\n");
	return;
}

/**
 * tcpcp_call_flush_hook - call hook function to flush socket or timeout.
 *
 * [Args]
 *   (IN)  *sk        : socket
 *   (OUT) *user_timeout : user-provided timeout for flush
 *
 * [Return]
 *   0      : OK.
 *  -ENOSYS : NG. (function not implemented)
 *   other  : (skip)
 **/
int tcpcp_call_flush_hook(struct sock *sk, struct timespec *user_timeout)
{
  int error = -ENOSYS;

  /* plug-in. */
  plugin_dummy();

  /* hook is not activate ? */
  if (NULL == __tcpcp_config.tcpcp_flush_hook) {
    /* no need to request module. */
    return error;
  }

  /* hook is activated ? */
  return __tcpcp_config.tcpcp_flush_hook(sk, user_timeout);
}

/* ----- Interface to ipv4/tcp_output.c ------------------------------------ */

/**
 * tcpcp_discard_xmit - queue packet transmission function change.
 *
 * [Args]
 *  (IN) *skb      : socket buffer
 *  (IN) ipfragok  : parameter used by function for exchange (not used)
 *                   Although ipfragok of the 2nd argument is not using
 *                   this function since this function is a pointer
 *                   function, it leaves in consideration of adjustment
 *                   with other functions.
 * [Return]
 *   0 : OK. (always in this case)
 **/
int tcpcp_discard_xmit(struct sk_buff *skb, int ipfragok)
{
	/* plug-in. */
	plugin_dummy();

	/* start. */
	TCPCP_DPRINTK("start. skb=0x%p\n", skb);

	kfree_skb(skb);

	/* end. */
	TCPCP_DPRINTK("end.\n");
	return 0;
}
EXPORT_SYMBOL(tcpcp_discard_xmit);


/* ----- Interface to ipv4/tcp_input.c ------------------------------------- */

struct sk_filter drop_all = {
	.refcnt = ATOMIC_INIT(1),	/* never kfree this ... */
	.len = 0,			/* return 0 */
};
EXPORT_SYMBOL(drop_all);

/**
 * tcpcp_fix_seq - corrects sequence number.
 *
 * [Args]
 *  (IN) *sk    : socket
 *  (IN) seq    : sequence number of packet transmitted from destination
 *  (IN) ackseq : ack sequence number of packet transmitted from destination
 **/
void tcpcp_fix_seq(struct sock *sk, __u32 seq, __u32 ackseq)
{
  struct tcp_sock *tp = tcp_sk(sk);

  /* plug-in. */
  plugin_dummy();

  /* start. */
  TCPCP_DPRINTK("start. sk=0x%p\n", sk);

  /* @@@ for DEBUG @@@ */
  TCPCP_DPRINTK("@@@ before fix seq. @@@\n");
  TCPCP_DPRINTK("tp->snd_nxt=%u\n", tp->snd_nxt);
  TCPCP_DPRINTK("ackseq     =%u\n", ackseq);
  TCPCP_DPRINTK("tp->rcv_nxt=%u\n", tp->rcv_nxt);
  TCPCP_DPRINTK("seq        =%u\n", seq);

  /* If next seq to send is stale (other end acks later send) */
  if (before(tp->snd_nxt, ackseq)) {
    /* XXX: should we be purging only up to ackseq if possible? We
       should only purge if there is missing data. In some cases, some
       of the data might need to be resent. */
    TCPCP_DPRINTK("purging write queue...\n");

    /* purge write_queue. */
    tcp_write_queue_purge(sk);
	  
    /* initialize packets_out, sk_send_head */
    tp->packets_out = 0;
    sk->sk_send_head = NULL;
	  
    /* fix sequence number (send). */
    tp->snd_nxt = ackseq;
    tp->snd_una = tp->snd_nxt;
    tp->write_seq = tp->snd_nxt;
    tp->snd_sml = tp->snd_nxt;
    tp->pushed_seq = tp->snd_nxt;
  }

  /* If expected seq to recv is stale (other end sending later) */
  if (before(tp->rcv_nxt, seq)) {
    /* XXX: should we be purging only up to seq? */
    TCPCP_DPRINTK("purginging recv queue...\n");

    /* purge receive_queue. */
    __skb_queue_purge(&sk->sk_receive_queue);
    __skb_queue_purge(&tp->out_of_order_queue);
#ifdef CONFIG_NET_DMA
    __skb_queue_purge(&sk->sk_async_wait_queue);
#endif
    /* fix sequence number (receive). */
    tp->rcv_nxt = seq;
    tp->rcv_wup = tp->rcv_nxt;
    tp->snd_wl1 = tp->rcv_nxt;
    tp->copied_seq = tp->rcv_nxt;
  }

  /* @@@ for DEBUG @@@ */
  TCPCP_DPRINTK("@@@ after fix seq. @@@\n");
  TCPCP_DPRINTK("tp->snd_nxt=%u\n", tp->snd_nxt);
  TCPCP_DPRINTK("tp->snd_una=%u\n", tp->snd_una);
  TCPCP_DPRINTK("tp->rcv_nxt=%u\n", tp->rcv_nxt);

  /* end. */
  TCPCP_DPRINTK("end.\n");
  return;
}
EXPORT_SYMBOL(tcpcp_fix_seq);

/**
 * tcpcp_fix_ts - corrects timestamp.
 *
 * [Args]
 *  (IN) *sk       : socket
 *  (IN) ack_tsecr : timestamp of the latest packet that succeeds in reception
 **/
void tcpcp_fix_ts(struct sock *sk, __u32 ack_tsecr)
{
	struct tcp_sock *tp = tcp_sk(sk);

	/* plug-in. */
	plugin_dummy();

	/* start. */
	TCPCP_DPRINTK("start. sk=0x%p\n", sk);

	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("@@@ before fix ts. @@@\n");
	TCPCP_DPRINTK("ack_tsecr     =%u\n", ack_tsecr);
	TCPCP_DPRINTK("tp->ts_offset =%u\n",
		tp->ts_offset);
	TCPCP_DPRINTK("tcp_time_stamp=%u\n",
		tcp_time_stamp);
	TCPCP_DPRINTK("tp_time_stamp =%u\n",
		tp_time_stamp(tp));

	/* Offset needs to be adjusted by difference between ack and
	 * current timestamp. We cannot simply set ts_offset to the
	 * difference between the ack and the current tcp_time_stamp
	 * because the ts_offset really is the offset to the original
	 * tcp_time_stamp used on the host where the connection was
	 * first made.
	 */
	tp->ts_offset -= (ack_tsecr -tcp_time_stamp);

	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("@@@ after fix ts. @@@\n");
	TCPCP_DPRINTK("tp->ts_offset = ack_tsecr - tcp_time_stamp!\n");
	TCPCP_DPRINTK("tp_time_stamp(tp) ="
		" tcp_time_stamp + tp->ts_offset\n");
	TCPCP_DPRINTK("tp->ts_offset=%u\n", tp->ts_offset);
	TCPCP_DPRINTK("tcp_time_stamp=%u\n",
		tcp_time_stamp);
	TCPCP_DPRINTK("tp_time_stamp =%u\n",
		tp_time_stamp(tp));

	/* end. */
	TCPCP_DPRINTK("end.\n");
	return;
}
EXPORT_SYMBOL(tcpcp_fix_ts);

#endif /* defined(CONFIG_TCPCP) || defined(CONFIG_TCPCP_MODULE) */
