/*
 * tcpcp.c - TCP connection passing
 *
 * Written 2002,2004 by Werner Almesberger
 *
 * Copyright (C) 2005 NTT Corporation
 */


#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/fs.h>
#include <linux/errno.h>
#include <linux/slab.h>
#include <linux/init.h>
#include <linux/net.h>
#include <linux/tcp.h>
#include <linux/security.h>
#include <linux/wait.h>
#include <linux/tcpcp.h>
#include <net/sock.h>
#include <net/tcp.h>
#include <net/inet_common.h>
#include <net/ipv6.h>
#include <asm/uaccess.h>

#include <linux/file.h>
#include <linux/fdtable.h>
#include <linux/syscalls.h>
#if defined(CONFIG_IPV6) || defined(CONFIG_IPV6_MODULE)
#include <net/ip6_route.h>
#include <net/transp_v6.h>
#include <net/addrconf.h>
#endif

#define TCPCP_CURRENT_SI_MAJOR	0
#define TCPCP_CURRENT_SI_MINOR	0

MODULE_LICENSE("GPL");

static struct tcpcp_list proc_list;
static struct tcpcp_list timeout_list;

 /* These values are defined in net/core/sock.c
  * We try to get the values using sysctl; otherwise,
  * we use the small constants.
  */
static int __sysctl_wmem_max = 32767;
static int __sysctl_rmem_max = 32767;

static void tcpcp_watch_timeout(unsigned long);
static struct timer_list tcpcp_timer
	= TIMER_INITIALIZER(tcpcp_watch_timeout, 0, 0);

 /* defined in tcp_input.c */
extern int tcpcp_recover_rcvbuf(struct sock *sk,
                                const struct tcpcp_si_skb_buf *rcv_buf);

DECLARE_RWSEM(tcpcp_sem);

#ifdef CONFIG_TCPCP_DEBUG
/* @@@ for DEBUG @@@ */
static void __print_list(void);
static void print_list(void);
static void skb_queue_check(struct sk_buff_head *queue);
static void check_fd(void);
#endif

 /* taken from net/core/sock.c */
inline static void sock_valbool_flag(struct sock *sk, int bit, int valbool)
{
        if (valbool)
                sock_set_flag(sk, bit);
        else
                sock_reset_flag(sk, bit);
}

/* ----- check socket ------------------------------------------------------ */

/**
 * tcpcp_check_sk_type - check type of TCP socket.
 *
 * [Args]
 *   (IN) *sk : TCP socket
 *
 * [Return]
 *   0               : OK.
 *  -EPFNOSUPPORT    : NG.
 *  -ESOCKTNOSUPPORT : NG.
 *  -EPROTONOSUPPORT : NG.
 **/
static int tcpcp_check_sk_type(struct sock *sk)
{
	/* plug-in. */
	plugin_dummy();

	/* start. */
	TCPCP_DPRINTK("start. sk=0x%p\n", sk);

#if defined(CONFIG_IPV6) || defined(CONFIG_IPV6_MODULE)
	if ((sk->sk_family != PF_INET) && (sk->sk_family != PF_INET6)) {
#else
	if (sk->sk_family != PF_INET) {
#endif
		TCPCP_DPRINTK("The specified socket family is not" \
			" supported.\n");
		return -EPFNOSUPPORT;
	}

	if (sk->sk_type != SOCK_STREAM) {
		TCPCP_DPRINTK("The specified socket type is not" \
			" supported.\n");
		return -ESOCKTNOSUPPORT;
	}

	if (sk->sk_protocol != IPPROTO_TCP) {
		TCPCP_DPRINTK("The specified socket protocol is not" \
			" supported.\n");
		return -EPROTONOSUPPORT;
	}

	/* end. */
	TCPCP_DPRINTK("end.\n");
	return 0;
}

/**
 * tcpcp_check_sk_state - check state of TCP socket.
 *
 * [Args]
 *   (IN) *sk  : TCP socket
 *   (IN) flag : TCPCP operation flag
 *
 * [Return]
 *   0      : OK.
 *  -EBADFD : NG.
 **/
static int tcpcp_check_sk_state(struct sock *sk, unsigned char flag)
{
	unsigned char cflag;

	/* plug-in. */
	plugin_dummy();

	/* start. */
	TCPCP_DPRINTK("start. sk=0x%p, flag=%u\n",
		sk, flag);

	/* The bit which shows ALL is set as 0. */
	cflag = flag & ~TCPCP_FLAG_ALL;

	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("sk->sk_socket->state=%d\n",
		sk->sk_socket->state);
	TCPCP_DPRINTK("sk->sk_state=%d\n",
		sk->sk_state);

	switch (cflag) {
	case TCPCP_FLAG_STOP:
		if ((sk->sk_socket->state != SS_CONNECTING)
		    && (sk->sk_socket->state != SS_CONNECTED)) {
			TCPCP_DPRINTK("Invalid socket state" \
				"(sk->sk_socket->state).\n");
			return -EBADFD;
		}
		if (sk->sk_state == TCP_CLOSE) {
			TCPCP_DPRINTK("Invalid socket state" \
				"(sk->sk_state).\n");
			return -EBADFD;
		}
		break;

	case TCPCP_FLAG_MAKE_SI:
		if ((sk->sk_socket->state != SS_CONNECTING)
		    && (sk->sk_socket->state != SS_CONNECTED)) {
			TCPCP_DPRINTK("Invalid socket state" \
				"(sk->sk_socket->state).\n");
			return -EBADFD;
		}
		if (sk->sk_state != TCP_ESTABLISHED) {
			TCPCP_DPRINTK("Invalid socket state" \
				"(sk->sk_state).\n");
			return -EBADFD;
		}
		break;

	case TCPCP_FLAG_SET_SI:
		if (!(flag & TCPCP_FLAG_ALL)) {
			if (sk->sk_state != TCP_CLOSE) {
				TCPCP_DPRINTK("Invalid socket state" \
					"(sk->sk_state).\n");
				return -EBADFD;
			}
		}
		break;

	case TCPCP_FLAG_START:
	case TCPCP_FLAG_CLOSE:
		break;

	default:
		TCPCP_DPRINTK("Invalid operation flag.\n");
		return -EBADFD;
	}

	/* end. */
	TCPCP_DPRINTK("end.\n");
	return 0;
}

/**
 * tcpcp_get_sk_i - check FD and get SOCKET_I().
 *
 * [Args]
 *   (IN) *sk      : TCP socket
 *   (IN) fd       : FD
 *   (IN) *r_entry : Request Entry (to get flag "ONE" / "ALL" and set FD)
 *
 * [Return]
 *   normal : SOCKET_I()
 *   NULL   : not target FD
 **/
static struct sock *tcpcp_get_sk_i(struct sock *sk, int fd,
				   struct tcpcp_req_entry *r_entry)
{
	struct files_struct *files = current->files;
	struct file *filp;
	struct sock *sk_i;
	unsigned char flag = r_entry->flag;
	int error = 0;

	/* plug-in. */
	plugin_dummy();

	/* start. */
	TCPCP_DPRINTK("start. sk=0x%p, fd=%d, r_entry=0x%p\n",
		sk, fd, r_entry);

	/* get file pointer. */
	filp = fcheck_files(files, fd);
	if (filp == NULL) {
		TCPCP_DPRINTK("Can't get file pointer.\n");
		return NULL;
	}

	if (!S_ISSOCK(filp->f_dentry->d_inode->i_mode)) {
		TCPCP_DPRINTK("Can't get socket.\n");
		return NULL;
	}

	/* get socket from FD. */
	sk_i = SOCKET_I(filp->f_dentry->d_inode)->sk;
	if (sk_i == NULL) {
		TCPCP_DPRINTK("Can't get SOCKET_I.\n");
		return NULL;
	}

	if (flag & TCPCP_FLAG_ALL) {
		/* check dummy FD. (only "ALL") */
		if (sk_i == sk) {
			r_entry->fd = fd;
			TCPCP_DPRINTK("Dummy FD number is %d.\n",
				fd);
			return NULL;
		}

		/* check type of socket. (only "ALL") */
		error = tcpcp_check_sk_type(sk_i);
		if (error != 0) {
			TCPCP_DPRINTK("This socket is not target.\n");
			return NULL;
		}

		/* check state of TCP socket. (only "ALL") */
		error = tcpcp_check_sk_state(sk_i, flag);
		if (error != 0) {
			TCPCP_DPRINTK("This socket is not target.\n");
			return NULL;
		}
	} else {
		/* check target FD. (only "ONE") */
		if (sk_i != sk) {
			TCPCP_DPRINTK("This socket is not target.\n");
			return NULL;
		}

		r_entry->fd = fd;
		TCPCP_DPRINTK("Target FD number is %d.\n",
			fd);
	}

	/* end. */
	TCPCP_DPRINTK("end. sk_i=0x%p\n", sk_i);
	return sk_i;
}


/* ----- Entry management (non-API) ---------------------------------------- */

/* ----- common entry ----- */

/**
 * tcpcp_add_com_entry - add and change common entry.
 *
 * [Args]
 *   (IN) flag        : "ONE" / "ALL"
 *   (IN/OUT) **entry : (IN)  now common entry (or NULL)
 *                      (OUT) new common entry
 *
 * [Return]
 *   0      : OK.
 *  -ENOMEM : NG.
 **/
static int tcpcp_add_com_entry(unsigned char flag,
			       struct tcpcp_com_entry **entry)
{
	struct tcpcp_com_entry *new_entry;
	unsigned int entry_size;

	/* plug-in. */
	plugin_dummy();

	/* start. */
	TCPCP_DPRINTK("start. flag=%u\n", flag);

#ifdef CONFIG_TCPCP_DEBUG
	/* @@@ print Process List with lock. (for DEBUG) @@@ */
	print_list();
#endif

	/* select entry size. */
	if (flag & TCPCP_FLAG_ALL)
		entry_size = TCPCP_COM_ENTRY_SIZE_FOR_ALL;
	else
		entry_size = TCPCP_COM_ENTRY_SIZE_FOR_ONE;
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("entry_size=%u\n", entry_size);

	/* allocate common entry. */
	new_entry = kmalloc(entry_size, GFP_KERNEL);
	if (new_entry == NULL) {
		TCPCP_DPRINTK("[Error] can't allocate common entry.\n");
		return -ENOMEM;
	}

	/* initialize common entry. */
	memset(new_entry, 0, entry_size);
	new_entry->next = NULL;
	new_entry->max_data_len = entry_size - sizeof(struct tcpcp_com_entry);
	new_entry->data_len = 0;
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("new_entry=0x%p\n",
		new_entry);
	TCPCP_DPRINTK("new_entry->max_data_len=%u\n",
		new_entry->max_data_len);
	TCPCP_DPRINTK("new_entry->data_len    =%u\n",
		new_entry->data_len);


	if ((*entry) != NULL) {
		/* link common entry. */
		(*entry)->next = new_entry;
		/* @@@ for DEBUG @@@ */
		TCPCP_DPRINTK(" *entry       =0x%p\n",
			*entry);
		TCPCP_DPRINTK("(*entry)->next=0x%p\n",
			(*entry)->next);
	}

	/* change now entry. */
	(*entry) = new_entry;

#ifdef CONFIG_TCPCP_DEBUG
	/* @@@ print Process List with lock. (for DEBUG) @@@ */
	print_list();
#endif

	/* end. */
	TCPCP_DPRINTK("end.\n");
	return 0;
}

/**
 * tcpcp_del_all_com_entry - delete all common entry.
 *
 * [Args]
 *   (IN/OUT) **entry : (IN) first common entry
 *                      (OUT) NULL
 **/
static void tcpcp_del_all_com_entry(struct tcpcp_com_entry **entry)
{
	struct tcpcp_com_entry *next_entry;

	/* plug-in. */
	plugin_dummy();

	/* start. */
	TCPCP_DPRINTK("start.\n");

#ifdef CONFIG_TCPCP_DEBUG
	/* @@@ print Process List without lock. (for DEBUG) @@@ */
	__print_list();
#endif

	while ((*entry) != NULL) {
		/* free common entry. */
		next_entry = (*entry)->next;
		kfree(*entry);
		(*entry) = next_entry;
		/* @@@ for DEBUG @@@ */
		TCPCP_DPRINTK("common entry is deleted.\n");
	}

#ifdef CONFIG_TCPCP_DEBUG
	/* @@@ print Process List without lock. (for DEBUG) @@@ */
	__print_list();
#endif

	/* end. */
	TCPCP_DPRINTK("end.\n");
	return;
}

/* ----- Process Entry ----- */

/**
 * tcpcp_add_proc_entry - allocate and add Process Entry.
 *
 * [Args]
 *   (IN) *sk  : TCP socket
 *   (IN) flag : TCPCP operation flag
 *
 * [Return]
 *   normal : new Process Entry
 *   NULL   : -ENOMEM
 **/
static struct tcpcp_proc_entry *tcpcp_add_proc_entry(struct sock *sk,
						     unsigned char flag)
{
	struct tcpcp_proc_entry *p_entry;
	unsigned int p_entry_size;

	/* plug-in. */
	plugin_dummy();

	/* start. */
	TCPCP_DPRINTK("start. sk=0x%p, flag=%u\n",
		sk, flag);

	/* allocate Process Entry. */
	p_entry_size = sizeof(struct tcpcp_proc_entry);
	p_entry = kmalloc(p_entry_size, GFP_ATOMIC);
	if (p_entry == NULL) {
		TCPCP_DPRINTK("[Error] can't allocate Process Entry.\n");
		return NULL;
	}

	/* initialize Process Entry. */
	memset(p_entry, 0, p_entry_size);
	p_entry->tgid = current->tgid;
	p_entry->flag = flag;
	INIT_LIST_HEAD(&(p_entry->r_list_head));

	/* link Process Entry to Process List. */
	list_add_tail(&(p_entry->node), &(proc_list.head));

	/* end. */
	TCPCP_DPRINTK("end. p_entry=0x%p\n", p_entry);
	return p_entry;
}

/**
 * tcpcp_del_proc_entry
 *   - delete and free Process Entry, if it doesn't have Request Entry.
 *
 * [Args]
 *   (IN) *p_entry : Process Entry (not have Request Entry)
 **/
void tcpcp_del_proc_entry(struct tcpcp_proc_entry *p_entry)
{
	/* plug-in. */
	plugin_dummy();

	/* start. */
	TCPCP_DPRINTK("start.\n");

	/* check that Process Entry is NULL. */
	if (p_entry == NULL) {
		TCPCP_DPRINTK("Process entry is NULL.\n");
		return;
	}

	/* check that Process Entry has Request Entry. */
	if (list_empty(&(p_entry->r_list_head)) == 0) {
		TCPCP_DPRINTK("Process entry still holds some" \
			" Request Entries elsewhere.\n");
		return;
	}

	/* unlink Process Entry from Process List and free. */
	list_del_init(&(p_entry->node));
	kfree(p_entry);
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("Process Entry is deleted.\n");

#ifdef CONFIG_TCPCP_DEBUG
	/* @@@ print Process List without lock. (for DEBUG) @@@ */
	__print_list();
#endif

	/* end. */
	TCPCP_DPRINTK("end.\n");
	return;
}

/* ----- Request Entry ----- */

/**
 * __tcpcp_add_req_entry - allocate and add Request Entry.
 *
 * [Args]
 *   (IN) *sk      : TCP socket
 *   (IN) flag     : TCPCP operation flag
 *   (IN) *p_entry : Process Entry
 *
 * [Return]
 *   normal : new Request Entry
 *   NULL   : -ENOMEM
 **/
static struct tcpcp_req_entry *__tcpcp_add_req_entry(struct sock *sk,
			unsigned char flag, struct tcpcp_proc_entry *p_entry)
{
	struct tcpcp_req_entry *r_entry;
	unsigned int r_entry_size;

	/* plug-in. */
	plugin_dummy();

	/* start. */
	TCPCP_DPRINTK("start. sk=0x%p, flag=%u, p_entry=0x%p\n",
		sk, flag, p_entry);

	/* allocate Request Entry. */
	r_entry_size = sizeof(struct tcpcp_req_entry);
	r_entry = kmalloc(r_entry_size, GFP_ATOMIC);
	if (r_entry == NULL) {
		TCPCP_DPRINTK("[Error] can't allocate Request Entry.\n");

		/* delete Process Entry. */
		tcpcp_del_proc_entry(p_entry);
		return NULL;
	}

	/* initialize Request Entry. */
	memset(r_entry, 0, r_entry_size);
	r_entry->sk = sk;
	r_entry->flag = flag;
	r_entry->pid = current->pid;
	r_entry->s_entry = NULL;
	r_entry->f_entry = NULL;
	INIT_LIST_HEAD(&(r_entry->timer_node));
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("r_entry->sk=0x%p\n",
		r_entry->sk);
	TCPCP_DPRINTK("r_entry->flag=%d\n",
		r_entry->flag);
	TCPCP_DPRINTK("r_entry->pid=%d\n",
		r_entry->pid);
	TCPCP_DPRINTK("r_entry->s_entry=0x%p\n",
		r_entry->s_entry);
	TCPCP_DPRINTK("r_entry->f_entry=0x%p\n",
		r_entry->f_entry);

	/* set parent Process Entry and lint Request Entry to Request List. */
	r_entry->p_entry = p_entry;
	list_add_tail(&(r_entry->node), &(p_entry->r_list_head));
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("r_entry->p_entry=0x%p\n",
		r_entry->p_entry);

	/* end. */
	TCPCP_DPRINTK("end. r_entry=0x%p\n", r_entry);
	return r_entry;
}

/**
 * __tcpcp_del_req_entry
 *   - delete and free Request Entry with common entries.
 *     (and delete and free parent Process Entry, if it doesn't have Request
 *      Entry.)
 *
 * [Args]
 *   (IN) *r_entry : now Request Entry
 **/
static void __tcpcp_del_req_entry(struct tcpcp_req_entry *r_entry)
{
	struct tcpcp_proc_entry *p_entry;

	/* plug-in. */
	plugin_dummy();

	/* start. */
	TCPCP_DPRINTK("start.\n");

	/* check that Request Entry is NULL. */
	if (r_entry == NULL) {
		TCPCP_DPRINTK("Request entry is NULL.\n");
		return;
	}

	/* get parent Process Entry. */
	p_entry = r_entry->p_entry;

	/* delete TCP-SIs Entry and FDR Info. Entry. */
	tcpcp_del_all_com_entry(&(r_entry->s_entry));
	tcpcp_del_all_com_entry(&(r_entry->f_entry));

	/* unlink Request Entry from Timeout List. */
	list_del_init(&(r_entry->timer_node));

	/* unlink Request Entry from Request List and free. */
	list_del_init(&(r_entry->node));
	kfree(r_entry);

	/* delete Process Entry. */
	tcpcp_del_proc_entry(p_entry);

	/* end. */
	TCPCP_DPRINTK("end.\n");
	return;
}

/* ----- check conflict ----- */

/**
 * tcpcp_check_conflict_one_all
 *   - search Request List and check conflict. (only "ONE/ALL")
 *
 * [Args]
 *   (IN) *sk     : TCP socket
 *   (IN) flag    : TCPCP operation flag
 *   (IN) p_entry : Process Entry
 *
 * [Return]
 *   0     : OK. (not conflict)
 *  -EBUSY : NG. (conflict)
 **/
static int tcpcp_check_conflict_one_all(struct sock *sk, unsigned char flag,
					struct tcpcp_proc_entry *p_entry)
{
	struct tcpcp_req_entry *r_entry;
	unsigned char cflag;

	/* plug-in. */
	plugin_dummy();

	/* start. */
	TCPCP_DPRINTK("start. sk=0x%p, flag=%u, p_entry=0x%p\n",
		sk, flag, p_entry);

	/* search Request List. */
	list_for_each_entry(r_entry, &(p_entry->r_list_head), node) {
		/* lookup Request Entry what has same socket. */
		if (r_entry->sk == sk) {
			TCPCP_DPRINTK("[Error] conflict has occurred"
				" between batch processing and individual"
				" processing.\n");
			return -EBUSY;
		}
		/* The bit which shows ALL is set as 0. */
		cflag = r_entry->flag & ~TCPCP_FLAG_ALL;

		/* check conflict. */
		switch (cflag) {
		case TCPCP_FLAG_MAKE_SI:
			if ((flag & TCPCP_FLAG_CLOSE)
			    || (flag & TCPCP_FLAG_SET_SI)) {
				TCPCP_DPRINTK("[Error] conflict has"
				 	" occurred, because TCPCP close"
					" processing or TCPCP set SI"
					" processing are under operation.\n");
				return -EBUSY;
			}
			break;

		case TCPCP_FLAG_STOP:
		case TCPCP_FLAG_START:
			if (flag & TCPCP_FLAG_MAKE_SI)
				break;
			TCPCP_DPRINTK("[Error] conflict has occurred,"
				" because TCPCP stop processing or TCPCP start"
				" processing are under operation.\n");
			return -EBUSY;

		case TCPCP_FLAG_CLOSE:
		case TCPCP_FLAG_SET_SI:
		default:
			TCPCP_DPRINTK("[Error] conflict has occurred,"
				" because TCPCP close processing or TCPCP"
				" set-SIs processing are under operation.\n");
			return -EBUSY;
		}
	}

	/* end. */
	TCPCP_DPRINTK("end.\n");
	return 0;
}

/**
 * tcpcp_check_conflict_one_one
 *   - search Request List and check conflict (only "ONE/ONE").
 *
 * [Args]
 *   (IN) *sk      : TCP socket
 *   (IN) *p_entry : Process Entry
 *
 * [Return]
 *   0     : OK. (not conflict)
 *  -EBUSY : NG. (conflict)
 **/
static int tcpcp_check_conflict_one_one(struct sock *sk,
					struct tcpcp_proc_entry *p_entry)
{
	struct tcpcp_req_entry *r_entry;

	/* plug-in. */
	plugin_dummy();

	/* start. */
	TCPCP_DPRINTK("start. sk=0x%p, p_entry=0x%p\n",
		sk, p_entry);

	/* search Request Entry. */
	list_for_each_entry(r_entry, &(p_entry->r_list_head), node) {
		/* lookup Request Entry what has same socket. */
		if (r_entry->sk == sk) {
			TCPCP_DPRINTK("[Error] conflict has occurred"
				" in individual processings.\n");
			return -EBUSY;
		}
	}

	/* end. */
	TCPCP_DPRINTK("end.\n");
	return 0;
}

/**
 * tcpcp_check_conflict - check conflict and add Process Entry.
 *
 * [Args]
 *   (IN) *sk        : TCP socket
 *   (IN) flag       : TCPCP operation flag
 *   (OUT) **p_entry : Process Entry
 *
 * [Return]
 *   0               : OK.
 *  -EBUSY           : NG. (conflict)
 *  -ENOMEM          : NG. (no memory)
 **/
static int tcpcp_check_conflict(struct sock *sk, unsigned char flag,
				struct tcpcp_proc_entry **p_entry)
{
	unsigned char p_flag;
	int error;

	/* plug-in. */
	plugin_dummy();

	/* start. */
	TCPCP_DPRINTK("start. sk=0x%p, flag=%u\n",
		sk, flag);

	/* search Process List. */
	list_for_each_entry((*p_entry), &(proc_list.head), node) {
		/* lookup Process Entry what has same Thread Group ID. */
		if ((*p_entry)->tgid == current->tgid) {
			TCPCP_DPRINTK("lookup Process Entry.\n");
			goto search_req_entry;
		}
	}
	TCPCP_DPRINTK("can't lookup Process Entry.\n");

	/* add Process Entry. */
	(*p_entry) = tcpcp_add_proc_entry(sk, flag);
	if ((*p_entry) == NULL) {
		TCPCP_DPRINTK("[Error] can't add Process Entry.\n");
		return -ENOMEM;
	}
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("Process Entry is added.\n");

#ifdef CONFIG_TCPCP_DEBUG
	/* @@@ print Process List without lock. (for DEBUG) @@@ */
	__print_list();
#endif

	/* end. */
	TCPCP_DPRINTK("end.\n");
	return 0;

search_req_entry:
	/* check conflict. */
	p_flag = (*p_entry)->flag;

	if ((p_flag & TCPCP_FLAG_ALL) && (flag & TCPCP_FLAG_ALL)) {
		/* conflict each batch processing. */
		TCPCP_DPRINTK("[Error] conflict has occurred in batch"
			" processing.\n");
		error = -EBUSY;

	} else if ((p_flag & TCPCP_FLAG_ALL) || (flag & TCPCP_FLAG_ALL))
		/* conflict batch processing and individual processing. */
		error = tcpcp_check_conflict_one_all(sk, flag, (*p_entry));

	else
		/* conflict each individual processing. */
		error = tcpcp_check_conflict_one_one(sk, (*p_entry));

	/* end. */
	TCPCP_DPRINTK("end. error=%d\n", error);
	return error;
}

/**
 * tcpcp_lookup_req_entry - lookup Request Entry.
 *
 * [Args]
 *   (IN) *sk : TCP socket
 *
 * [Return]
 *   normal : Request Entry
 *   NULL   : not lookup
 **/
static struct tcpcp_req_entry *tcpcp_lookup_req_entry(struct sock *sk)
{
	struct tcpcp_proc_entry *p_entry = NULL;
	struct tcpcp_req_entry *r_entry = NULL;

	/* plug-in. */
	plugin_dummy();

	/* start. */
	TCPCP_DPRINTK("start. sk=0x%p\n", sk);

	/* search Process List. */
	list_for_each_entry(p_entry, &(proc_list.head), node) {
		/* lookup Process Entry what has same Thread Group ID. */
		if (p_entry->tgid == current->tgid)
			goto check_sk;
	}

	TCPCP_DPRINTK("[Error] there is no information to get."
		" (no Process Entry)\n");
	return NULL;

check_sk:
	/* search Request List. */
	TCPCP_DPRINTK("search Request List.\n");
	list_for_each_entry(r_entry, &(p_entry->r_list_head), node) {
		/* lookup Request Entry what has same socket. */
		if (r_entry->sk == sk)
			goto check_pid;
	}

	TCPCP_DPRINTK("[Error] there is no information to get."
		" (no Request Entry)\n");
	return NULL;

check_pid:
	TCPCP_DPRINTK("check process ID.\n");
	if (r_entry->pid != current->pid) {
		TCPCP_DPRINTK("[Error] there is no information to get."
			" (different PID)\n");
		return NULL;
	}

	/* end. */
	TCPCP_DPRINTK("end. r_entry=0x%p\n", r_entry);
	return r_entry;
}

/* ----- Timeout-watcher ----- */

/**
 * tcpcp_watch_timeout - watch whether the time-out generation.
 **/
static void tcpcp_watch_timeout(unsigned long arg)
{
	struct tcpcp_req_entry *r_entry, *next_r_entry;
	__u32 cur_time = tcp_time_stamp;

	/* plug-in. */
	plugin_dummy();

	/* start. */
	TCPCP_DPRINTK("start.\n");

	/* lock Process List (and Timeout List). */
	if (!write_trylock(&(proc_list.lock))) {
		TCPCP_DPRINTK("failed lock for Process List.\n");
		mod_timer(&tcpcp_timer,
			  tcp_time_stamp + TCPCP_RETRY_WATCH_INTERVAL);
		return;
	}
	TCPCP_DPRINTK("Process List is locked.\n");

#ifdef CONFIG_TCPCP_DEBUG
	/* @@@ print Process List without lock. (for DEBUG) @@@ */
	__print_list();
#endif

	/* search Timeout Request List. */
	list_for_each_entry_safe(r_entry, next_r_entry, &(timeout_list.head),
				 timer_node) {
		/* @@@ for DEBUG @@@ */
		TCPCP_DPRINTK("r_entry->timeout=%d\n",
			r_entry->timeout);

		if (r_entry->timeout > cur_time) {
			TCPCP_DPRINTK("search is finished.\n");
			break;
		}

		/* delete Request Entry with common entries. */
		__tcpcp_del_req_entry(r_entry);
		/* @@@ for DEBUG @@@ */
		TCPCP_DPRINTK("delete Request Entry with common"
			" entries.\n");
	}

#ifdef CONFIG_TCPCP_DEBUG
	/* @@@ print Process List without lock. (for DEBUG) @@@ */
	__print_list();
#endif

	if ((list_empty(&(timeout_list.head)) == 0)
	    && (timer_pending(&tcpcp_timer) == 0)) {
		/* regist Timeout-watcher. */
		tcpcp_timer.expires = tcp_time_stamp + TCPCP_WATCH_INTERVAL;
		add_timer(&tcpcp_timer);
		TCPCP_DPRINTK("Timeout watcher is registered.\n");
	}

	/* release Process List (and Timeout List). */
	write_unlock(&(proc_list.lock));
	TCPCP_DPRINTK("Process List is unlocked.\n");

	/* end. */
	TCPCP_DPRINTK("end.\n");
	return;
}

/**
 * tcpcp_regist_timer - regist timer.
 **/
static void tcpcp_regist_timer(void)
{
	/* plug-in. */
	plugin_dummy();

	/* start. */
	TCPCP_DPRINTK("start.\n");

	/* regist Timeout-watcher. */
	tcpcp_timer.expires = tcp_time_stamp + TCPCP_WATCH_INTERVAL;
	add_timer(&tcpcp_timer);
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("tcpcp_timer.expires=%lu\n",
		tcpcp_timer.expires);

	/* end. */
	TCPCP_DPRINTK("end.\n");
	return;
}


/* ----- Entry management (API) -------------------------------------------- */
 
/* ----- common entry ----- */

/**
 * tcpcp_get_w_ptr : get write-point of common entry.
 *
 * [Args]
 *   (IN) flag         : "ONE" / "ALL" (to add common entry)
 *   (IN) new_data_len : length of new data
 *   (IN/OUT) **entry  : now common entry
 *
 * [Return]
 *   normal : write-point
 *   NULL   : -ENOMEM
 **/
static char *tcpcp_get_w_ptr(unsigned char flag, unsigned int new_data_len,
			     struct tcpcp_com_entry **entry)
{
	char *w_ptr;
	int error;

	/* plug-in. */
	plugin_dummy();

	/* start. */
	TCPCP_DPRINTK("start. flag=%d, new_data_len=%u\n", 
		flag, new_data_len);

	/* check 1st entry. */
	/* check free of common entry. */
	if (((*entry) == NULL)
	    || (*entry)->max_data_len - (*entry)->data_len < new_data_len) {
		/* add and change common entry. */
		error = tcpcp_add_com_entry(flag, entry);
		if (error != 0) {
			TCPCP_DPRINTK("[Error] can't add common entry."
				" error=%d\n", error);
			return NULL;
		}
		TCPCP_DPRINTK("common entry is added.\n");
	}

	/* get write-point. */
	w_ptr = (*entry)->data + (*entry)->data_len;
	(*entry)->data_len += new_data_len;
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("(*entry)->data_len=%u\n",
		(*entry)->data_len);

	/* end. */
	TCPCP_DPRINTK("end. w_ptr=0x%p\n", w_ptr);
	return w_ptr;
}

/* ----- Request Entry ----- */

/**
 * tcpcp_add_req_entry - check conflict and add Request Entry.
 *
 * [Args]
 *   (IN) *sk        : TCP socket
 *   (IN) flag       : TCPCP operation flag
 *   (OUT) **r_entry : Request Entry
 *
 * [Return]
 *   0               : OK.
 *  -EPFNOSUPPORT    : NG. (not support socket family)
 *  -ESOCKTNOSUPPORT : NG. (not support socket type)
 *  -EPROTONOSUPPORT : NG. (not support socket protocol)
 *  -EBADFD          : NG. (bad state of TCP socket)
 *  -EBUSY           : NG. (conflict)
 *  -ENOMEM          : NG. (no memory)
 **/
static int tcpcp_add_req_entry(struct sock *sk, unsigned char flag,
			       struct tcpcp_req_entry **r_entry)
{
	struct tcpcp_proc_entry *p_entry;
	int error;

	/* plug-in. */
	plugin_dummy();

	/* start. */
	TCPCP_DPRINTK("start. sk=0x%p, flag=%d\n",
		sk, flag);

	(*r_entry) = NULL;

	/* check type of socket. */
	error = tcpcp_check_sk_type(sk);
	if (error != 0) {
		TCPCP_DPRINTK("end. error=%d\n", error);
		return error;
	}

	if (!(flag & TCPCP_FLAG_ALL)) {
		/* check state of TCP socket. (only "ONE") */
		error = tcpcp_check_sk_state(sk, flag);
		if (error != 0) {
			/* only -EBADFD. */
			TCPCP_DPRINTK("end. error=%d\n",
				error);
			return error;
		}
	}

	/* lock Process List. */
	write_lock(&(proc_list.lock));
	TCPCP_DPRINTK("Process List is locked.\n");

	/* check conflict and get Process Entry. */
	error = tcpcp_check_conflict(sk, flag, &p_entry);
	if (error != 0)
		goto out;

	/* add Request Entry. */
	(*r_entry) = __tcpcp_add_req_entry(sk, flag, p_entry);
	if ((*r_entry) == NULL)
		error = -ENOMEM;

out:
	/* release Process List. */
	write_unlock(&(proc_list.lock));
	TCPCP_DPRINTK("Process List is unlocked.\n");

	/* end. */
	TCPCP_DPRINTK("end. error=%d\n", error);
	return error;
}

/**
 * tcpcp_del_req_entry
 *   - delete and free Request Entry with common entries.
 *     (and delete and free parent Process Entry, if it doesn't have Request
 *      Entry.)
 *
 * [Args]
 *   (IN) *r_entry : Request Entry
 **/
static void tcpcp_del_req_entry(struct tcpcp_req_entry *r_entry)
{
	/* plug-in. */
	plugin_dummy();

	/* start. */
	TCPCP_DPRINTK("start. r_entry=0x%p\n", r_entry);

	/* lock Process List. */
	write_lock(&(proc_list.lock));
	TCPCP_DPRINTK("Process List is locked.\n");

	/* delete Request Entry. */
	__tcpcp_del_req_entry(r_entry);

	/* release Process List. */
	write_unlock(&(proc_list.lock));
	TCPCP_DPRINTK("Process List is unlocked.\n");

	/* end. */
	TCPCP_DPRINTK("end.\n");
	return;
}

/**
 * tcpcp_del_all_entry - delete and free all entry.
 **/
static void tcpcp_del_all_entry(void)
{
	struct tcpcp_proc_entry *p_entry, *next_p_entry;
	struct tcpcp_req_entry *r_entry, *next_r_entry;

	/* plug-in. */
	plugin_dummy();

	/* start. */
	TCPCP_DPRINTK("start.\n");

	/* lock Process List. */
	write_lock(&(proc_list.lock));
	TCPCP_DPRINTK("Process List is locked.\n");

#ifdef CONFIG_TCPCP_DEBUG
	/* @@@ print Process List without lock. (for DEBUG) @@@ */
	__print_list();
#endif

	/* search Process List. */
	list_for_each_entry_safe(p_entry, next_p_entry, &(proc_list.head),
				 node) {		
		/* @@@ for DEBUG @@@ */
		TCPCP_DPRINTK("Process List Loop.\n");

		/* search Reqest List. */
		list_for_each_entry_safe(r_entry, next_r_entry,
		    &(p_entry->r_list_head), node) {
			/* @@@ for DEBUG @@@ */
			TCPCP_DPRINTK("Request List Loop.\n");

			/* delete Request Entry with common entries. */
			__tcpcp_del_req_entry(r_entry);
		}
	}

#ifdef CONFIG_TCPCP_DEBUG
	/* @@@ print Process List without lock. (for DEBUG) @@@ */
	__print_list();
#endif

	/* release Process List. */
	write_unlock(&(proc_list.lock));
	TCPCP_DPRINTK("Process List is unlocked.\n");

	/* end. */
	TCPCP_DPRINTK("end.\n");
	return;
}

/* ----- Timeout-watcher ----- */

/**
 * tcpcp_quick_timeout
 *   - unlink Request Entry from Request List, and re-link to Timeout List.
 *
 * [Args]
 *   (IN) *sk : TCP socket
 **/
static void tcpcp_quick_timeout(struct sock *sk)
{
	struct tcpcp_proc_entry *p_entry;
	struct tcpcp_req_entry *r_entry;

	/* plug-in. */
	plugin_dummy();

	/* start. */
	TCPCP_DPRINTK("start. sk=0x%p\n", sk);

	/* lock Process List (and Timeout List). */
	write_lock(&(proc_list.lock));
	TCPCP_DPRINTK("Process List is locked.\n");

#ifdef CONFIG_TCPCP_DEBUG
	/* @@@ print Process List without lock. (for DEBUG) @@@ */
	__print_list();
#endif

	/* search Process List. */
	list_for_each_entry(p_entry, &(proc_list.head), node) {
		/* @@@ for DEBUG @@@ */
		TCPCP_DPRINTK("Process List Loop.\n");

		/* lookup Process Entry what has same Thread Group ID. */
		if (p_entry->tgid == current->tgid)
			goto check_sk;
	}

	/* not lookup ! */
	TCPCP_DPRINTK("there is no process entry to lookup.\n");
	goto out;

check_sk:
	/* search Request List. */
	TCPCP_DPRINTK("search Request List.\n");

	list_for_each_entry(r_entry, &(p_entry->r_list_head), node) {
		/* @@@ for DEBUG @@@ */
		TCPCP_DPRINTK("Request List Loop.\n");

		/* lookup Request Entry what has same socket. */
		if (r_entry->sk == sk)
			/* lookup !*/
			goto quick_timeout;
	}

	/* not lookup ! */
	goto out;

quick_timeout:
	TCPCP_DPRINTK("looked up Request Entry what" \
		"  has same socket.\n");

	/* unlink Request Entry from Timeout List. */
	list_del_init(&(r_entry->timer_node));

	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("r_entry->timeout=%u\n",
		r_entry->timeout);

	r_entry->timeout = tcp_time_stamp;
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("r_entry->timeout=%u\n",
		r_entry->timeout);

	/* unlink Request Entry from Request List. */
	list_del_init(&(r_entry->node));

#ifdef CONFIG_TCPCP_DEBUG
	/* @@@ print Process List without lock. (for DEBUG) @@@ */
	__print_list();
#endif

	/* delete Process Entry. */
	tcpcp_del_proc_entry(r_entry->p_entry);

	r_entry->p_entry = NULL;

#ifdef CONFIG_TCPCP_DEBUG
	/* @@@ print Process List without lock. (for DEBUG) @@@ */
	__print_list();
#endif

	/* link Request Entry to Timeout List. */
	list_add(&(r_entry->timer_node), &(timeout_list.head));

out:
#ifdef CONFIG_TCPCP_DEBUG
	/* @@@ print Process List without lock. (for DEBUG) @@@ */
	__print_list();
#endif

	/* release Process List (and Timeout List). */
	write_unlock(&(proc_list.lock));
	TCPCP_DPRINTK("Process List is unlocked.\n");

	/* end. */
	TCPCP_DPRINTK("end.\n");
	return;
}

/**
 * tcpcp_set_timeout - link Request Entry to Timeout List.
 *
 * [Args]
 *   (IN) *r_entry : Request Entry
 **/
static void tcpcp_set_timeout(struct tcpcp_req_entry *r_entry)
{
	/* plug-in. */
	plugin_dummy();

	/* start. */
	TCPCP_DPRINTK("start. r_entry=0x%p\n", r_entry);

	r_entry->timeout = tcp_time_stamp + TCPCP_REQ_ENTRY_TIMEOUT;
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("tcp_time_stamp  =%u\n",
		tcp_time_stamp);
	TCPCP_DPRINTK("r_entry->timeout=%u\n",
		r_entry->timeout);
	TCPCP_DPRINTK("r_emtry->timeout-tcp_time_stamp=%u\n",
		r_entry->timeout - tcp_time_stamp);

	/* lock Process List (and Timeout List). */
	write_lock(&(proc_list.lock));
	TCPCP_DPRINTK("Process List is locked.\n");

	if ((list_empty(&(timeout_list.head)) != 0)
	    && (timer_pending(&tcpcp_timer) == 0)) {
		/* regist Timeout-watcher. */
		tcpcp_regist_timer();
		/* @@@ for DEBUG @@@ */
		TCPCP_DPRINTK("Wake up Timeout-watcher.\n");
	}

	/* link Request Entry to Timeout List. */
	list_add_tail(&(r_entry->timer_node), &(timeout_list.head));

	/* release Process List (and Timeout List). */
	write_unlock(&(proc_list.lock));
	TCPCP_DPRINTK("Process List is unlocked.\n");

	/* end. */
	TCPCP_DPRINTK("end.\n");
	return;
}

/**
 * tcpcp_reset_timeout - lookup Request Entry and unlink from Timeout List.
 *
 * [Args]
 *   (IN) *sk : socket
 *
 * [Return]
 *   normal : Request Entry
 *   NULL   : not lookup
 **/
static struct tcpcp_req_entry *tcpcp_reset_timeout(struct sock *sk)
{
	struct tcpcp_req_entry *r_entry;

	/* plug-in. */
	plugin_dummy();

	/* start. */
	TCPCP_DPRINTK("start. sk=0x%p\n", sk);

	/* lock Process List (and Timeout List). */
	write_lock(&(proc_list.lock));
	TCPCP_DPRINTK("Process List is locked.\n");

	/* lookup Request Entry. */
	r_entry = tcpcp_lookup_req_entry(sk);
	if (r_entry == NULL) {
		TCPCP_DPRINTK("Not lookup request entry.\n");

		goto out;
	}

	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("lookup! r_entry=0x%p\n", r_entry);

	/* unlink Request Entry from Timeout List. */
	list_del_init(&(r_entry->timer_node));

out:	
	/* release Process List (and Timeout List). */
	write_unlock(&(proc_list.lock));
	TCPCP_DPRINTK("Process List is unlocked.\n");

	/* end. */
	TCPCP_DPRINTK("end. r_entry=0x%p\n", r_entry);
	return r_entry;
}


/* ----- make FD Result (FDR) Information ---------------------------------- */

/**
 * tcpcp_make_res - make FDR Info. (only "ALL")
 *
 * Caller must be "ALL".
 * 
 * [Args]
 *   (IN) fd                 : FD of TCP socket
 *   (IN) ret                : result
 *   (IN/OUT) **fdr_entry    : FDR Info. Entry
 *
 * [Return]
 *   0      : OK.
 *  -ENOMEM : NG.
 **/
static int tcpcp_make_res(unsigned int fd, int ret,
			  struct tcpcp_com_entry **fdr_entry)
{
	struct fdr *fdrs;
	unsigned int fdr_len;
	char *w_ptr;

	/* plug-in. */
	plugin_dummy();

	/* start. */
	TCPCP_DPRINTK("start. fd=%d, ret=%d\n", fd, ret);

	/* get write-point of FDR Info. Entry. */
	fdr_len = sizeof(struct fdr);
	w_ptr = tcpcp_get_w_ptr(TCPCP_FLAG_ALL, fdr_len, fdr_entry);
	if (w_ptr == NULL)
		return -ENOMEM;

	/* make FDR Info. */
	fdrs = (struct fdr *)w_ptr;
	fdrs->fd = fd;
	fdrs->result = ret;
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("fdrs->fd=%d\n", fdrs->fd);
	TCPCP_DPRINTK("fdrs->result=%d\n", fdrs->result);

	/* end. */
	TCPCP_DPRINTK("end.\n");
	return 0;
}


/* ----- make TCP Socket Informations (TCP-SIs) ---------------------------- */

/**
 * tcpcp_make_si_skb - make TCP-SI packet info.
 *
 * This function is based on TCPCP-9:put_buffer().
 * 
 * [Args]
 *   (IN) type     : "write_queue" / "receive_queue"
 *   (IN) *skb     : socket buffer (linerized)
 *   (IN) data_len : length of packet.
 *   (IN) *si_skb  : TCP-SIs Entry
 **/
static void tcpcp_make_si_skb(uint8_t type, struct sk_buff *skb,
			      unsigned int data_len,
			      struct tcpcp_si_skb *si_skb)
{
	/* plug-in. */
	plugin_dummy();

	/* start. */
	TCPCP_DPRINTK("start. type=%u, skb=0x%p, data_len=%u,"
		" si_skb=0x%p\n", type, skb, data_len, si_skb);

	/* set packet info. */
	si_skb->type = type;
	si_skb->length = htons(data_len);
	si_skb->flags = TCP_SKB_CB(skb)->flags;
	si_skb->sacked = TCP_SKB_CB(skb)->sacked;
	si_skb->gso_segs = htons(skb_shinfo(skb)->gso_segs);
	si_skb->gso_size = htons(skb_shinfo(skb)->gso_size);
	si_skb->seq = htonl(TCP_SKB_CB(skb)->seq);
	memcpy(si_skb->data, skb->tail - data_len, data_len);
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("si_skb->type=%u\n",
		si_skb->type);
	TCPCP_DPRINTK("si_skb->length=%u\n",
		si_skb->length);
	TCPCP_DPRINTK("si_skb->flags=%u\n",
		si_skb->flags);
	TCPCP_DPRINTK("si_skb->sacked=%u\n",
		si_skb->sacked);
	TCPCP_DPRINTK("si_skb->gso_segs=%u\n",
		si_skb->gso_segs);
	TCPCP_DPRINTK("si_skb->gso_size=%u\n",
		si_skb->gso_size);
	TCPCP_DPRINTK("si_skb->seq=%u\n",
		si_skb->seq);
	TCPCP_DPRINTK("si_skb->data=0x%p\n",
		si_skb->data);
	TCPCP_DPRINTK("TCP_SKB_CB(skb)->when=%u\n",
		      TCP_SKB_CB(skb)->when);

	/* end. */
	TCPCP_DPRINTK("end.\n");
	return;
}

/**
 * tcpcp_copy_queue - copy write_queue / receive_queue.
 *
 * This function is based on TCPCP-9:copy_queue()
 *
 * [Args]
 *   (IN) *queue         : write_queue / receive_queue of TCP socket
 *   (IN) type           : type of queue (to make packet info.)
 *   (IN) flag           : "ONE" / "ALL" (to add new TCP-SIs Entry) 
 *   (IN/OUT) **si_entry : TCP-SIs Entry
 *   (IN/OUT) *si_size   : total size of TCP-SI (include packet info.)
 *
 * [Return]
 *   0      : OK.
 *  -ENOMEM : NG. (Caller must delete all TCP-SIs Entry.)
 **/
static int tcpcp_copy_queue(struct sk_buff_head *queue, uint8_t type,
		unsigned char flag, struct tcpcp_com_entry **si_entry,
		unsigned int *si_size)
{
	struct tcpcp_si_skb *si_skb;
	struct sk_buff *skb, *clone;
	unsigned int data_len;
	unsigned int si_skb_len;
	char *w_ptr;
	int error;

#ifdef CONFIG_TCPCP_DEBUG
	/* @@@ for printing counter of packets. (for DEBUG) @@@ */
	int pkt_cnt = 0;
#endif

	/* plug-in. */
	plugin_dummy();

	/* start. */
	TCPCP_DPRINTK("start. queue=0x%p, type=%u, flag=%u\n",
		queue, type, flag);

	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("queue->qlen=%u\n", queue->qlen);

	/* search socket buffer. */
	skb_queue_walk(queue, skb) {

#ifdef CONFIG_TCPCP_DEBUG
		/* @@@ print counter of packets. (for DEBUG) @@@ */
		pkt_cnt++;
		TCPCP_DPRINTK("packet count=%d\n",
			pkt_cnt);
#endif

		/* clone socket buffer. */
		clone = skb_clone(skb, GFP_KERNEL);
		if (!clone) {
			TCPCP_DPRINTK("[Error] can't clone skb.\n");
			return -ENOMEM;
		}

		/* linearize socket buffer. */
		error = skb_linearize_cow(clone);
		if (error != 0) {
                        TCPCP_DPRINTK("[Error] can't linearize skb."
				" error=%d\n", error);
			kfree_skb(clone);
			return -ENOMEM;
		}

		/* calculate data length of packet. */
		data_len = TCP_SKB_CB(clone)->end_seq - TCP_SKB_CB(clone)->seq
			- ((TCP_SKB_CB(clone)->flags & TCPCB_FLAG_SYN) ? 1 : 0)
			- ((TCP_SKB_CB(clone)->flags & TCPCB_FLAG_FIN) ? 1 : 0);
		/* @@@ for DEBUG @@@ */
		TCPCP_DPRINTK("data_len=%u\n", data_len);
		TCPCP_DPRINTK("TCP_SKB_CB(clone)->seq    =%u\n",
			TCP_SKB_CB(clone)->seq);
		TCPCP_DPRINTK("TCP_SKB_CB(clone)->end_seq=%u\n",
			TCP_SKB_CB(clone)->end_seq);
		TCPCP_DPRINTK("TCP_SKB_CB(clone)->flags=%u\n",
			TCP_SKB_CB(clone)->flags);

		/* calculate length of packet info. */
		si_skb_len = sizeof(struct tcpcp_si_skb)
			     + ((data_len + 3) & ~3);
		/* @@@ for DEBUG @@@ */
		TCPCP_DPRINTK("si_skb_len=%u\n",
			si_skb_len);

		/* get write-point of TCP-SIs Entry. */
		w_ptr = tcpcp_get_w_ptr(flag, si_skb_len, si_entry);
		if (w_ptr == NULL) {
			TCPCP_DPRINTK("[Error] can't get w_ptr.\n");
			kfree_skb(clone);
			return -ENOMEM;
		}

		/* make packet info. */
		si_skb = (struct tcpcp_si_skb *)w_ptr;
		(*si_size) += si_skb_len;
		/* @@@ for DEBUG @@@ */
		TCPCP_DPRINTK("*si_size=%u\n", *si_size);

		tcpcp_make_si_skb(type, clone, data_len, si_skb);
		/* @@@ for DEBUG @@@ */
		TCPCP_DPRINTK("clone->truesize=%u\n",
			clone->truesize);

		kfree_skb(clone);
	}

	/* end. */
	TCPCP_DPRINTK("end.\n");
	return 0;
}

/**
 * tcpcp_make_si_hdr - make TCP-SI header.
 *
 * This function is based on TCPCP-9:do_getici().
 *
 * [Args]
 *   (IN) *sk     : TCP socket
 *   (IN) si_size : total length of TCP-SI (include packet info.)
 *   (IN) fd      : FD of TCP socket
 *   (IN) *si_hdr : TCP-SIs Entry
 **/
static void tcpcp_make_si_hdr(struct sock *sk, unsigned int si_size,
			      unsigned int fd, struct tcpcp_si_hdr *si_hdr)
{
	struct inet_sock *inet = inet_sk(sk);
#if defined(CONFIG_IPV6) || defined(CONFIG_IPV6_MODULE)
	struct ipv6_pinfo *inet6 = inet6_sk(sk);
#endif
	struct tcp_sock *tp = tcp_sk(sk);

	/* plug-in. */
	plugin_dummy();

	/* start. */
	TCPCP_DPRINTK("start. sk=0x%p, si_size=%u, fd=%u, si_hdr=0x%p\n",
		 sk, si_size, fd, si_hdr);

	/* get total length of TCP-SI and FD. */
	si_hdr->si_length = htonl(si_size / 4);
	si_hdr->fd = htonl(fd);
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("si_size          =%u\n",
		si_size);
	TCPCP_DPRINTK("si_hdr->si_length=%u\n",
		si_hdr->si_length);
	TCPCP_DPRINTK("fd        =%u\n", fd);
	TCPCP_DPRINTK("si_hdr->fd=%u\n", si_hdr->fd);

	/* get globally unique TCP connection ID. */
#if defined(CONFIG_IPV6) || defined(CONFIG_IPV6_MODULE)
	if (sk->sk_family == PF_INET6) {
		int addr_type = ipv6_addr_type(&inet6->daddr);
		/* @@@ for DEBUG @@@ */
		TCPCP_DPRINTK("inet6->daddr=%04x:%04x:%04x:%04x:"
			"%04x:%04x:%04x:%04x\n",
			ntohs(inet6->daddr.s6_addr16[0]),
			ntohs(inet6->daddr.s6_addr16[1]),
			ntohs(inet6->daddr.s6_addr16[2]),
			ntohs(inet6->daddr.s6_addr16[3]),
			ntohs(inet6->daddr.s6_addr16[4]),
			ntohs(inet6->daddr.s6_addr16[5]),
			ntohs(inet6->daddr.s6_addr16[6]),
			ntohs(inet6->daddr.s6_addr16[7]));
		TCPCP_DPRINTK("addr_type=%u\n", addr_type);

		if (addr_type == IPV6_ADDR_MAPPED)
			si_hdr->id.ip_version = 4;
		else
			si_hdr->id.ip_version = 6;
	} else
#endif
		si_hdr->id.ip_version = 4;
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("si_hdr->id.ip_version=%u\n",
		si_hdr->id.ip_version);

#if defined(CONFIG_IPV6) || defined(CONFIG_IPV6_MODULE)
	if (si_hdr->id.ip_version == 6) {
		ipv6_addr_copy(&(si_hdr->id.ip.v6.ip_src), &(inet6->saddr));
		ipv6_addr_copy(&(si_hdr->id.ip.v6.ip_dst), &(inet6->daddr));
		si_hdr->id.ip.v6.scope_id = htonl(sk->sk_bound_dev_if);
		/* @@@ for DEBUG @@@ */
		TCPCP_DPRINTK("si_hdr->id.ip.v6.ip_src=%04x:%04x:%04x:"
			"%04x:%04x:%04x:%04x:%04x\n",
			ntohs(si_hdr->id.ip.v6.ip_src.s6_addr16[0]),
			ntohs(si_hdr->id.ip.v6.ip_src.s6_addr16[1]),
			ntohs(si_hdr->id.ip.v6.ip_src.s6_addr16[2]),
			ntohs(si_hdr->id.ip.v6.ip_src.s6_addr16[3]),
			ntohs(si_hdr->id.ip.v6.ip_src.s6_addr16[4]),
			ntohs(si_hdr->id.ip.v6.ip_src.s6_addr16[5]),
			ntohs(si_hdr->id.ip.v6.ip_src.s6_addr16[6]),
			ntohs(si_hdr->id.ip.v6.ip_src.s6_addr16[7]));
		TCPCP_DPRINTK("si_hdr->id.ip.v6.ip_dst=%04x:%04x:%04x:"
			"%04x:%04x:%04x:%04x:%04x\n",
			ntohs(si_hdr->id.ip.v6.ip_dst.s6_addr16[0]),
			ntohs(si_hdr->id.ip.v6.ip_dst.s6_addr16[1]),
			ntohs(si_hdr->id.ip.v6.ip_dst.s6_addr16[2]),
			ntohs(si_hdr->id.ip.v6.ip_dst.s6_addr16[3]),
			ntohs(si_hdr->id.ip.v6.ip_dst.s6_addr16[4]),
			ntohs(si_hdr->id.ip.v6.ip_dst.s6_addr16[5]),
			ntohs(si_hdr->id.ip.v6.ip_dst.s6_addr16[6]),
			ntohs(si_hdr->id.ip.v6.ip_dst.s6_addr16[7]));
		TCPCP_DPRINTK("sk->sk_bound_dev_if      =%u\n",
			sk->sk_bound_dev_if);
		TCPCP_DPRINTK("si_hdr->id.ip.v6.scope_id=%u\n",
			si_hdr->id.ip.v6.scope_id);
	} else
#endif
	{
		si_hdr->id.ip.v4.ip_src.s_addr = inet->saddr;
		si_hdr->id.ip.v4.ip_dst.s_addr = inet->daddr;
		/* @@@ for DEBUG @@@ */
		TCPCP_DPRINTK("si_hdr->id.ip.v4.ip_src=%u.%u.%u.%u\n",
			
			((uint8_t*)&(si_hdr->id.ip.v4.ip_src))[0],
			((uint8_t*)&(si_hdr->id.ip.v4.ip_src))[1],
			((uint8_t*)&(si_hdr->id.ip.v4.ip_src))[2],
			((uint8_t*)&(si_hdr->id.ip.v4.ip_src))[3]);
		TCPCP_DPRINTK("si_hdr->id.ip.v4.ip_dst=%u.%u.%u.%u\n",
			
			((uint8_t*)&(si_hdr->id.ip.v4.ip_dst))[0],
			((uint8_t*)&(si_hdr->id.ip.v4.ip_dst))[1],
			((uint8_t*)&(si_hdr->id.ip.v4.ip_dst))[2],
			((uint8_t*)&(si_hdr->id.ip.v4.ip_dst))[3]);
	}
	si_hdr->id.tcp_sport = inet->sport;
	si_hdr->id.tcp_dport = inet->dport;
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("si_hdr->id.tcp_sport=%u\n",
		      (unsigned int)si_hdr->id.tcp_sport);
	TCPCP_DPRINTK("si_hdr->id.tcp_dport=%u\n",
		      (unsigned int)si_hdr->id.tcp_dport);

	/* get fixed general data. */
	if (tp->rx_opt.tstamp_ok)
		si_hdr->fixgen.tcp_flags |= TCPI_OPT_TIMESTAMPS;
	if (tp->rx_opt.sack_ok)
		si_hdr->fixgen.tcp_flags |= TCPI_OPT_SACK;
	if (tp->rx_opt.wscale_ok) {
		si_hdr->fixgen.tcp_flags |= TCPI_OPT_WSCALE;
		si_hdr->fixgen.snd_wscale = tp->rx_opt.snd_wscale;
		si_hdr->fixgen.rcv_wscale = tp->rx_opt.rcv_wscale;
	} else {
		si_hdr->fixgen.snd_wscale = 0;
		si_hdr->fixgen.rcv_wscale = 0;
	}
	if (tp->ecn_flags & TCP_ECN_OK)
		si_hdr->fixgen.tcp_flags |= TCPI_OPT_ECN;
	si_hdr->fixgen.snd_mss = htons(tp->advmss);
	si_hdr->fixgen.rcv_mss = htons(tp->rx_opt.mss_clamp);
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("tp->rx_opt.tstamp_ok=%u\n",
		tp->rx_opt.tstamp_ok);
	TCPCP_DPRINTK("tp->rx_opt.sack_ok  =%u\n",
		tp->rx_opt.sack_ok);
	TCPCP_DPRINTK("tp->rx_opt.wscale_ok=%u\n",
		tp->rx_opt.wscale_ok);
	TCPCP_DPRINTK("tp->ecn_flags=%u\n",
		tp->ecn_flags);
	TCPCP_DPRINTK("si_hdr->fixgen.tcp_flags =%u\n",
		si_hdr->fixgen.tcp_flags);
	TCPCP_DPRINTK("si_hdr->fixgen.snd_wscale=%u\n",
		si_hdr->fixgen.snd_wscale);
	TCPCP_DPRINTK("si_hdr->fixgen.rcv_wscale=%u\n",
		si_hdr->fixgen.rcv_wscale);
	TCPCP_DPRINTK("tp->advmss            =%u\n",
		tp->advmss);
	TCPCP_DPRINTK("si_hdr->fixgen.snd_mss=%u\n",
		si_hdr->fixgen.snd_mss);
	TCPCP_DPRINTK("tp->rx_opt.mss_clamp         =%u\n",
		tp->rx_opt.mss_clamp);
	TCPCP_DPRINTK("si_hdr->fixgen.rcv_mss=%u\n",
		si_hdr->fixgen.rcv_mss);
	TCPCP_DPRINTK("inet_csk(sk)->icsk_ack.rcv_mss =%u\n",
		      inet_csk(sk)->icsk_ack.rcv_mss);

	/* get variable general data. */
	si_hdr->vargen.state = sk->sk_state;
	si_hdr->vargen.snd_nxt = htonl(tp->snd_nxt);
	si_hdr->vargen.rcv_nxt = htonl(tp->rcv_nxt);
	si_hdr->vargen.snd_wnd = htonl(tp->snd_wnd);
	si_hdr->vargen.rcv_wnd = htonl(tcp_receive_window(tp)); /* tp->rcv_wnd); */
	si_hdr->vargen.ts_gen = htonl(tp_time_stamp(tp));
	si_hdr->vargen.ts_recent = htonl(tp->rx_opt.saw_tstamp ? tp->rx_opt.ts_recent : 0);
	si_hdr->vargen.copied_seq = htonl(tp->copied_seq);
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("si_hdr->vargen.state=%u\n",
		      si_hdr->vargen.state);
	TCPCP_DPRINTK("si_hdr->vargen.snd_nxt=%u\n",
		      si_hdr->vargen.snd_nxt);
	TCPCP_DPRINTK("si_hdr->vargen.rcv_nxt=%u\n",
		      si_hdr->vargen.rcv_nxt);
	TCPCP_DPRINTK("tp->snd_wnd           =%lu\n",
		      (unsigned long) tp->snd_wnd);
	TCPCP_DPRINTK("si_hdr->vargen.snd_wnd=%u\n",
		      si_hdr->vargen.snd_wnd);
	TCPCP_DPRINTK("tcp_receive_window(tp)=%lu\n",
		      (unsigned long) tcp_receive_window(tp));
	TCPCP_DPRINTK("si_hdr->vargen.rcv_wnd=%u\n",
		      si_hdr->vargen.rcv_wnd);
	TCPCP_DPRINTK("si_hdr->vargen.ts_gen   =%u\n",
		      ntohl(si_hdr->vargen.ts_gen));
	TCPCP_DPRINTK("si_hdr->vargen.ts_recent=%u\n",
		      ntohl(si_hdr->vargen.ts_recent));
	TCPCP_DPRINTK("tcp_time_stamp         =%u\n",
		      tcp_time_stamp);
	TCPCP_DPRINTK("si_hdr->vargen.copied_seq=%u\n",
		      ntohl(si_hdr->vargen.copied_seq));

	/* get socket options. (socket) */
	si_hdr->sockopt.sk.flags = 0;

	if (sk->sk_reuse)
		si_hdr->sockopt.sk.flags |= TCPCP_SIE_SOCK_REUSEADDR;
	if (sock_flag(sk, SOCK_KEEPOPEN))
		si_hdr->sockopt.sk.flags |= TCPCP_SIE_SOCK_KEEPOPEN;
	if (sock_flag(sk, SOCK_LINGER)) {
		si_hdr->sockopt.sk.flags |= TCPCP_SIE_SOCK_LINGER;
		si_hdr->sockopt.sk.l_linger = htonl(sk->sk_lingertime / HZ);
	}
	si_hdr->sockopt.sk.userlocks = sk->sk_userlocks;
	si_hdr->sockopt.sk.sndbuf = htonl(sk->sk_sndbuf / 2);
	si_hdr->sockopt.sk.rcvbuf = htonl(sk->sk_rcvbuf / 2);
	si_hdr->sockopt.sk.priority = htonl(sk->sk_priority);
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("sk->sk_reuse=%u\n", sk->sk_reuse);
	TCPCP_DPRINTK("sk->sk_flags SOCK_KEEPALIVE=%u\n",
		!!sock_flag(sk, SOCK_KEEPOPEN));
	TCPCP_DPRINTK("sk->sk_flags SOCK_LINGER   =%u\n",
		!!sock_flag(sk, SOCK_LINGER));
	TCPCP_DPRINTK("sk->sk_priority=%lu\n",
		(unsigned long) sk->sk_priority);
	TCPCP_DPRINTK("sk->sk_lingertime/HZ=%lu\n",
                (unsigned long) sk->sk_lingertime / HZ);
        TCPCP_DPRINTK("sk->sk_userlocks=%u\n",
		sk->sk_userlocks);
        TCPCP_DPRINTK("sk->sk_sndbuf/2=%lu\n",
		(unsigned long) sk->sk_sndbuf / 2);
	TCPCP_DPRINTK("sk->sk_rcvbuf/2=%lu\n",
		(unsigned long) sk->sk_rcvbuf / 2);
	TCPCP_DPRINTK("sk->sk_priority=%lu\n",
		(unsigned long) sk->sk_priority);

#if defined(CONFIG_IPV6) || defined(CONFIG_IPV6_MODULE)
	if (si_hdr->id.ip_version == 6) {
		/* get socket options. (IPv6) */
		si_hdr->sockopt.ip.v6.flags = 0;
		if (inet6->hop_limit == -1)
			si_hdr->sockopt.ip.v6.flags
				|= TCPCP_SIE_IPV6_DEFAULT_HOPLIMIT;
		else
			si_hdr->sockopt.ip.v6.hop_limit = inet6->hop_limit;
		si_hdr->sockopt.ip.v6.pmtudisc = inet6->pmtudisc;
		/* @@@ for DEBUG @@@ */
		TCPCP_DPRINTK("inet6->ttl=%d\n",
			inet6->hop_limit);
                TCPCP_DPRINTK("inet6->pmtudisc=%u\n",
			inet6->pmtudisc);
	} else
#endif
	{
		unsigned char optbuf[sizeof(struct ip_options) + 40];
		struct ip_options *opt = (struct ip_options *)optbuf;

		/* get socket options. (IPv4) */
		si_hdr->sockopt.ip.v4.flags = 0;
		if (inet->uc_ttl == -1)
			si_hdr->sockopt.ip.v4.flags |= TCPCP_SIE_IP_DEFAULT_TTL;
		else
			si_hdr->sockopt.ip.v4.uc_ttl = inet->uc_ttl;
		si_hdr->sockopt.ip.v4.pmtudisc = inet->pmtudisc;
		si_hdr->sockopt.ip.v4.tos = inet->tos;
		/* @@@ for DEBUG @@@ */
		TCPCP_DPRINTK("inet->ttl %d\n",
			inet->uc_ttl);
		TCPCP_DPRINTK("inet->pmtudisc %u\n",
			inet->pmtudisc);
		TCPCP_DPRINTK("inet->tos %d\n",
			inet->tos);

		/* get socket options. (IPv4: IPv4 options) */
		opt->optlen = 0;
		if (inet->opt) {
			memcpy(optbuf, inet->opt,
			       sizeof(struct ip_options) + inet->opt->optlen);
			/* @@@ for DEBUG @@@ */
			TCPCP_DPRINTK("inet->opt->optlen=%u\n",
				inet->opt->optlen);
		}
		/* @@@ for DEBUG @@@ */
		TCPCP_DPRINTK("opt->optlen=%u\n",
			opt->optlen);

		si_hdr->sockopt.ip.v4.opt_len = opt->optlen;
		if (opt->optlen) {
			ip_options_undo(opt);
			memcpy(&si_hdr->sockopt.ip.v4.opt_data, opt->__data,
				opt->optlen);
		}
	}

	/* get socket options. (TCP) */
	si_hdr->sockopt.tcp.nonagle = tp->nonagle;
	si_hdr->sockopt.tcp.keepcnt = tp->keepalive_probes;
	si_hdr->sockopt.tcp.keepidle = htons(tp->keepalive_time / HZ);
	si_hdr->sockopt.tcp.keepintvl = htons(tp->keepalive_intvl / HZ);
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("tp->nonagle=%u\n", tp->nonagle);
	TCPCP_DPRINTK("tp->keepalive_probes=%u\n",
		tp->keepalive_probes);
	TCPCP_DPRINTK("tp->keepalive_time/HZ=%d\n",
		tp->keepalive_time / HZ);
	TCPCP_DPRINTK("tp->keepalive_intvl/HZ=%d\n",
		tp->keepalive_intvl / HZ);

	/* end. */
	TCPCP_DPRINTK("end.\n");
	return;
}

/**
 * do_make_si - make TCP-SI.
 *
 * This function is based on TCPCP-9:do_getici().
 * Caller must lock socket.
 * 
 * [Args]
 *   (IN) *sk            : TCP socket
 *   (IN) fd             : FD of TCP socket (to make TCP-SI header)
 *   (IN) flag           : "ONE" / "ALL" (to add new TCP-SIs Entry)
 *   (IN/OUT) **si_entry : TCP-SIs Entry
 *   (OUT) *si_size      : total length of TCP-SI (include pack.)
 *
 * [Return]
 *   0      : OK.
 *  -EBADFD : NG.
 *  -ENOMEM : NG.
 **/
static int do_make_si(struct sock *sk, unsigned int fd, unsigned char flag,
		      struct tcpcp_com_entry **si_entry,
		      unsigned int *si_size)
{
	struct tcpcp_si_hdr *si_hdr;
	unsigned int si_hdr_len; 
	char *w_ptr;
	int error;

	/* plug-in. */
	plugin_dummy();

	/* start. */
	TCPCP_DPRINTK("start. sk=0x%p, fd=%u, flag=%u\n",
		sk, fd, flag);

	/* check socket error. */
	error = sock_error(sk);
	if (error != 0) {
		TCPCP_DPRINTK("[Error] sock_error failure. error=%d\n",
			error);
		return -EBADFD;
	}
	/* @@@ for DEBUG @@@ */
        TCPCP_DPRINTK("check socket error.\n");

	/* get write-point of TCP-SIs Entry. */
	si_hdr_len = sizeof(struct tcpcp_si_hdr);
	w_ptr = tcpcp_get_w_ptr(flag, si_hdr_len, si_entry);
	if (w_ptr == NULL) {
		TCPCP_DPRINTK("[Error] can't get w_ptr.\n");
		return -ENOMEM;
	}

	/* reserve TCP-SI header. */
	si_hdr = (struct tcpcp_si_hdr *)w_ptr;
	(*si_size) = si_hdr_len;
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("*si_size=%u\n", *si_size);

	/* copy write_queue. */
	error = tcpcp_copy_queue(&sk->sk_write_queue, TCPCP_SIE_BUF_SND, flag,
				 si_entry, si_size);
	if (error != 0) {
		TCPCP_DPRINTK("[Error] can't copy write_queue."
			" error=%d\n", error);
		return -ENOMEM;
	}
	/* @@@ for DEBUG @@@ */
        TCPCP_DPRINTK("copy write_queue.\n");
	TCPCP_DPRINTK("sk->sk_sndbuf     =%u\n",
		sk->sk_sndbuf);
	TCPCP_DPRINTK("sk->sk_wmem_queued=%u\n",
		sk->sk_wmem_queued);

	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("*si_size=%u\n", *si_size);

	/* copy receive_queue. */
	error = tcpcp_copy_queue(&sk->sk_receive_queue, TCPCP_SIE_BUF_RCV, flag,
				 si_entry, si_size);
	if (error != 0) {
		TCPCP_DPRINTK("[Error] can't copy receive_queue."
			" error=%d\n", error);
		return -ENOMEM;
	}
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("copy receive_queue.\n");

	/* make TCP-SI header. */
	tcpcp_make_si_hdr(sk, (*si_size), fd, si_hdr);	

	/* end. */
	TCPCP_DPRINTK("end.\n");
	return 0;
}

/**
 * do_tcpcp_make_si - make TCP-SIs.
 *
 * [Args]
 *   (IN) *sk       : TCP socket
 *   (IN) *r_entry  : Request Entry
 *   (IN) *user_res : Processing Result Info.
 *
 * [Return]
 *   0      : OK.
 *  -ENOMEM : NG.
 **/
static int do_tcpcp_make_si(struct sock *sk,
			    struct tcpcp_req_entry *r_entry,
			    struct result *user_res)
{
	struct task_struct *tsk = current;
	struct files_struct *files = tsk->files;
	struct tcpcp_com_entry *si_entry = NULL;
	struct tcpcp_com_entry *fdr_entry = NULL;
	struct tcpcp_total_si_hdr *total_si_hdr;
	unsigned int total_si_hdr_len;
	unsigned int total_si_size;
	unsigned int fd;
	char *w_ptr;
	unsigned int fd_all = 0;
	unsigned int fd_ng = 0;
	int error;
	struct fdtable *fdt;
	/* plug-in. */
	plugin_dummy();

	/* start. */
	TCPCP_DPRINTK("start. sk=0x%p, r_entry=0x%p, user_res=0x%p\n",
		sk, r_entry, user_res);

	/* get write-point of TCP-SIs Entry. */
	total_si_hdr_len = sizeof(struct tcpcp_total_si_hdr);
	w_ptr = tcpcp_get_w_ptr(r_entry->flag, total_si_hdr_len,
				&(r_entry->s_entry));
	if (w_ptr == NULL) {
		TCPCP_DPRINTK("[Error] can't get w_ptr.\n");
		return -ENOMEM;
	}

	if (r_entry->flag & TCPCP_FLAG_ALL) {
		/* add FDR Info. Entry. (only "ALL")*/
		error = tcpcp_add_com_entry(r_entry->flag, &(r_entry->f_entry));
		if (error != 0)
			return -ENOMEM;
	}
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("r_entry->f_entry=0x%p\n",
		r_entry->f_entry);

	/* initalize now TCP-SIs and FDR Info. Entry. */
	si_entry = r_entry->s_entry;
	fdr_entry = r_entry->f_entry;

	/* reserve TCP-SIs total header. */
	total_si_hdr = (struct tcpcp_total_si_hdr *)w_ptr;
	total_si_size = total_si_hdr_len;
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("total_si_size=%d\n", total_si_size);
	fdt = files_fdtable(files);
	/* search FD. */
	for (fd = 0; fd < fdt->max_fds; fd++) {
		unsigned int si_size = 0;
 		struct sock *sk_i;
 		int ret;
 
		/* @@@ for DEBUG @@@ */
		TCPCP_DPRINTK("fd=%d\n", fd);

		/* check FD and get SOCKET_I(). */
		sk_i = tcpcp_get_sk_i(sk, fd, r_entry);
		if (sk_i == NULL)
			continue;
		/* @@@ for DEBUG @@@ */
		TCPCP_DPRINTK("nothing continue after tcpcp_get_sk_i.\n");

		/* make TCP-SI. */
		lock_sock(sk_i);
		ret = do_make_si(sk_i, fd, r_entry->flag, &si_entry, &si_size);
		release_sock(sk_i);
		if (ret == -ENOMEM) {
			TCPCP_DPRINTK("[Error] can't make TCP-SI."
				" ret=%d\n", ret);
			return ret;
		}
		/* @@@ for DEBUG @@@ */
		TCPCP_DPRINTK("si_size=%d\n", si_size);

		total_si_size += si_size;
		/* @@@ for DEBUG @@@ */
		TCPCP_DPRINTK("total_si_size=%d\n",
			total_si_size);

		fd_all++;
		/* @@@ for DEBUG @@@ */
		TCPCP_DPRINTK("fd_all=%u\n", fd_all);

		if (!(r_entry->flag & TCPCP_FLAG_ALL)) {
			/* @@@ for DEBUG @@@ */
			TCPCP_DPRINTK("Break for loop.\n");

			if (ret != 0)
				return ret;
			break;
		}

		/* Only ALL, the following processings are performed. */
		if (ret != 0)
			fd_ng++;
		/* @@@ for DEBUG @@@ */
		TCPCP_DPRINTK("fd_ng=%u\n", fd_ng);

		/* make FDR Info. (only "ALL") */
		error = tcpcp_make_res(fd, ret, &fdr_entry);
		if (error != 0) {
			/* only -ENOMEM. */
			TCPCP_DPRINTK("[Error] can't make FDR Info."
				" error=%d\n", error);
			/* Caller delete Request Entry with common entries. */
			return error;
		}
	}

	/* make TCP-SIs header. */
	total_si_hdr->total_si_length = htonl(total_si_size / 4);
	total_si_hdr->major = TCPCP_CURRENT_SI_MAJOR;
	total_si_hdr->minor = TCPCP_CURRENT_SI_MINOR;
	total_si_hdr->si_cnt = htons(fd_all);
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("total_si_hdr->total_si_length=%u\n",
		total_si_hdr->total_si_length);
	TCPCP_DPRINTK("total_si_hdr->major=%u\n",
		total_si_hdr->major);
	TCPCP_DPRINTK("total_si_hdr->minor=%u\n",
		total_si_hdr->minor);
	TCPCP_DPRINTK("total_si_hdr->si_cnt=%u\n",
		total_si_hdr->si_cnt);

	/* set Processing Result Info. */
	user_res->fd_all = fd_all;
	user_res->fd_ng = fd_ng;
	user_res->buf_size = total_si_size;
	user_res->fdrs_size = sizeof(struct fdr) * fd_all;
	memcpy(&(r_entry->res), user_res, sizeof(struct result));
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("user_res->fd_all=%u\n",
		user_res->fd_all);
	TCPCP_DPRINTK("user_res->fd_ng =%u\n",
		user_res->fd_ng);
	TCPCP_DPRINTK("user_res->buf_size =%u\n",
		user_res->buf_size);
	TCPCP_DPRINTK("user_res->fdrs_size=%u\n",
		user_res->fdrs_size);
	TCPCP_DPRINTK("r_entry->res.fd_all=%u\n",
		r_entry->res.fd_all);
	TCPCP_DPRINTK("r_entry->res.fd_ng =%u\n",
		r_entry->res.fd_ng);
	TCPCP_DPRINTK("r_entry->res.buf_size =%u\n",
		r_entry->res.buf_size);
	TCPCP_DPRINTK("r_entry->res.fdrs_size=%u\n",
		r_entry->res.fdrs_size);

	/* end. */
	TCPCP_DPRINTK("end.\n");
	return 0;
}


/* ----- stop TCP sockets -------------------------------------------------- */

/**
 * tcpcp_change_xmit - cue packet transmission function change.
 *
 * [Args]
 *   (IN) *sk  : TCP socket
 *   (IN) flag : "TCPCP_FLAG_STOP" / "TCPCP_FLAG_START"
 **/
static void tcpcp_change_xmit(struct sock *sk, unsigned char flag)
{
  struct tcp_sock *tp = tcp_sk(sk);
  tcpcp_config_t *config = tcpcp_get_config();

  /* plug-in. */
  plugin_dummy();

  /* start. */
  TCPCP_DPRINTK("start. sk=0x%p, flag=%d\n",
		sk, flag);

  /* @@@ for DEBUG @@@ */
#if defined(CONFIG_IPV6) || defined(CONFIG_IPV6_MODULE)
  TCPCP_DPRINTK("&tcpcp_ipv6_mapped=0x%p\n",
		&config->tcpcp_ipv6_mapped);
  TCPCP_DPRINTK("&ipv6_mapped         =0x%p\n",
		config->orig_ipv6_mapped_ptr);
  TCPCP_DPRINTK("&tcpcp_ipv6_specific=0x%p\n",
		&config->tcpcp_ipv6_specific);
  TCPCP_DPRINTK("&ipv6_specific         =0x%p\n",
		config->orig_ipv6_specific_ptr);
#endif
  TCPCP_DPRINTK("&tcpcp_ipv4_specific=0x%p\n",
		&config->tcpcp_ipv4_specific);
  TCPCP_DPRINTK("&ipv4_specific         =0x%p\n",
		config->orig_ipv4_specific_ptr);

  if (flag == TCPCP_FLAG_STOP) {
    /* stop. */
#if defined(CONFIG_IPV6) || defined(CONFIG_IPV6_MODULE)
    if (sk->sk_family == PF_INET6) {
      struct ipv6_pinfo *inet6 = inet6_sk(sk);
      int addr_type = ipv6_addr_type(&inet6->daddr);
      if (addr_type == IPV6_ADDR_MAPPED) {
	tp->inet_conn.icsk_af_ops = &config->tcpcp_ipv6_mapped;
#ifdef CONFIG_TCP_MD5SIG
        tp->af_specific = (struct tcp_sock_af_ops *)&config->tcpcp_sock_ipv6_mapped_specific;
#endif /* CONFIG_TCP_MD5SIG */
      } else {
	tp->inet_conn.icsk_af_ops = &config->tcpcp_ipv6_specific;
#ifdef CONFIG_TCP_MD5SIG
        tp->af_specific = (struct tcp_sock_af_ops *)&config->tcpcp_sock_ipv6_specific;
#endif /* CONFIG_TCP_MD5SIG */
      }
    } else 
#endif
      {
	tp->inet_conn.icsk_af_ops = &config->tcpcp_ipv4_specific;
#ifdef CONFIG_TCP_MD5SIG
	tp->af_specific = (struct tcp_sock_af_ops *)&config->tcpcp_sock_ipv4_specific;
#endif /* CONFIG_TCP_MD5SIG */
      }
  } else {
    /* start. */
#if defined(CONFIG_IPV6) || defined(CONFIG_IPV6_MODULE)
    if (sk->sk_family == PF_INET6) {
      struct ipv6_pinfo *inet6 = inet6_sk(sk);
      int addr_type = ipv6_addr_type(&inet6->daddr);
      if (addr_type == IPV6_ADDR_MAPPED) {
	tp->inet_conn.icsk_af_ops = config->orig_ipv6_mapped_ptr;
#ifdef CONFIG_TCP_MD5SIG
        tp->af_specific = (struct tcp_sock_af_ops *)config->orig_sock_ipv6_mapped_specific_ptr;
#endif /* CONFIG_TCP_MD5SIG */
      } else {
	tp->inet_conn.icsk_af_ops = config->orig_ipv6_specific_ptr;
#ifdef CONFIG_TCP_MD5SIG
        tp->af_specific = (struct tcp_sock_af_ops *)config->orig_sock_ipv6_specific_ptr;
#endif /* CONFIG_TCP_MD5SIG */
      }
    } else
#endif
      {
	tp->inet_conn.icsk_af_ops = config->orig_ipv4_specific_ptr;
#ifdef CONFIG_TCP_MD5SIG
	tp->af_specific = (struct tcp_sock_af_ops *)config->orig_sock_ipv4_specific_ptr;
#endif /* CONFIG_TCP_MD5SIG */
      }
  }

  /* @@@ for DEBUG @@@ */
  TCPCP_DPRINTK("tp->af_specific=0x%p\n",
		tp->af_specific);
  TCPCP_DPRINTK("tp->inet_conn.icsk_ca_ops=0x%p\n",
		tp->inet_conn.icsk_ca_ops);

  /* end. */
  TCPCP_DPRINTK("end.\n");
  return;
}

/**
 * do_stop - stop TCP socket.
 *
 * Caller must lock socket.
 *
 * [Args]
 *   (IN) *sk : TCP socket
 **/
static void do_stop(struct sock *sk)
{
	/* plug-in. */
	plugin_dummy();

	/* start. */
	TCPCP_DPRINTK("start. sk=0x%p\n", sk);

	/* set up dummy socket filter that just drops all inbound packets. */
	if (sk->sk_filter != NULL) {
		sk_filter_release(sk->sk_filter);
		/* @@@ for DEBUG @@@ */
		TCPCP_DPRINTK("socket filter is released.\n");
		TCPCP_DPRINTK("drop_all.refcnt=%d\n",
			atomic_read(&(drop_all.refcnt)));
	}
	sk->sk_filter = &drop_all;
	sk_filter_charge(sk, sk->sk_filter);
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("&drop_all    =0x%p\n", &drop_all);
	TCPCP_DPRINTK("sk->sk_filter=0x%p\n",
		sk->sk_filter);
	TCPCP_DPRINTK("drop_all.refcnt=%d\n",
		atomic_read(&(drop_all.refcnt)));

	/* prevent socket from sending by inserting a fake set of
	 * AF-specific functions. Very ugly, of course. */
	tcpcp_change_xmit(sk, TCPCP_FLAG_STOP);

	/* since our socket is now deaf and mute,
	 * there's no point in lingering */
	sock_set_flag(sk, SOCK_LINGER);
	sk->sk_lingertime = 0;
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("sk->sk_flags=%lu (if value is 16, OK.)\n",
		sk->sk_flags);
	TCPCP_DPRINTK("sk->sk_lingertime=%lu\n",
		sk->sk_lingertime);

	/* end. */
	TCPCP_DPRINTK("end.\n");
	return;
}

/**
 * do_tcpcp_stop - stop TCP sockets.
 *
 * [Args]
 *   (IN) *sk       : TCP socket
 *   (IN) *r_entry  : Request Entry
 *   (IN) *user_res : Processing Result Info.
 *
 * [Return]
 *   0      : OK.
 *  -ENOMEM : NG.
 **/
static int do_tcpcp_stop(struct sock *sk, struct tcpcp_req_entry *r_entry,
			 struct result *user_res)
{
	struct task_struct *tsk = current;
	struct files_struct *files = tsk->files;
	struct tcpcp_com_entry *fdr_entry;
	unsigned int fd;
	unsigned int fd_all = 0;
	int error;
	struct fdtable *fdt;
	/* plug-in. */
	plugin_dummy();

	/* start. */
	TCPCP_DPRINTK("start. sk=0x%p, r_entry=0x%p, user_res=0x%p\n",
		sk, r_entry, user_res);

	if (r_entry->flag & TCPCP_FLAG_ALL) {
		/* add FDR Info. Entry. (only "ALL") */
		error = tcpcp_add_com_entry(r_entry->flag, &(r_entry->f_entry));
		if (error != 0)
			return -ENOMEM;
	}
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("r_entry->f_entry=0x%p\n",
		r_entry->f_entry);

	/* initialize now FDR Info. Entry. */
	fdr_entry = r_entry->f_entry;
	fdt = files_fdtable(files);
	/* search FD. */
	for (fd = 0; fd < fdt->max_fds; fd++) {
		struct sock *sk_i;

		/* check FD and get SOCKET_I(). */
		sk_i = tcpcp_get_sk_i(sk, fd, r_entry);
		if (sk_i == NULL) {
			/* @@@ for DEBUG @@@ */
			TCPCP_DPRINTK("sk_i is NULL.\n");

			continue;
		}
		/* stop TCP socket. */
		lock_sock(sk_i);
		do_stop(sk_i);
		release_sock(sk_i);

		fd_all++;

		if (!(r_entry->flag & TCPCP_FLAG_ALL)) {
			/* @@@ for DEBUG @@@ */
			TCPCP_DPRINTK("Break for loop.\n");

			break;
		}

		/* make FDR Info. (only "ALL") */
		error = tcpcp_make_res(fd, 0, &fdr_entry);
		if (error != 0) {
			/* only -ENOMEM. */
			TCPCP_DPRINTK("[Error] can't make FDR Info."
				" error=%d\n", error);
			/* Caller delete Request Entry with common entries. */
			return error;
		}
	}
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("Loop count=%d\n", fd);

	/* set Processing Result Info. */
	user_res->fd_all = fd_all;
	user_res->fd_ng = 0;
	user_res->buf_size = 0;
	user_res->fdrs_size = sizeof(struct fdr) * fd_all;
	memcpy(&(r_entry->res), user_res, sizeof(struct result));
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("r_entry->res->fd_all=%u\n",
		r_entry->res.fd_all);
	TCPCP_DPRINTK("r_entry->res->fd_ng =%u\n",
		r_entry->res.fd_ng);
	TCPCP_DPRINTK("r_entry->res->buf_size =%u\n",
		r_entry->res.buf_size);
	TCPCP_DPRINTK("r_entry->res->fdrs_size=%u\n",
		r_entry->res.fdrs_size);

	/* end. */
	TCPCP_DPRINTK("end.\n");
	return 0;
}


/* ----- close TCP sockets ------------------------------------------------- */

/**
 * do_close - close TCP socket.
 *
 * [Args]
 *   (IN) fd : FD of TCP socket
 *
 * [Return]
 *   0     : OK.
 *   other : NG.
 **/
static int do_close(unsigned int fd)
{
	struct files_struct *files = current->files;
	struct file *filp;
	int error;
	struct fdtable *fdt;
	/* plug-in. */
	plugin_dummy();

	/* start. */
	TCPCP_DPRINTK("start. fd=%u\n", fd);

	spin_lock(&files->file_lock);
	fdt = files_fdtable(files);
	/* get file pointer. */
	filp = fdt->fd[fd];
	rcu_assign_pointer(fdt->fd[fd], NULL);
	FD_CLR(fd, fdt->close_on_exec);
	__FD_CLR(fd, fdt->open_fds);
	if (fd < files->next_fd) {
		/* @@@ for DEBUG @@@ */
		TCPCP_DPRINTK("files->next_fd=%d\n",
			files->next_fd);

		files->next_fd = fd;
		/* @@@ for DEBUG @@@ */
		TCPCP_DPRINTK("fd            =%d\n", fd);
		TCPCP_DPRINTK("files->next_fd=%d\n",
			files->next_fd);
	}

	spin_unlock(&files->file_lock);

	/* close file. */
	error = filp_close(filp, files);

	/* end. */
	TCPCP_DPRINTK("end. error=%d\n", error);
	return error;
}

/**
 * do_tcpcp_close - close TCP sockets.
 *
 * [Args]
 *   (IN) *sk       : TCP socket
 *   (IN) *r_entry  : Request Entry
 *   (IN) *user_res : Processing Result Info.
 *
 * [Return]
 *   0      : OK.
 *  -ENOMEM : NG.
 **/
static int do_tcpcp_close(struct sock *sk, struct tcpcp_req_entry *r_entry,
			  struct result *user_res)
{
	struct task_struct *tsk = current;
	struct files_struct *files = tsk->files;
	struct tcpcp_com_entry *fdr_entry;
	unsigned int fd;
	unsigned int fd_all = 0;
	unsigned int fd_ng = 0;
	int error;
	struct fdtable *fdt;
	/* plug-in. */
	plugin_dummy();

	/* start. */
	TCPCP_DPRINTK("start. sk=0x%p, r_entry=0x%p, user_res=0x%p\n",
		sk, r_entry, user_res);

	/* add FDR Info. Entry. (only "ALL")*/
	error = tcpcp_add_com_entry(r_entry->flag, &(r_entry->f_entry));
	if (error != 0)
		return -ENOMEM;
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("r_entry->f_entry=0x%p\n",
		r_entry->f_entry);

	/* initialize now FDR Info. Entry. */
	fdr_entry = r_entry->f_entry;
	fdt = files_fdtable(files);
	/* search FD. */
	for (fd = 0; fd < fdt->max_fds; fd++) {
 		struct sock *sk_i;
 		int ret;
 
		/* check FD and get SOCKET_I(). */
		sk_i = tcpcp_get_sk_i(sk, fd, r_entry);
		if (sk_i == NULL) {
			/* @@@ for DEBUG @@@ */
			TCPCP_DPRINTK("sk_i is NULL.\n");

			continue;
		}
		/* close TCP socket. */
		/* 
		 * lock_sock and release_sock are unnecessary.
		 * because spin_lock to files is in do_close.
		 */
		ret = do_close(fd);

		fd_all++;
		if (ret != 0)
			fd_ng++;

		/* make FDR Info. (only "ALL") */
		error = tcpcp_make_res(fd, ret, &fdr_entry);
		if (error != 0) {
			/* only -ENOMEM. */
                        TCPCP_DPRINTK("[Error] can't make FDR Info."
				" error=%d\n", error);
			/* Caller delete Request Entry with common entries. */
			return error;
		}
	}
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("Loop count=%d\n", fd);

#ifdef CONFIG_TCPCP_DEBUG
	/* @@@ print Process List with lock. (for DEBUG) @@@ */
	print_list();
#endif

	/* set Processing Result Info. */
	user_res->fd_all = fd_all;
	user_res->fd_ng = fd_ng;
	user_res->buf_size = 0;
	user_res->fdrs_size = sizeof(struct fdr) * fd_all;
	memcpy(&(r_entry->res), user_res, sizeof(struct result));
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("r_entry->res->fd_all=%u\n",
		r_entry->res.fd_all);
	TCPCP_DPRINTK("r_entry->res->fd_ng =%u\n",
		r_entry->res.fd_ng);
	TCPCP_DPRINTK("r_entry->res->buf_size =%u\n",
		r_entry->res.buf_size);
	TCPCP_DPRINTK("r_entry->res->fdrs_size=%u\n",
		r_entry->res.fdrs_size);

	/* end. */
	TCPCP_DPRINTK("end.\n");
	return 0;
}


/* ----- set TCP Socket Informations (TCP-SIs) ----------------------------- */

/* ----- TCP-SI buffer management ----- */

/**
 * tcpcp_clear_si_buf - clear TCP-SI buffer.
 *
 * [Args]
 *   (IN) *si_buf : TCP-SI buffer
 **/
static void tcpcp_clear_si_buf(struct tcpcp_si_buf *si_buf)
{
	/* plug-in. */
	plugin_dummy();

	/* start. */
	TCPCP_DPRINTK("start. si_buf=0x%p\n", si_buf);

	/* clear TCP-SI header buffer and packet info. buffer. */
	memset(&(si_buf->si_hdr), 0, sizeof(struct tcpcp_si_hdr));
	memset(si_buf->snd_buf.si_skb, 0, TCPCP_SI_SKB_BUF_SIZE);
	memset(si_buf->rcv_buf.si_skb, 0, TCPCP_SI_SKB_BUF_SIZE);

	/* clear length and count of snd/rcv buffer. */
	si_buf->snd_buf.buf_len = 0;
	si_buf->snd_buf.skb_cnt = 0;
	si_buf->rcv_buf.buf_len = 0;
	si_buf->rcv_buf.skb_cnt = 0;
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("si_buf->snd_buf.buf_len=%u\n",
		si_buf->snd_buf.buf_len);
	TCPCP_DPRINTK("si_buf->snd_buf.skb_cnt=%u\n",
		si_buf->snd_buf.skb_cnt);
	TCPCP_DPRINTK("si_buf->rcv_buf.buf_len=%u\n",
		si_buf->rcv_buf.buf_len);
	TCPCP_DPRINTK("si_buf->rcv_buf.skb_cnt=%u\n",
		si_buf->rcv_buf.skb_cnt);

	/* end. */
	TCPCP_DPRINTK("end.\n");
	return;
}

/**
 * tcpcp_alloc_si_buf - allocate TCP-SI buffer.
 *
 * [Args]
 *   (IN) *si_buf : TCP-SI buffer
 *
 * [Return]
 *   0      : OK.
 *  -ENOMEM : NG.
 **/
static int tcpcp_alloc_si_buf(struct tcpcp_si_buf *si_buf)
{
	/* plug-in. */
	plugin_dummy();

	/* start. */
	TCPCP_DPRINTK("start. si_buf=0x%p\n", si_buf);

	/* allocate TCP-SI buffer for send-packet info. */
	si_buf->snd_buf.si_skb = kmalloc(TCPCP_SI_SKB_BUF_SIZE, GFP_KERNEL);
	if (si_buf->snd_buf.si_skb == NULL) {
		TCPCP_DPRINTK("[Error] can't allocate TCP-SI buffer"
			" (1).\n");
		return -ENOMEM;
	}
	si_buf->snd_buf.max_buf_len = TCPCP_SI_SKB_BUF_SIZE;
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("TCPCP_SI_SKB_BUF_SIZE      =%lu\n",
		TCPCP_SI_SKB_BUF_SIZE);
	TCPCP_DPRINTK("si_buf->snd_buf.max_buf_len=%u\n",
		si_buf->snd_buf.max_buf_len);

	/* allocate TCP-SI buffer for receive-packet info. */
	si_buf->rcv_buf.si_skb = kmalloc(TCPCP_SI_SKB_BUF_SIZE, GFP_KERNEL);
	if (si_buf->rcv_buf.si_skb == NULL) {
		TCPCP_DPRINTK("[Error] can't allocate TCP-SI buffer"
			" (2).\n");

		kfree(si_buf->snd_buf.si_skb);
		si_buf->snd_buf.si_skb = NULL;
		/* @@@ for DEBUG @@@ */
		TCPCP_DPRINTK("si_buf->snd_buf.si_skb is NULL.\n");

		return -ENOMEM;
	}
	si_buf->rcv_buf.max_buf_len = TCPCP_SI_SKB_BUF_SIZE;
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("si_buf->rcv_buf.max_buf_len=%u\n",
		si_buf->rcv_buf.max_buf_len);

	/* clear TCP-SI buffer. */
	tcpcp_clear_si_buf(si_buf);

	/* end. */
	TCPCP_DPRINTK("end.\n");
	return 0;
}

/**
 * tcpcp_free_si_buf - free TCP-SI buffer.
 *
 * [Args]
 *   (IN) *si_buf : TCP-SI buffer
 **/
static void tcpcp_free_si_buf(struct tcpcp_si_buf *si_buf)
{
	/* plug-in. */
	plugin_dummy();

	/* start. */
	TCPCP_DPRINTK("start. si_buf=0x%p\n", si_buf);

	/* clear TCP-SI buffer. */
	tcpcp_clear_si_buf(si_buf);

	/* free TCP-SI buffer for packet info. */
	kfree(si_buf->snd_buf.si_skb);
	kfree(si_buf->rcv_buf.si_skb);
	si_buf->snd_buf.max_buf_len = 0;
	si_buf->rcv_buf.max_buf_len = 0;
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("si_buf->snd_buf.max_buf_len=%u\n",
		si_buf->snd_buf.max_buf_len);
	TCPCP_DPRINTK("si_buf->rcv_buf.max_buf_len=%u\n",
		si_buf->rcv_buf.max_buf_len);

	/* end. */
	TCPCP_DPRINTK("end.\n");
	return;
}

/**
 * tcpcp_copy_si_from_user - copy TCP-SI from userland to Kernel.
 *
 * This function is based on TCPCP-9:tcpcp_buffers().
 *
 * [Args]
 *   (IN) *user_si    : user TCP-SI
 *   (IN) *si_buf     : TCP-SI buffer
 *   (IN) total_ucopy : un-copied TCP-SIs
 *
 * [Return]
 *   0      : OK.
 *  -EFAULT : NG.
 *  -ENOMEM : NG.
 *  -EINVAL : NG.
 **/
static int tcpcp_copy_si_from_user(const struct tcpcp_si_hdr *user_si,
				   struct tcpcp_si_buf *si_buf,
				   unsigned int total_ucopy)
{
	struct tcpcp_si_skb *si_skb;
	unsigned int si_hdr_len;
	unsigned int si_len;
	unsigned int ucopy;
	int error = 0;
	char *r_ptr;

	/* plug-in. */
	plugin_dummy();

	/* start. */
	TCPCP_DPRINTK("start. user_si=0x%p, si_buf=0x%p"
		" total_ucopy=%u\n",
		user_si, si_buf, total_ucopy);

	/* set read-point. */
	r_ptr = (char *)user_si;
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("r_ptr=0x%p\n", r_ptr);

	si_hdr_len = sizeof(struct tcpcp_si_hdr);
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("total_ucopy=%u\n", total_ucopy);
	TCPCP_DPRINTK("si_hdr_len =%u\n", si_hdr_len);

	/* check length of TCP-SI header. */
	if (total_ucopy < si_hdr_len)
		return -EINVAL;

	/* get length of TCP-SI. (include TCP-SI header) */
	si_len = ntohl(user_si->si_length) * 4;
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("si_len=%u\n", si_len);

	/* check length of TCP-SI. */
	if (si_len < si_hdr_len)
		return -EINVAL;
	if (total_ucopy < si_len)
		return -EINVAL;

	/* copy TCP-SI header from userland to Kernel. */
	error = copy_from_user(&(si_buf->si_hdr), user_si, si_hdr_len);
	if (error != 0) {
		TCPCP_DPRINTK("[Error] can't copy TCP-SI header from"
			"userland. error=%d\n", error);
		return -EFAULT;
	}

	/* move read-point. */
	r_ptr += si_hdr_len;
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("r_ptr=0x%p\n", r_ptr);

	/* set un-copied TCP-SI. */
	ucopy = si_len - si_hdr_len;
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("ucopy=%u\n", ucopy);

	while (ucopy != 0) {
		struct tcpcp_si_skb_buf *si_skb_buf;
		unsigned int si_skb_len;

		/* check length of packet info. */
		if (ucopy < sizeof(struct tcpcp_si_skb))
			return -EINVAL;

		si_skb = (struct tcpcp_si_skb *)r_ptr;

		/* calculate length of packet info. */
		si_skb_len = sizeof(struct tcpcp_si_skb)
			     + ((ntohs(si_skb->length) + 3) & ~3);
		/* @@@ for DEBUG @@@ */
		TCPCP_DPRINTK("si_skb_len=%u\n",
			si_skb_len);

		/* check length of packet info. */
		if (ucopy < si_skb_len)
			return -EINVAL;

		/* @@@ for DEBUG @@@ */
		TCPCP_DPRINTK("si_skb->type=%u\n",
			si_skb->type);

		switch (si_skb->type) {
		case TCPCP_SIE_BUF_SND:
			si_skb_buf = &(si_buf->snd_buf);
			break;
		case TCPCP_SIE_BUF_RCV: 
			si_skb_buf = &(si_buf->rcv_buf);
			break;
		default:
			return -EINVAL;
		}
		/* @@@ for DEBUG @@@ */
		TCPCP_DPRINTK("&(si_buf->snd_buf)=0x%p\n",
			&(si_buf->snd_buf));
		TCPCP_DPRINTK("&(si_buf->rcv_buf)=0x%p\n",
			&(si_buf->rcv_buf));
		TCPCP_DPRINTK("si_skb_buf        =0x%p\n",
			si_skb_buf);

		/* @@@ for DEBUG @@@ */
		TCPCP_DPRINTK("si_skb_buf->max_buf_len=%u\n",
			si_skb_buf->max_buf_len);
		TCPCP_DPRINTK("si_skb_buf->buf_len    =%u\n",
			si_skb_buf->buf_len);

		/* check free of packet info. buffer. */
		if (si_skb_buf->max_buf_len - si_skb_buf->buf_len < si_skb_len)
			return -EINVAL;

		/* copy packet info. from userland. */
		error = copy_from_user(si_skb_buf->si_skb + si_skb_buf->buf_len,
				       r_ptr, si_skb_len);
		if (error != 0) {
			TCPCP_DPRINTK("[Error] can't copy packet info."
				" from userland. error=%d\n",
				error);
			return -EFAULT;
		}

		si_skb_buf->buf_len += si_skb_len;
		si_skb_buf->skb_cnt++;
		/* @@@ for DEBUG @@@ */
		TCPCP_DPRINTK("si_skb_len         =%u\n",
			si_skb_len);
		TCPCP_DPRINTK("si_skb_buf->buf_len=%u\n",
			si_skb_buf->buf_len);
		TCPCP_DPRINTK("si_skb_buf->skb_cnt=%u\n",
			si_skb_buf->skb_cnt);

		/* move read-point. */
		r_ptr += si_skb_len;
		ucopy -= si_skb_len;
		/* @@@ for DEBUG @@@ */
		TCPCP_DPRINTK("r_ptr=0x%p\n", r_ptr);
		TCPCP_DPRINTK("ucopy=%u\n", ucopy);
	}

	/*  end. */
	TCPCP_DPRINTK("end.\n");
	return 0;
}

/* ----- set TCP-SI ----- */

/**
 * tcpcp_check_compatibility - Step 1 : check compatibility.
 *
 * This function is based on TCPCP-9:tcpcp_compatibility().
 * It's checks that a connection with the characteristics described in TCP-SI
 * header is compatible with the functionality available on the local system.
 * If not, the operation fails.
 * (E.g. if SACK is disabled here, we cannot negotiate out of using it.)
 * 
 * [Args]
 *   (IN) *si_hdr : TCP-SI header
 *
 * [Return]
 *   0           : OK.
 *  -ENOPROTOOPT : NG.
 **/
static int tcpcp_check_compatibility(const struct tcpcp_si_hdr *si_hdr)
{
	/* plug-in. */
	plugin_dummy();

	/* start. */
	TCPCP_DPRINTK("start. si_hdr=0x%p\n", si_hdr);

	if ((si_hdr->fixgen.tcp_flags & TCPI_OPT_TIMESTAMPS)
	    && (sysctl_tcp_timestamps == 0))
		return -ENOPROTOOPT;
	if ((si_hdr->fixgen.tcp_flags & TCPI_OPT_SACK)
	    && (sysctl_tcp_sack == 0))
		return -ENOPROTOOPT;
	if ((si_hdr->fixgen.tcp_flags & TCPI_OPT_WSCALE)
	    && (sysctl_tcp_window_scaling == 0))
		return -ENOPROTOOPT;
	if ((si_hdr->fixgen.tcp_flags & TCPI_OPT_ECN) && (sysctl_tcp_ecn == 0))
		return -ENOPROTOOPT;

	/* end. */
	TCPCP_DPRINTK("end.\n");
	return 0;
}

/**
 * tcpcp_setopt_before_bind - Step 2 : set socket options. (before pre-bind)
 *
 * This function is based on sock.c:sock_setsockopt(),
 *                           ip_sockglue.c:ip_setsockopt(),
 *                           ipv6_sockglue.c:ipv6_setsockopt(),
 *                           tcp.c:tcp_setsockopt().
 * Caller must lock socket.
 *
 * [Args]
 *   (IN) *sk     : TCP socket
 *   (IN) *si_hdr : TCP-SI header
 **/
static void tcpcp_setopt_before_bind(struct sock *sk,
				     const struct tcpcp_si_hdr *si_hdr)
{
	struct inet_sock *inet = inet_sk(sk);
#if defined(CONFIG_IPV6) || defined(CONFIG_IPV6_MODULE)
	struct ipv6_pinfo *inet6 = inet6_sk(sk);
#endif
	int valbool;
	uint8_t val8 = 0;
	uint32_t val32 = 0;

	/* plug-in. */
	plugin_dummy();

	/* start. */
	TCPCP_DPRINTK("start. sk=0x%p, si_hdr=0x%p\n",
		sk, si_hdr);

	/*
	 * set socket options. (socket)
	 */

	/* set SO_REUSEADDR flag. */
	if (si_hdr->sockopt.sk.flags & TCPCP_SIE_SOCK_REUSEADDR)
		sk->sk_reuse = 1;
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("si_hdr->sockopt.sk.flags=%d\n",
		si_hdr->sockopt.sk.flags);
	TCPCP_DPRINTK("sk->sk_reuse=%d\n", sk->sk_reuse);

	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("sk->sk_flags=%lu\n", sk->sk_flags);

	/* set SO_KEEPOPEN flag. */
	valbool = (si_hdr->sockopt.sk.flags & TCPCP_SIE_SOCK_KEEPOPEN);
	tcpcp_set_keepalive(sk, valbool);
	sock_valbool_flag(sk, SOCK_KEEPOPEN, valbool);
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("sk->sk_flags=%lu\n", sk->sk_flags);

	/* set SOCK_LINGER flag and time. */
	if (!(si_hdr->sockopt.sk.flags & TCPCP_SIE_SOCK_LINGER))
		sock_reset_flag(sk, SOCK_LINGER);
	else {
		val32 = ntohl(si_hdr->sockopt.sk.l_linger);
		sk->sk_lingertime = val32 * HZ;
		sock_set_flag(sk, SOCK_LINGER);
	}
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("si_hdr->sockopt.sk.flags=%u\n",
		si_hdr->sockopt.sk.flags);
	TCPCP_DPRINTK("sk->sk_flags=%lu\n",
		sk->sk_flags);
	TCPCP_DPRINTK("si_hdr->sockopt.sk.l_linger=%u\n",
		si_hdr->sockopt.sk.l_linger);
	TCPCP_DPRINTK("val32                      =%u\n",
		val32);
	TCPCP_DPRINTK("sk->sk_lingertime          =%lu\n",
		sk->sk_lingertime);

	/* set size of buffer. */
	sk->sk_userlocks = si_hdr->sockopt.sk.userlocks;
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("si_hdr->sockopt.sk.userlocks=%u\n",
		si_hdr->sockopt.sk.userlocks);
	TCPCP_DPRINTK("sk->sk_userlocks            =%u\n",
		sk->sk_userlocks);

	/* set size of send buffer in bytes. */
	val32 = ntohl(si_hdr->sockopt.sk.sndbuf);
  if (val32 > __sysctl_wmem_max)
    val32 = __sysctl_wmem_max;
   /* if (val32 > sk->sk_prot->sysctl_wmem[1])
      val32 = sk->sk_prot->sysctl_wmem[1]; */
	if (val32 * 2 < SOCK_MIN_SNDBUF)
		sk->sk_sndbuf = SOCK_MIN_SNDBUF;
	else
		sk->sk_sndbuf = val32 * 2;
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("si_hdr->sockopt.sk.sndbuf=%u\n",
		si_hdr->sockopt.sk.sndbuf);
	TCPCP_DPRINTK("val32                    =%u\n",
		val32);
	TCPCP_DPRINTK("__sysctl_wmem_max          =%u\n",
		__sysctl_wmem_max);
	TCPCP_DPRINTK("sk->sk_sndbuf            =%u\n",
		sk->sk_sndbuf);

	/* set size of receive buffer in bytes. */
	val32 = ntohl(si_hdr->sockopt.sk.rcvbuf);
  if (val32 > __sysctl_rmem_max)
      val32 = __sysctl_rmem_max;
   /* if (val32 > sk->sk_prot->sysctl_rmem[1])
      val32 = sk->sk_prot->sysctl_rmem[1]; */
	if (val32 * 2 < SOCK_MIN_RCVBUF)
		sk->sk_rcvbuf = SOCK_MIN_RCVBUF;
	else
		sk->sk_rcvbuf = val32 * 2;
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("si_hdr->sockopt.sk.rcvbuf=%u\n",
		si_hdr->sockopt.sk.rcvbuf);
	TCPCP_DPRINTK("val32                    =%u\n",
		val32);
	TCPCP_DPRINTK("__sysctl_rmem_max          =%u\n",
		__sysctl_rmem_max);
	TCPCP_DPRINTK("sk->sk_rcvbuf            =%u\n",
		sk->sk_rcvbuf);

	/* set SO_PRIORITY. */
	sk->sk_priority = ntohl(si_hdr->sockopt.sk.priority);
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("si_hdr->sockopt.sk.priority=%u\n",
		si_hdr->sockopt.sk.priority);
	TCPCP_DPRINTK("sk->sk_priorityi           =%u\n",
		sk->sk_priority);

#if defined(CONFIG_IPV6) || defined(CONFIG_IPV6_MODULE)
	if (si_hdr->id.ip_version == 6) {
		/*
		 * set socket options. (IPv6)
		 */

		/* @@@ for DEBUG @@@ */
		TCPCP_DPRINTK("si_hdr->sockopt.ip.v6.flags=%u\n",
			si_hdr->sockopt.ip.v6.flags);

		/* set unicast hop limit. */
		if (si_hdr->sockopt.ip.v6.flags
		    & TCPCP_SIE_IPV6_DEFAULT_HOPLIMIT)
			inet6->hop_limit = -1;
		else {
			val8 = si_hdr->sockopt.ip.v6.hop_limit;
			inet6->hop_limit = val8;
		}
		/* @@@ for DEBUG @@@ */
		TCPCP_DPRINTK("si_hdr->sockopt.ip.v6.hop_limit=%u\n",
			si_hdr->sockopt.ip.v6.hop_limit);
		TCPCP_DPRINTK("val8                           =%u\n",
			val8);
		TCPCP_DPRINTK("inet6->hop_limit               =%u\n",
			inet6->hop_limit);

		/* set IP_MTU_DISCOVER state. */
		val8 = si_hdr->sockopt.ip.v6.pmtudisc;
		if (val8 <= 2)
			inet6->pmtudisc = val8;
		/* @@@ for DEBUG @@@ */
		TCPCP_DPRINTK("si_hdr->sockopt.ip.v6.pmtudisc=%u\n",
			si_hdr->sockopt.ip.v6.pmtudisc);
		TCPCP_DPRINTK("val8                          =%u\n",
			val8);
		TCPCP_DPRINTK("inet6->pmtudisc               =%u\n",
			inet6->pmtudisc);

	} else
#endif
	{
		/*
		 * set socket options. (IPv4)
		 */

		/* set unicast TTL. */
		if (si_hdr->sockopt.ip.v4.flags & TCPCP_SIE_IP_DEFAULT_TTL)
			inet->uc_ttl = -1;
		else {
			val8 = si_hdr->sockopt.ip.v4.uc_ttl;
			if (val8 != 0)
				inet->uc_ttl = val8;
		}
		/* @@@ for DEBUG @@@ */
		TCPCP_DPRINTK("si_hdr->sockopt.ip.v4.flags =%u\n",
			si_hdr->sockopt.ip.v4.flags);
		TCPCP_DPRINTK("si_hdr->sockopt.ip.v4.uc_ttl=%u\n",
			si_hdr->sockopt.ip.v4.uc_ttl);
		TCPCP_DPRINTK("val8                        =%u\n",
			val8);
		TCPCP_DPRINTK("inet->uc_ttl                =%d\n",
			inet->uc_ttl);

		/* set IP_MTU_DISCOVER state. */
		val8 = si_hdr->sockopt.ip.v4.pmtudisc;
		if (val8 <= 2)
			inet->pmtudisc = val8;
		/* @@@ for DEBUG @@@ */
		TCPCP_DPRINTK("si_hdr->sockopt.ip.v4.pmtudisc=%u\n",
			si_hdr->sockopt.ip.v4.pmtudisc);
		TCPCP_DPRINTK("val8                          =%u\n",
			val8);
		TCPCP_DPRINTK("inet->pmtudisc                =%u\n",
			inet->pmtudisc);

		/* set TOS. */
		val8 = si_hdr->sockopt.ip.v4.tos;
		val8 &= ~3;
		val8 |= inet->tos & 3;
		inet->tos = val8;
		/* @@@ for DEBUG @@@ */
		TCPCP_DPRINTK("si_hdr->sockopt.ip.v4.tos=%u\n",
			si_hdr->sockopt.ip.v4.tos);
		TCPCP_DPRINTK("val8                     =%u\n",
			val8);
		TCPCP_DPRINTK("inet->tos                =%d\n",
			inet->tos);
	}

	/* end. */
	TCPCP_DPRINTK("end.\n");
	return;
}

/**
 * tcpcp_inet_addr_type - get type of IPv4 address.
 * 
 * This function is based on fib_frontend.c:inet_addr_type().
 * 
 * [Args]
 *   (IN) addr : IPv4 address
 * 
 * [Return]
 *   RTN_LOCAL
 *   RTN_BROADCAST
 *   RTN_MULTICAST
 **/
static unsigned tcpcp_inet_addr_type(u32 addr)
{
	unsigned ret = RTN_LOCAL;

	/* plug-in. */
	plugin_dummy();

	/* start. */
	TCPCP_DPRINTK("start. addr=0x%x\n", addr);

	if (ipv4_is_zeronet(addr) || ipv4_is_lbcast(addr))
		return RTN_BROADCAST;
	if (ipv4_is_multicast(addr))
		return RTN_MULTICAST;

	/* end. */
	TCPCP_DPRINTK("end.\n");
	return ret;
}

#if defined(CONFIG_IPV6) || defined(CONFIG_IPV6_MODULE)
/**
 * tcpcp_pre_v6bind - Step 3 : pre-processing of TCP/IPv6 bind.
 *
 * This function is based on af_inet6.c:inet6_bind().
 * Caller must lock socket.
 * 
 * [Args]
 *   (IN) *sk     : TCP socket
 *   (IN) *si_hdr : TCP-SI header
 * 
 * [Return]
 *   0             : OK.
 *  -EINVAL        : NG.
 *  -EADDRNOTAVAIL : NG.
 *  -EACCES        : NG.
 **/
static int tcpcp_pre_v6bind(struct sock *sk, const struct tcpcp_si_hdr *si_hdr)
{
	struct inet_sock *inet = inet_sk(sk);
	struct ipv6_pinfo *inet6 = inet6_sk(sk);

	struct in6_addr ip_src;
	unsigned short snum;
	__u32 scope_id;
	__u32 v4addr = 0;
	int addr_type = 0;

	/* plug-in. */
	plugin_dummy();

	/* start. */
	TCPCP_DPRINTK("start. sk=0x%p, si_hdr=0x%p\n",
		sk, si_hdr);

	/* check Bind hash. */
	if (inet_csk(sk)->icsk_bind_hash != NULL)
		return -EBADFD;

	/* get IPv6 source address and TCP source port. */
	ipv6_addr_copy(&ip_src, &(si_hdr->id.ip.v6.ip_src));
	snum = ntohs(si_hdr->id.tcp_sport);
	scope_id = ntohl(si_hdr->id.ip.v6.scope_id); 
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("si_hdr->id.ip.v6.ip_src=%04x:%04x:%04x:%04x:"
		"%04x:%04x:%04x:%04x\n",
		ntohs(si_hdr->id.ip.v6.ip_src.s6_addr16[0]),
		ntohs(si_hdr->id.ip.v6.ip_src.s6_addr16[1]),
		ntohs(si_hdr->id.ip.v6.ip_src.s6_addr16[2]),
		ntohs(si_hdr->id.ip.v6.ip_src.s6_addr16[3]),
		ntohs(si_hdr->id.ip.v6.ip_src.s6_addr16[4]),
		ntohs(si_hdr->id.ip.v6.ip_src.s6_addr16[5]),
		ntohs(si_hdr->id.ip.v6.ip_src.s6_addr16[6]),
		ntohs(si_hdr->id.ip.v6.ip_src.s6_addr16[7]));
	TCPCP_DPRINTK("ip_src                 =%04x:%04x:%04x:%04x:"
		"%04x:%04x:%04x:%04x\n",
		ntohs(ip_src.s6_addr16[0]),
		ntohs(ip_src.s6_addr16[1]),
		ntohs(ip_src.s6_addr16[2]),
		ntohs(ip_src.s6_addr16[3]),
		ntohs(ip_src.s6_addr16[4]),
		ntohs(ip_src.s6_addr16[5]),
		ntohs(ip_src.s6_addr16[6]),
		ntohs(ip_src.s6_addr16[7]));
	TCPCP_DPRINTK("si_hdr->id.tcp_sport=%u\n",
		si_hdr->id.tcp_sport);
	TCPCP_DPRINTK("snum                =%u\n",
		snum);
	TCPCP_DPRINTK("si_hdr->id.ip.v6.scope_id=%u\n",
		si_hdr->id.ip.v6.scope_id);
	TCPCP_DPRINTK("scope_id                 =%u\n",
		scope_id);

	/* get type of IPv6 source address. */
	addr_type = ipv6_addr_type(&ip_src);
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("addr_type=%u\n", addr_type);

	/* check type of IPv6 source address. */
	if ((addr_type == IPV6_ADDR_ANY) || (addr_type & IPV6_ADDR_MULTICAST)) {
		/* TCPCP not support IPV6_ADDR_ANY and multicast. */
		TCPCP_DPRINTK("[Error] type of ip_src is ANY or"
			" multicast.\n");
		return -EADDRNOTAVAIL;
	}

	/* check TCP source port. */
	if (snum == 0) {
		TCPCP_DPRINTK("[Error] tcp_sport is 0.\n");
		return -EINVAL;
	}
	if ((snum < PROT_SOCK) && (!capable(CAP_NET_BIND_SERVICE)))
		return -EACCES;

	/* check if the address belongs to the host. */
	if (addr_type == IPV6_ADDR_MAPPED) {
		/* mapped address. */
		v4addr = ip_src.s6_addr32[3];
		/* @@@ for DEBUG @@@ */
		TCPCP_DPRINTK("ip_src=%04x:%04x:%04x:%04x:"
			"%04x:%04x:%04x:%04x\n",
			ntohs(ip_src.s6_addr16[0]),
			ntohs(ip_src.s6_addr16[1]),
			ntohs(ip_src.s6_addr16[2]),
			ntohs(ip_src.s6_addr16[3]),
			ntohs(ip_src.s6_addr16[4]),
			ntohs(ip_src.s6_addr16[5]),
			ntohs(ip_src.s6_addr16[6]),
			ntohs(ip_src.s6_addr16[7]));
		TCPCP_DPRINTK("v4addr=0x%x\n", v4addr);

		if (tcpcp_inet_addr_type(v4addr) != RTN_LOCAL)
			return -EADDRNOTAVAIL;

	} else if (addr_type & IPV6_ADDR_LINKLOCAL) {
		/* link local address. */
		if (scope_id != 0)
			/* override any existing binding,
			 * if another one is supplied by user. */
			sk->sk_bound_dev_if = scope_id;
		/* @@@ for DEBUG @@@ */
		TCPCP_DPRINTK("sk->sk_bound_dev_if=%u\n",
			sk->sk_bound_dev_if);

		/* binding to link-local address requires an interface. */
		if (sk->sk_bound_dev_if == 0)
			return -EINVAL;

		/* IPv4 address of the socket is invalid.
		 * only the unspecified and mapped address have a IPv4
		 * equivalent. */
		v4addr = LOOPBACK4_IPV6;
		/* @@@ for DEBUG @@@ */
		TCPCP_DPRINTK("v4addr=0x%x\n", v4addr);
	}

	/* set IPv4 source address. */
	inet->rcv_saddr = v4addr;
	inet->saddr = v4addr;
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("v4addr         =0x%x\n",
		v4addr);
	TCPCP_DPRINTK("inet->rcv_saddr=0x%x\n",
		inet->rcv_saddr);
	TCPCP_DPRINTK("inet->saddr    =0x%x\n",
		inet->saddr);

	/* set IPv6 source address. */
	ipv6_addr_copy(&(inet6->rcv_saddr), &ip_src);
	ipv6_addr_copy(&(inet6->saddr), &ip_src);
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("inet6->rcv_saddr=%04x:%04x:%04x:%04x:"
		"%04x:%04x:%04x:%04x\n",
		ntohs(inet6->rcv_saddr.s6_addr16[0]),
		ntohs(inet6->rcv_saddr.s6_addr16[1]),
		ntohs(inet6->rcv_saddr.s6_addr16[2]),
		ntohs(inet6->rcv_saddr.s6_addr16[3]),
		ntohs(inet6->rcv_saddr.s6_addr16[4]),
		ntohs(inet6->rcv_saddr.s6_addr16[5]),
		ntohs(inet6->rcv_saddr.s6_addr16[6]),
		ntohs(inet6->rcv_saddr.s6_addr16[7]));
	TCPCP_DPRINTK("inet6->saddr    =%04x:%04x:%04x:%04x:"
		"%04x:%04x:%04x:%04x\n",
		ntohs(inet6->saddr.s6_addr16[0]),
		ntohs(inet6->saddr.s6_addr16[1]),
		ntohs(inet6->saddr.s6_addr16[2]),
		ntohs(inet6->saddr.s6_addr16[3]),
		ntohs(inet6->saddr.s6_addr16[4]),
		ntohs(inet6->saddr.s6_addr16[5]),
		ntohs(inet6->saddr.s6_addr16[6]),
		ntohs(inet6->saddr.s6_addr16[7]));

	/* set TCP source port. */
	inet->sport = htons(snum);
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("inet->sport=%u\n", inet->sport);

	/* initialize IPv4 destination address and TCP destination port. */
	inet->daddr = 0;
	inet->dport = 0;
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("inet->daddr=0x%x\n", inet->daddr);
	TCPCP_DPRINTK("inet->dport=%u\n", inet->dport);

	/* end. */
	TCPCP_DPRINTK("end.\n");
	return 0;
}
#endif /* defined(CONFIG_IPV6) || defined(CONFIG_IPV6_MODULE) */

/**
 * tcpcp_pre_v4bind - Step 3 : pre-processing of TCP/IPv4 bind.
 * 
 * This function is based on af_inet.c:inet_bind().
 * Caller must lock socket.
 * 
 * [Args]
 *   (IN) *sk     : TCP socket
 *   (IN) *si_hdr : TCP-SI header
 * 
 * [Return]
 *   0             : OK.
 *  -EINVAL        : NG.
 *  -EACCES        : NG.
 *  -EADDRNOTAVAIL : NG.
 **/
static int tcpcp_pre_v4bind(struct sock *sk, const struct tcpcp_si_hdr *si_hdr)
{
	struct inet_sock *inet = inet_sk(sk);

	u32 s_addr;
	unsigned short snum;
	int chk_addr_ret;

	/* plug-in. */
	plugin_dummy();

	/* start. */
	TCPCP_DPRINTK("start. si_hdr=0x%p\n", si_hdr);

	/* check Bind hash. */
	if (inet_csk(sk)->icsk_bind_hash != NULL)
		return -EBADFD;

	/* get IPv4 source address and TCP source port. */
	s_addr = si_hdr->id.ip.v4.ip_src.s_addr;
	snum = ntohs(si_hdr->id.tcp_sport);
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("si_hdr->id.ip.v4.ip_src=0x%x\n",
		si_hdr->id.ip.v4.ip_src.s_addr);
	TCPCP_DPRINTK("s_addr                 =0x%x\n",
		s_addr);
	TCPCP_DPRINTK("si_hdr->id.tcp_sport=%u\n",
		si_hdr->id.tcp_sport);
	TCPCP_DPRINTK("snum                =%u\n",
		snum);

	/* check IPv4 source address. */
	if (s_addr == 0) {
		TCPCP_DPRINTK("[Error] ip_src is 0.0.0.0.\n");
		return -EINVAL;
	}

	/* get type of IPv4 source address. */
	chk_addr_ret = tcpcp_inet_addr_type(s_addr);
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("addr_type=%d\n", chk_addr_ret);

	/* check type of IPv4 source address. */
	if ((chk_addr_ret == RTN_MULTICAST)
	    || (chk_addr_ret == RTN_BROADCAST)) {
		/* TCPCP not support multicast and broadcast. */
		TCPCP_DPRINTK("[Error] type of ip_src is"
			" multi/broadcast.\n");
		return -EADDRNOTAVAIL;
	}

	/* check TCP source port. */
	if (snum == 0) {
		TCPCP_DPRINTK("[Error] tcp_sport is 0.\n");
		return -EINVAL;
	}
	if ((snum < PROT_SOCK) && (!capable(CAP_NET_BIND_SERVICE)))
		return -EACCES;

	/* set IPv4 source address. */
	inet->rcv_saddr = inet->saddr = s_addr;
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("inet->saddr    =0x%x\n",
		inet->saddr);
	TCPCP_DPRINTK("inet->rcv_saddr=0x%x\n",
		inet->rcv_saddr);

	/* set TCP source port. */
	inet->sport = htons(snum);
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("inet->sport=%u\n", inet->sport);

	/* initialize IPv4 destination address and TCP destination port. */
	inet->daddr = 0;
	inet->dport = 0;
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("inet->daddr=0x%x\n", inet->daddr);
	TCPCP_DPRINTK("inet->dport=%u\n", inet->dport);

	/* reset destination. */
	sk_dst_reset(sk);
	/* @@@ for DEBUG @@@ */
	if (sk->sk_dst_cache == NULL)
		TCPCP_DPRINTK("sk->sk_dst_cache is NULL.\n");

	/* end. */
	TCPCP_DPRINTK("end.\n");
	return 0;
}

/**
 * tcpcp_setopt_after_bind - Step 4 : set socket options. (after pre-bind)
 *
 * This function is based on sock.c:sock_setsockopt(),
 *                           ip_sockglue.c:ip_setsockopt(),
 *                           tcp.c:tcp_setsockopt().
 * Caller must lock socket.
 *
 * [Args]
 *   (IN) *sk     : TCP socket
 *   (IN) *si_hdr : TCP-SI header
 **/
static void tcpcp_setopt_after_bind(struct sock *sk,
				    const struct tcpcp_si_hdr *si_hdr)
{
	struct inet_sock *inet = inet_sk(sk);
	struct tcp_sock *tp = tcp_sk(sk);
	struct ip_options *opt = NULL;
	int valbool;
	uint8_t val8;
	uint16_t val16;
	int error;

	/* plug-in. */
	plugin_dummy();

	/* start. */
	TCPCP_DPRINTK("start. sk=0x%p, si_hdr=0x%p\n",
		sk, si_hdr);

	/*
	 * set socket options. (sock.c)
	 */

	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("sk->sk_flags=%lu\n",
		sk->sk_flags);

	/* set SO_KEEPOPEN flag. */
	valbool = (si_hdr->sockopt.sk.flags & TCPCP_SIE_SOCK_KEEPOPEN);
#ifdef CONFIG_INET
  if (sk->sk_protocol == IPPROTO_TCP) {
    tcpcp_set_keepalive(sk, valbool);
  }
#endif
	sock_valbool_flag(sk, SOCK_KEEPOPEN, valbool);
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("sk->sk_flags=%lu\n",
		sk->sk_flags);

	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("si_hdr->id.ip_version=%u\n",
		si_hdr->id.ip_version);

   /*
    * set socket options (ip_sockglue.c)
    */

   /* set IP options. */
  val8 = si_hdr->sockopt.ip.v4.opt_len;
   /* @@@ for DEBUG @@@ */
  TCPCP_DPRINTK("si_hdr->sockopt.ip.v4.opt_len=%u\n",
          si_hdr->sockopt.ip.v4.opt_len);

  if ((0 < val8) && (val8 <= 40)) {
     /* copy IP options. */
    error = ip_options_get_from_user(sock_net(sk), &opt,
                                     (unsigned char *)si_hdr->sockopt.ip.v4.opt_data,
                                     val8);
    if ( 0 == error ) {
      if (inet->is_icsk) {
        struct inet_connection_sock *icsk = inet_csk(sk);
#if defined(CONFIG_IPV6) || defined(CONFIG_IPV6_MODULE)
        if (sk->sk_family == PF_INET ||
            (!((1 << sk->sk_state) &
               (TCPF_LISTEN | TCPF_CLOSE)) &&
             inet->daddr != LOOPBACK4_IPV6)) {
#endif
          if (inet->opt) {
            icsk->icsk_ext_hdr_len -= inet->opt->optlen;
          }
          if (opt) {
            icsk->icsk_ext_hdr_len += opt->optlen;
          }
          icsk->icsk_sync_mss(sk, icsk->icsk_pmtu_cookie);
#if defined(CONFIG_IPV6) || defined(CONFIG_IPV6_MODULE)
        }
#endif
      }
      opt = xchg(&inet->opt, opt);
      if( NULL != opt ) {
        kfree(opt);
      }
    }
  }

	/*
	 * set socket options. (tcp.c)
	 */

	/* disable Nagle algorithm ? */
	tp->nonagle = si_hdr->sockopt.tcp.nonagle;
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("si_hdr->sockopt.tcp.nonagle=%u\n",
		si_hdr->sockopt.tcp.nonagle);
	TCPCP_DPRINTK("tp->nonagle                =%u\n",
		tp->nonagle);

	/* set num of allowed keep alive probes. */
	val8 = si_hdr->sockopt.tcp.keepcnt;
	if ((0 < val8) && (val8 <= MAX_TCP_KEEPCNT)) {
		tp->keepalive_probes = val8;
  }
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("si_hdr->sockopt.tcp.keepcnt=%u\n",
		si_hdr->sockopt.tcp.keepcnt);
	TCPCP_DPRINTK("val8                       =%u\n",
		val8);
	TCPCP_DPRINTK("tp->keepalive_probes       =%u\n",
		tp->keepalive_probes);

	/* set time before keep alive takes place. */
	val16 = ntohs(si_hdr->sockopt.tcp.keepidle);
	if ((0 < val16) && (val16 <= MAX_TCP_KEEPIDLE)) {
		tp->keepalive_time = val16 * HZ;
  }
   /* XXX: inet_csk_reset_keepalive_timer(sk, elapsed);? */
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("si_hdr->sockopt.tcp.keepidle=%u\n",
		si_hdr->sockopt.tcp.keepidle);
	TCPCP_DPRINTK("val16                       =%u\n",
		val16);
	TCPCP_DPRINTK("tp->keepalive_time/HZ       =%u\n",
		tp->keepalive_time / HZ);

	/* set time interval between keep alive probes. */
	val16 = ntohs(si_hdr->sockopt.tcp.keepintvl);
	if ((0 < val16) && (val16 <= MAX_TCP_KEEPINTVL)) {
		tp->keepalive_intvl = val16 * HZ;
  }
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("si_hdr->sockopt.tcp.keepintvl=%u\n",
		si_hdr->sockopt.tcp.keepintvl);
	TCPCP_DPRINTK("val16                        =%u\n",
		val16);
	TCPCP_DPRINTK("tp->keepalive_intvl/HZ       =%u\n",
		tp->keepalive_intvl / HZ);

	/* end. */
	TCPCP_DPRINTK("end.\n");
	return;
}

/**
 * tcpcp_init_connect - initialize TCP.
 *
 * This function is based on tcp_output.c:tcp_connect_init(),
 *                           tcp_output.c:tcp_connect().
 * 
 * [Args]
 *   (IN) *sk     : TCP socket
 *   (IN) *si_hdr : TCP-SI header
 **/
static void tcpcp_init_connect(struct sock *sk,
			       const struct tcpcp_si_hdr *si_hdr)
{
	struct tcp_sock *tp = tcp_sk(sk);
	/* plug-in. */
	plugin_dummy();

	/* start. */
	TCPCP_DPRINTK("start. si_hdr=0x%p\n", si_hdr);

	/*
	 * it's based on tcp_output.c:tcp_connect_init().
	 */

	/* initialize tcp_header_len. */
	tp->tcp_header_len = sizeof(struct tcphdr) +
		(sysctl_tcp_timestamps ? TCPOLEN_TSTAMP_ALIGNED : 0);
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("sysctl_tcp_timestamps=%u\n",
		sysctl_tcp_timestamps);
	TCPCP_DPRINTK("tp->tcp_header_len=%u\n",
		tp->tcp_header_len);

	/* initialize max_window. */
	tp->max_window = 0;
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("tp->max_window=%u\n",
		tp->max_window);

	/* set advmss. */
	tp->advmss = ntohs(si_hdr->fixgen.snd_mss); 
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("si_hdr->fixgen.snd_mss=%u\n",
		si_hdr->fixgen.snd_mss);
	TCPCP_DPRINTK("tp->advmss            =%u\n",
		tp->advmss);

	/* .... */
	tcp_initialize_rcv_mss(sk);
	tcpcp_ca_init(tp); /* ??? */
	TCPCP_DPRINTK("inet_csk(sk)->icsk_ack.rcv_mss =%u\n",
		      inet_csk(sk)->icsk_ack.rcv_mss);

	/* initialize window. sets window_clamp, rcv_wnd, and rcv_wscale */
	{
	  __u8 rcv_wscale;
	  tcpcp_select_initial_window(tcp_full_space(sk),
				      tp->advmss - (tp->rx_opt.ts_recent_stamp ?
						    tp->tcp_header_len
						    - sizeof(struct tcphdr) : 0),
				      &(tp->rcv_wnd),
				      &(tp->window_clamp),
				      sysctl_tcp_window_scaling,
				      &rcv_wscale);
	}

	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("tp->window_clamp=%u\n",
		tp->window_clamp);

	/* overwrite rcv_wnd and set rcv_ssthresh. */
	tp->rcv_wnd = ntohl(si_hdr->vargen.rcv_wnd);
	tp->rcv_ssthresh = tp->rcv_wnd;
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("si_hdr->vargen.rcv_wnd=%u\n",
		si_hdr->vargen.rcv_wnd);
	TCPCP_DPRINTK("tcp_receive_window(tp)=%u\n",
		      tcp_receive_window(tp));
	TCPCP_DPRINTK("tp->rcv_wnd           =%u\n",
		      tp->rcv_wnd);
	TCPCP_DPRINTK("tp->rcv_ssthresh      =%u\n",
		tp->rcv_ssthresh);

	/* overwrite rcv_wscale. */
	tp->rx_opt.rcv_wscale = si_hdr->fixgen.rcv_wscale;
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("si_hdr->fixgen.rcv_wscale=%d\n",
		si_hdr->fixgen.rcv_wscale);
	TCPCP_DPRINTK("tp->rx_opt.rcv_wscale           =%d\n",
		tp->rx_opt.rcv_wscale);


	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("sk->sk_flags=%lu\n",
		sk->sk_flags);

	/* initialize sk_err. */
	sk->sk_err = 0;
	sock_reset_flag(sk, SOCK_DONE);
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("sk->sk_err=%d\n",
		sk->sk_err);
	TCPCP_DPRINTK("sk->sk_flags=%lu\n",
		sk->sk_flags);

	/* initialize snd_wnd, snd_wl1, snd_una, snd_sml, rcv_nxt, rcv_wup and
	 * copied_seq. */
	tp->snd_wnd = 0;
	tp->snd_wl1 = tp->write_seq;
	tp->snd_una = tp->write_seq;
	tp->snd_sml = tp->write_seq;
	/* The next elements are set in tcpcp_recover_si_hdr */
	tp->rcv_nxt = 0;
	tp->rcv_wup = 0;
	tp->copied_seq = 0;
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("tp->snd_wnd   =%u\n",
		tp->snd_wnd);
	TCPCP_DPRINTK("tp->write_seq =%u\n",
		tp->write_seq);
	TCPCP_DPRINTK("tp->snd_wl1   =%u\n",
		tp->snd_wl1);
	TCPCP_DPRINTK("tp->snd_una   =%u\n",
		tp->snd_una);
	TCPCP_DPRINTK("tp->snd_sml   =%u\n",
		tp->snd_sml);

	/* initialize rto, retransmits. */
	inet_csk(sk)->icsk_rto = TCP_TIMEOUT_INIT;
	inet_csk(sk)->icsk_retransmits = 0;
	tcpcp_clear_retrans(tp);
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("inet_csk(sk)->icsk_rto=%u\n",
		inet_csk(sk)->icsk_rto);
	TCPCP_DPRINTK("inet_csk(sk)->icsk_retransmits=%d\n",
		inet_csk(sk)->icsk_retransmits);
	TCPCP_DPRINTK("tp->undo_marker=%u\n",
		tp->undo_marker);
	TCPCP_DPRINTK("tp->undo_retrans=%d\n",
		tp->undo_retrans);
	TCPCP_DPRINTK("tp->retrans_stamp=%u\n",
		tp->retrans_stamp);
	TCPCP_DPRINTK("tp->total_retrans=%u\n",
		tp->total_retrans);
	TCPCP_DPRINTK("tp->fackets_out=%u\n",
		tp->fackets_out);
	TCPCP_DPRINTK("tp->sacked_out =%u\n",
		tp->sacked_out);
	TCPCP_DPRINTK("tp->lost_out   =%u\n",
		tp->lost_out);

	/*
	 * it's based on tcp_input.c:tcp_connect().
	 */

	/* initialize snd_nxt, pushed_seq. */
	tp->snd_nxt = tp->write_seq;
	tp->pushed_seq = tp->write_seq;
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("tp->write_seq =%u\n",
		tp->write_seq);
	TCPCP_DPRINTK("tp->snd_nxt   =%u\n",
		tp->snd_nxt);
	TCPCP_DPRINTK("tp->pushed_seq=%u\n",
		tp->pushed_seq);

#if 1
	/*
	 * Estimate initial timestamp offset. This is only useful for
	 * the first packet. The first response (handled in
	 * tcp_input.c:tcp_rcv_established()) will call
	 * tcpcp_hooks.c:tcp_fix_ts() to set with fresh ack
	 * packet. Add offset (1) to make sure new packet is
	 * later. Note that vargen.ts_gen is the last tcp_time_stamp
	 * seen at the donor of the handoff.
	 */
	tp->ts_offset = 1 +ntohl(si_hdr->vargen.ts_gen) -tcp_time_stamp;
#endif
	TCPCP_DPRINTK("tp->ts_offset           =%u\n",
		      tp->ts_offset);
	TCPCP_DPRINTK("si_hdr->vargen.ts_gen   =%u\n",
		      ntohl(si_hdr->vargen.ts_gen));
	TCPCP_DPRINTK("tcp_time_stamp          =%u\n",
		      tcp_time_stamp);

	/* end. */
	TCPCP_DPRINTK("end.\n");
	return;
}

#if defined(CONFIG_IPV6) || defined(CONFIG_IPV6_MODULE)
/**
 * tcpcp_pre_v6connect - Step 5 : pre-processing of TCP/IPv6 connect.
 *
 * This function is based on tcp_ipv6.c:tcp_v6_connect().
 * Caller must lock socket.
 *
 * [Args]
 *   (IN) *sk     : TCP socket
 *   (IN) *si_hdr : TCP-SI header
 * 
 * [Return]
 *   0            : OK.
 *  -EINVAL       : NG.
 *  -EAFNOSUPPORT :
 *  -ENETUNREACH  :
 **/
static int tcpcp_pre_v6connect(struct sock *sk,
			       const struct tcpcp_si_hdr *si_hdr)
{
  struct inet_sock *inet = inet_sk(sk);
  struct ipv6_pinfo *inet6 = inet6_sk(sk);
  struct tcp_sock *tp = tcp_sk(sk);
  struct in6_addr ip_dst;
  __u32 scope_id;
  int addr_type;

  /* plug-in. */
  plugin_dummy();

  /* start. */
  TCPCP_DPRINTK("start. sk=0x%p, si_hdr=0x%p\n",
		sk, si_hdr);

  /* get IPv6 destination address. */
  ipv6_addr_copy(&ip_dst, &(si_hdr->id.ip.v6.ip_dst));
  scope_id = ntohl(si_hdr->id.ip.v6.scope_id);
  /* @@@ for DEBUG @@@ */
  TCPCP_DPRINTK("ip_dst=%04x:%04x:%04x:%04x:%04x:%04x:%04x:%04x\n",
		
		ntohs(ip_dst.s6_addr16[0]),
		ntohs(ip_dst.s6_addr16[1]),
		ntohs(ip_dst.s6_addr16[2]),
		ntohs(ip_dst.s6_addr16[3]),
		ntohs(ip_dst.s6_addr16[4]),
		ntohs(ip_dst.s6_addr16[5]),
		ntohs(ip_dst.s6_addr16[6]),
		ntohs(ip_dst.s6_addr16[7]));
  TCPCP_DPRINTK("si_hdr->id.ip.v6.scope_id=%u\n",
		si_hdr->id.ip.v6.scope_id);
  TCPCP_DPRINTK("scope_id                 =%u\n",
		scope_id);

  /* check IPv6 destination address. */
  if (ipv6_addr_any(&ip_dst) != 0) {
    ip_dst.s6_addr[15] = 0x1;
  }
  /* @@@ for DEBUG @@@ */
  TCPCP_DPRINTK("ip_dst=%04x:%04x:%04x:%04x:%04x:%04x:%04x:%04x\n",
		
		ntohs(ip_dst.s6_addr16[0]),
		ntohs(ip_dst.s6_addr16[1]),
		ntohs(ip_dst.s6_addr16[2]),
		ntohs(ip_dst.s6_addr16[3]),
		ntohs(ip_dst.s6_addr16[4]),
		ntohs(ip_dst.s6_addr16[5]),
		ntohs(ip_dst.s6_addr16[6]),
		ntohs(ip_dst.s6_addr16[7]));

  /* get type of IPv6 destination address. */
  addr_type = ipv6_addr_type(&ip_dst);
  /* @@@ for DEBUG @@@ */
  TCPCP_DPRINTK("addr_type=%u\n", addr_type);

  /* check tupe of IPv6 destination address. */
  if (addr_type & IPV6_ADDR_MULTICAST)
    return -ENETUNREACH;

  /* @@@ for DEBUG @@@ */
  TCPCP_DPRINTK("sk->sk_bound_dev_if=%u\n",
		sk->sk_bound_dev_if);

  if (addr_type & IPV6_ADDR_LINKLOCAL) {
    if (scope_id) {
      /* if interface is set while binding,
       * indices must coincide. */
      if ((sk->sk_bound_dev_if != 0) &&
	  (sk->sk_bound_dev_if != scope_id))
	return -EINVAL;

      sk->sk_bound_dev_if = scope_id;
    }
    /* @@@ for DEBUG @@@ */
    TCPCP_DPRINTK("sk->sk_bound_dev_if=%u\n",
		  sk->sk_bound_dev_if);

    /* connect to link-local address requires an interface. */
    if (sk->sk_bound_dev_if == 0) {
      return -EINVAL;
    }
  }

  /* @@@ for DEBUG @@@ */
  TCPCP_DPRINTK("tp->rx_opt.ts_recent_stamp=%lu\n",
		tp->rx_opt.ts_recent_stamp);
  TCPCP_DPRINTK("inet6->daddr=%04x:%04x:%04x:%04x:"
		"%04x:%04x:%04x:%04x\n",
		ntohs(inet6->daddr.s6_addr16[0]),
		ntohs(inet6->daddr.s6_addr16[1]),
		ntohs(inet6->daddr.s6_addr16[2]),
		ntohs(inet6->daddr.s6_addr16[3]),
		ntohs(inet6->daddr.s6_addr16[4]),
		ntohs(inet6->daddr.s6_addr16[5]),
		ntohs(inet6->daddr.s6_addr16[6]),
		ntohs(inet6->daddr.s6_addr16[7]));

  if ((tp->rx_opt.ts_recent_stamp != 0) &&
      (ipv6_addr_cmp(&(inet6->daddr), &ip_dst) != 0)) {
    /* reset inherited state. */
    tp->rx_opt.ts_recent = 0;
    tp->rx_opt.ts_recent_stamp = 0;
    tp->write_seq = 0;
  }
  /* @@@ for DEBUG @@@ */
  TCPCP_DPRINTK("tp->rx_opt.ts_recent=%u\n",
		tp->rx_opt.ts_recent);
  TCPCP_DPRINTK("tp->rx_opt.ts_recent_stamp=%lu\n",
		tp->rx_opt.ts_recent_stamp);
  TCPCP_DPRINTK("tp->write_seq=%u\n",
		tp->write_seq);

  /* set IPv6 destination address and TCP destination port. */
  ipv6_addr_copy(&(inet6->daddr), &ip_dst);
  inet->dport = si_hdr->id.tcp_dport;
  /* @@@ for DEBUG @@@ */
  TCPCP_DPRINTK("inet6->daddr=%04x:%04x:%04x:%04x:"
		"%04x:%04x:%04x:%04x\n",
		ntohs(inet6->daddr.s6_addr16[0]),
		ntohs(inet6->daddr.s6_addr16[1]),
		ntohs(inet6->daddr.s6_addr16[2]),
		ntohs(inet6->daddr.s6_addr16[3]),
		ntohs(inet6->daddr.s6_addr16[4]),
		ntohs(inet6->daddr.s6_addr16[5]),
		ntohs(inet6->daddr.s6_addr16[6]),
		ntohs(inet6->daddr.s6_addr16[7]));
  TCPCP_DPRINTK("si_hdr->id.tcp_dport=%u\n",
		si_hdr->id.tcp_dport);
  TCPCP_DPRINTK("inet->dport         =%u\n",
		inet->dport);

  /* set ext_header_len. */
  {
    struct inet_connection_sock *icsk = inet_csk(sk);
    icsk->icsk_ext_hdr_len = 0;
    if (inet6->opt != 0) {
      icsk->icsk_ext_hdr_len
	= inet6->opt->opt_flen + inet6->opt->opt_nflen;
      /* @@@ for DEBUG @@@ */
      TCPCP_DPRINTK("inet6->opt->opt_flen =%u\n",
		    inet6->opt->opt_flen);
      TCPCP_DPRINTK("inet6->opt->opt_nflen=%u\n",
		    inet6->opt->opt_nflen);
    }
    /* @@@ for DEBUG @@@ */
    TCPCP_DPRINTK("icsk->icsk_ext_hdr_len=%u\n",
		  icsk->icsk_ext_hdr_len);
  }

  /* initialize mss_clamp. */
  tp->rx_opt.mss_clamp = IPV6_MIN_MTU - sizeof(struct tcphdr)
    - sizeof(struct ipv6hdr);
  /* @@@ for DEBUG @@@ */
  TCPCP_DPRINTK("tp->rx_opt.mss_clamp=%u\n",
		tp->rx_opt.mss_clamp);

  /* set write_seq. */
  tp->write_seq = ntohl(si_hdr->vargen.snd_nxt);
  /* @@@ for DEBUG @@@ */
  TCPCP_DPRINTK("si_hdr->vargen.snd_nxt=%u\n",
		si_hdr->vargen.snd_nxt);
  TCPCP_DPRINTK("tp->write_seq         =%u\n",
		tp->write_seq);

  /* set copied_seq. Might be adjusted during rcv_buf recovery */
  tp->copied_seq = ntohl(si_hdr->vargen.copied_seq);
  TCPCP_DPRINTK("tp->copied_seq        =%u\n",
		tp->copied_seq);

  /* initialize TCP. */
  tcpcp_init_connect(sk, si_hdr);

  /* end. */
  TCPCP_DPRINTK("end.\n");
  return 0;
}
#endif /* defined(CONFIG_IPV6) || defined(CONFIG_IPV6_MODULE) */

/**
 * tcpcp_pre_v4connect - Step 5 : pre-processing of TCP/IPv4 connect.
 *
 * This function is based on tcp_ipv4.c:tcp_v4_connect().
 * Caller must lock socket.
 *
 * [Args]
 *   (IN) *sk     : TCP socket
 *   (IN) *si_hdr : TCP-SI header
 * 
 * [Return]
 *   0 : OK.
 **/
static int tcpcp_pre_v4connect(struct sock *sk,
			       const struct tcpcp_si_hdr *si_hdr)
{
  struct inet_sock *inet = inet_sk(sk);
  struct tcp_sock *tp = tcp_sk(sk);
  u32 daddr;

  /* plug-in. */
  plugin_dummy();

  /* start. */
  TCPCP_DPRINTK("start. sk=0x%p, si_hdr=0x%p\n",
		sk, si_hdr);

  /* get IPv4 destination address. */
  daddr = si_hdr->id.ip.v4.ip_dst.s_addr;
  /* @@@ for DEBUG @@@ */
  TCPCP_DPRINTK("si_hdr->id.ip.v4.ip_dst=0x%x\n",
		si_hdr->id.ip.v4.ip_dst.s_addr);
  TCPCP_DPRINTK("daddr                  =0x%x\n",
		daddr);

  inet->rcv_saddr = inet->saddr;
  /* @@@ for DEBUG @@@ */
  TCPCP_DPRINTK("inet->saddr    =0x%x\n",
		inet->saddr);
  TCPCP_DPRINTK("inet->rcv_saddr=0x%x\n",
		inet->rcv_saddr);

  /* @@@ for DEBUG @@@ */
  TCPCP_DPRINTK("tp->rx_opt.ts_recent_stamp=%lu\n",
		tp->rx_opt.ts_recent_stamp);
  TCPCP_DPRINTK("inet->daddr=0x%x\n",
		inet->daddr);

  if ((tp->rx_opt.ts_recent_stamp != 0) && (inet->daddr != daddr)) {
    /* reset inherited state. */
    tp->rx_opt.ts_recent       = 0;
    tp->rx_opt.ts_recent_stamp = 0;
    tp->write_seq       = 0;
  }
  /* @@@ for DEBUG @@@ */
  TCPCP_DPRINTK("tp->rx_opt.ts_recent=%u\n",
		tp->rx_opt.ts_recent);
  TCPCP_DPRINTK("tp->rx_opt.ts_recent_stamp=%lu\n",
		tp->rx_opt.ts_recent_stamp);
  TCPCP_DPRINTK("tp->write_seq=%u\n",
		tp->write_seq);

  /* set IPv4 destination address and TCP destination port. */
  inet->daddr = daddr;
  inet->dport = si_hdr->id.tcp_dport;
  /* @@@ for DEBUG @@@ */
  TCPCP_DPRINTK("inet->daddr=0x%x\n",
		inet->daddr);
  TCPCP_DPRINTK("si_hdr->id.tcp_dport=%u\n",
		si_hdr->id.tcp_dport);
  TCPCP_DPRINTK("inet->dport         =%u\n",
		inet->dport);

  /* initialize ext_header_len. */
  {
    struct inet_connection_sock *icsk = inet_csk(sk);
    icsk->icsk_ext_hdr_len = 0;
    if (inet->opt) {
      icsk->icsk_ext_hdr_len = inet->opt->optlen;
      /* @@@ for DEBUG @@@ */
      TCPCP_DPRINTK("inet->opt->optlen=%u\n",
		    inet->opt->optlen);
    }
    /* @@@ for DEBUG @@@ */
    TCPCP_DPRINTK("icsk->icsk_ext_hdr_len=%u\n",
		  icsk->icsk_ext_hdr_len);
  }

  /* initialize mss_clamp. */
  tp->rx_opt.mss_clamp = 536;
  /* @@@ for DEBUG @@@ */
  TCPCP_DPRINTK("tp->rx_opt.mss_clamp=%u\n",
		tp->rx_opt.mss_clamp);

  /* set write_seq. */
  tp->write_seq = ntohl(si_hdr->vargen.snd_nxt);
  /* @@@ for DEBUG @@@ */
  TCPCP_DPRINTK("si_hdr->vargen.snd_nxt=%u\n",
		si_hdr->vargen.snd_nxt);
  TCPCP_DPRINTK("tp->write_seq         =%u\n",
		tp->write_seq);

  /* set copied_seq. Might be adjust during recovery of rcv_buf */
  tp->copied_seq = ntohl(si_hdr->vargen.copied_seq);
  TCPCP_DPRINTK("tp->copied_seq        =%u\n",
		tp->copied_seq);

  /* @@@ for DEBUG @@@ */
  TCPCP_DPRINTK("jiffies (1)=%lu\n", jiffies);

  /* set IPv4 socket ID. */
  inet->id = tp->write_seq ^ jiffies;
  /* @@@ for DEBUG @@@ */
  TCPCP_DPRINTK("jiffies (2)=%lu\n", jiffies);
  TCPCP_DPRINTK("tp->write_seq=%u\n",
		tp->write_seq);
  TCPCP_DPRINTK("inet->id=%u\n", inet->id);

  /* initialize TCP. */
  tcpcp_init_connect(sk, si_hdr);

  /* end. */
  TCPCP_DPRINTK("end.\n");
  return 0;
}

/**
 * tcpcp_recover_si_hdr - Step 6 : recover TCP-SI header.
 *
 * This function is based on tcp_input.c:tcp_rcv_synsent_state_process(),
 *                           tcp_input.c:tcp_parse_options().
 * Caller must lock socket.
 *
 * [Args]
 *   (IN) *sk     : TCP socket
 *   (IN) *si_hdr : TCP-SI header
 **/
static void tcpcp_recover_si_hdr(struct sock *sk,
				 const struct tcpcp_si_hdr *si_hdr)
{
	struct tcp_sock *tp = tcp_sk(sk);

	/* plug-in. */
	plugin_dummy();

	/* start. */
	TCPCP_DPRINTK("start. sk=0x%p, si_hdr=0x%p\n",
		sk, si_hdr);
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("sk->sk_sndbuf=%u\n",
		sk->sk_sndbuf);
	TCPCP_DPRINTK("sk->sk_rcvbuf=%u\n",
		sk->sk_rcvbuf);

	/*
	 * it's based on tcp_input.c:tcp_parse_options().
	 */

	/* case TCPOPT_MSS. */
	tp->rx_opt.mss_clamp = ntohs(si_hdr->fixgen.rcv_mss);
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("si_hdr->fixgen.rcv_mss=%d\n",
		si_hdr->fixgen.rcv_mss);
	TCPCP_DPRINTK("tp->rx_opt.mss_clamp         =%d\n",
		tp->rx_opt.mss_clamp);

	/* case TCPOPT_WINDOW. */
	tp->rx_opt.wscale_ok = !!(si_hdr->fixgen.tcp_flags & TCPI_OPT_WSCALE);
	tp->rx_opt.snd_wscale = si_hdr->fixgen.snd_wscale;
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("si_hdr->fixgen.tcp_flags=%u\n",
		si_hdr->fixgen.tcp_flags);
	TCPCP_DPRINTK("tp->rx_opt.wscale_ok=%d\n",
		tp->rx_opt.wscale_ok);
	TCPCP_DPRINTK("si_hdr->fixgen.snd_wscale=%u\n",
		si_hdr->fixgen.snd_wscale);
	TCPCP_DPRINTK("tp->rx_opt.snd_wscale           =%d\n",
		tp->rx_opt.snd_wscale);

	if (tp->rx_opt.snd_wscale > 14)
		tp->rx_opt.snd_wscale = 14;
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("tp->rx_opt.snd_wscale=%d\n",
		tp->rx_opt.snd_wscale);

	/* case TCPOPT_TIMESTAMP. */
	tp->rx_opt.tstamp_ok = !!(si_hdr->fixgen.tcp_flags & TCPI_OPT_TIMESTAMPS);

	/* case TCPOPT_SACK_PERM. */
	tp->rx_opt.sack_ok = !!(si_hdr->fixgen.tcp_flags & TCPI_OPT_SACK);

	/*
	 * it's based on tcp_input.c:tcp_rcv_synsent_state_process().
	 */

	/* set ECN flags. */
	tp->ecn_flags
	    = si_hdr->fixgen.tcp_flags & TCPI_OPT_ECN ? TCP_ECN_OK : 0;
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("si_hdr->fixgen.tcp_flags=%u\n",
		si_hdr->fixgen.tcp_flags);
	TCPCP_DPRINTK("tp->rx_opt.tstamp_ok           =%u\n",
		tp->rx_opt.tstamp_ok);
	TCPCP_DPRINTK("tp->rx_opt.sack_ok             =%u\n",
		tp->rx_opt.sack_ok);
	TCPCP_DPRINTK("tp->ecn_flags           =%u\n",
		tp->ecn_flags);

	/* set rcv_nxt and initialize rcv_wup, snd_wl1. */
	tp->rcv_nxt = ntohl(si_hdr->vargen.rcv_nxt);
	tp->rcv_wup = tp->rcv_nxt;
	tp->snd_wl1 = tp->rcv_nxt;
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("si_hdr->vargen.rcv_nxt=%u\n",
		si_hdr->vargen.rcv_nxt);
	TCPCP_DPRINTK("tp->rcv_nxt           =%u\n",
		tp->rcv_nxt);
	TCPCP_DPRINTK("tp->rcv_wup           =%u\n",
		tp->rcv_wup);
	TCPCP_DPRINTK("tp->snd_wl1           =%u\n",
		tp->snd_wl1);

	/* set snd_wnd max_window. */
	tp->snd_wnd = ntohl(si_hdr->vargen.snd_wnd);
	tp->max_window = tp->snd_wnd;
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("si_hdr->vargen.snd_wnd=%u\n",
		si_hdr->vargen.snd_wnd);
	TCPCP_DPRINTK("tp->snd_wnd           =%u\n",
		tp->snd_wnd);
	TCPCP_DPRINTK("tp->max_window        =%u\n",
		tp->max_window);

	tp->tcp_header_len = sizeof(struct tcphdr);
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("tp->tcp_header_len=%d\n",
		tp->tcp_header_len);

	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("tp->rx_opt.tstamp_ok=%u\n",
          tp->rx_opt.tstamp_ok);

	if (tp->rx_opt.tstamp_ok) {
		tp->tcp_header_len += TCPOLEN_TSTAMP_ALIGNED;
		/* @@@ for DEBUG @@@ */
		TCPCP_DPRINTK("tp->tcp_header_len    =%d\n",
			tp->tcp_header_len);

		/* @@@ for DEBUG @@@ */
		TCPCP_DPRINTK("si_hdr->vargen.ts_recent=%u\n",
			      si_hdr->vargen.ts_recent);

		/* it's based on tcp_input.c:tcp_store_ts_recent(). */
		if (si_hdr->vargen.ts_recent) {
		  /* Init timestamp generation */
		  tp->rx_opt.ts_recent= ntohl(si_hdr->vargen.ts_recent);
		  tp->rx_opt.ts_recent_stamp = get_seconds();
		  /* Fake received timestamp */
		  /* tp->rx_opt.rcv_tsecr = tp->rx_opt.ts_recent; */
		  /* Fake sent timestamp */
		  /* tp->rx_opt.rcv_tsval = tcp_time_stamp; */
		}
		/* @@@ for DEBUG @@@ */
		TCPCP_DPRINTK("tp->rx_opt.ts_recent=%u\n",
			      tp->rx_opt.ts_recent);
		TCPCP_DPRINTK("tp->rx_opt.ts_recent_stamp=%lu\n",
			      tp->rx_opt.ts_recent_stamp);
		TCPCP_DPRINTK("get_seconds()      =%lu\n",
			get_seconds());
	}

	/* initialize copied_seq. */
	tp->copied_seq = tp->rcv_nxt;
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("tp->copied_seq=%u\n",
		      tp->copied_seq);
	TCPCP_DPRINTK("tp->rcv_nxt   =%u\n",
		      tp->rcv_nxt);

	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("tcp_time_stamp (1)   =%u\n",
		      tcp_time_stamp);

	/* prevent spurious tcp_cwnd_restart() on first data packet. */
	tp->lsndtime = tcp_time_stamp;
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("tcp_time_stamp (2)   =%u\n",
		      tcp_time_stamp);
	TCPCP_DPRINTK("tp->lsndtime         =%u\n",
		      tp->lsndtime);

	/* .... */
	tcpcp_init_buffer_space(sk);
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("tcp_time_stamp (3)   =%u\n",
		      tcp_time_stamp);
	TCPCP_DPRINTK("si_hdr->vargen.ts_gen=%u\n",
		      ntohl(si_hdr->vargen.ts_gen));

	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("sk->sk_sndbuf=%u\n",
		sk->sk_sndbuf);
	TCPCP_DPRINTK("sk->sk_rcvbuf=%u\n",
		sk->sk_rcvbuf);

	/* end. */
	TCPCP_DPRINTK("end.\n");
	return;
}

/**
 * tcpcp_recover_sndbuf - Step 7 : recover write_queue.
 *
 * This function is based on tcp.c:tcp_sendmsg(),
 *                           tcp.c:skb_entail(),
 *                           TCPCP-9:tcpcp_buffers(),
 *                           TCPCP-9:tcpcp_send_buf().
 * Caller must lock socket.
 *
 * [Args]
 *   (IN) *sk      : TCP socket
 *   (IN) *snd_buf : ....
 *
 * [Return]
 *   0      : OK.
 *  -ENOMEM : NG.
 *  -EINVAL : NG.
 **/
static int tcpcp_recover_sndbuf(struct sock *sk,
				const struct tcpcp_si_skb_buf *snd_buf)
{
	struct tcp_sock *tp = tcp_sk(sk);
	struct tcpcp_si_skb *si_skb;
	unsigned int cnt;

#ifdef CONFIG_TCPCP_DEBUG
	/* @@@ for printing payload. (for DEBUG) @@@ */
	unsigned int cnt1, cnt2 = 0;
	char *pload_data;
#endif

	/* plug-in. */
	plugin_dummy();

	/* start. */
	TCPCP_DPRINTK("start. sk=0x%p, snd_buf=0x%p\n",
		sk, snd_buf);

	/* purge write_queue. */
	tcp_write_queue_purge(sk);

	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("snd_buf->buf_len=%u\n",
		snd_buf->buf_len);
	TCPCP_DPRINTK("snd_buf->skb_cnt=%u\n",
		snd_buf->skb_cnt);

	/* check send buffer. */
	if ((snd_buf->buf_len == 0) || (snd_buf->skb_cnt == 0)) {
		TCPCP_DPRINTK("end.\n");
		return 0;
	}

	si_skb = (struct tcpcp_si_skb *)snd_buf->si_skb;

	for (cnt = 0; cnt < snd_buf->skb_cnt; cnt++) {
		struct sk_buff *skb;
		unsigned int data_len;
		uint32_t seq;

		/* @@@ for DEBUG @@@ */
		TCPCP_DPRINTK("cnt=%u\n", cnt);

		/*
		 * it's based on tcp.c:tcp_sendmsg().
		 */

		/* @@@ for DEBUG @@@ */
		TCPCP_DPRINTK("before sk_stream_memory_free.\n");

		/* check free of write_queue. */
		if (!sk_stream_memory_free(sk)) {
			TCPCP_DPRINTK("[Error] write_queue is not"
				" free.\n");
			return -ENOMEM;
		}

		/* get data_len. */
		data_len = ntohs(si_skb->length);
		/* @@@ for DEBUG @@@ */
		TCPCP_DPRINTK("si_skb->length=%u\n",
			si_skb->length);
		TCPCP_DPRINTK("data_len      =%u\n",
			data_len);

		/* allocate socket buffer. (TCP header and data of packet) */
		skb = sk_stream_alloc_skb(sk, data_len, sk->sk_allocation);
		if (!skb) {
			TCPCP_DPRINTK("[Error] can't allocate skb.\n");
			return -ENOMEM;
		}
		/* @@@ for DEBUG @@@ */
		TCPCP_DPRINTK("after sk_stream_alloc_pskb.\n");
 		TCPCP_DPRINTK("skb->head=0x%p\n",
			skb->head);
 		TCPCP_DPRINTK("skb->data=0x%p\n",
			skb->data);
 		TCPCP_DPRINTK("skb->tail=0x%p\n",
			skb->tail);
		TCPCP_DPRINTK("skb->len=%u\n", skb->len);

		/*
		 * it's based on tcp.c:skb_entail().
		 */

		skb->csum = 0;
		seq = ntohl(si_skb->seq);
		TCP_SKB_CB(skb)->seq = seq;
		TCP_SKB_CB(skb)->end_seq = seq + data_len
		  + ((si_skb->flags & TCPCB_FLAG_SYN) ? 1 : 0)
		  + ((si_skb->flags & TCPCB_FLAG_FIN) ? 1 : 0);
		TCP_SKB_CB(skb)->flags = si_skb->flags;
		TCP_SKB_CB(skb)->sacked = si_skb->sacked;
		skb_shinfo(skb)->gso_segs = ntohs(si_skb->gso_segs);
		skb_shinfo(skb)->gso_size = ntohs(si_skb->gso_size);
		/* @@@ for DEBUG @@@ */
		TCPCP_DPRINTK("skb->csum=%u\n", skb->csum);
		TCPCP_DPRINTK("si_skb->seq             =%u\n",
                        si_skb->seq);
		TCPCP_DPRINTK("TCP_SKB_CB(skb)->seq    =%u\n",
			TCP_SKB_CB(skb)->seq);
		TCPCP_DPRINTK("TCP_SKB_CB(skb)->end_seq=%u\n",
			TCP_SKB_CB(skb)->end_seq);
		TCPCP_DPRINTK("si_skb->flags         =%u\n",
			si_skb->flags);
		TCPCP_DPRINTK("TCP_SKB_CB(skb)->flags=%u\n",
			TCP_SKB_CB(skb)->flags);
		TCPCP_DPRINTK("si_skb->sacked         =%u\n",
			si_skb->sacked);
		TCPCP_DPRINTK("TCP_SKB_CB(skb)->sacked=%u\n",
			TCP_SKB_CB(skb)->sacked);
		TCPCP_DPRINTK("si_skb->gso_segs         =%u\n",
			si_skb->gso_segs);
		TCPCP_DPRINTK("skb_shinfo(skb)->gso_segs=%u\n",
			skb_shinfo(skb)->gso_segs);
		TCPCP_DPRINTK("si_skb->gso_size         =%u\n",
			si_skb->gso_size);
		TCPCP_DPRINTK("skb_shinfo(skb)->gso_size=%u\n",
			skb_shinfo(skb)->gso_size);
		TCPCP_DPRINTK("TCP_SKB_CB(skb)->when=%u\n",
			      TCP_SKB_CB(skb)->when);

#ifdef CONFIG_TCPCP_DEBUG
		/* @@@ print queue. (for DEBUG) @@@ */
		skb_queue_check(&sk->sk_write_queue);
#endif
    tcp_add_write_queue_tail(sk, skb);
    sk->sk_wmem_queued += skb->truesize;
    sk_mem_charge(sk, skb->truesize);
		/* @@@ for DEBUG @@@ */
		TCPCP_DPRINTK("sk->sk_wmem_queued=%d\n",
			sk->sk_wmem_queued);

#ifdef CONFIG_TCPCP_DEBUG
		/* @@@ print queue. (for DEBUG) @@@ */
		skb_queue_check(&sk->sk_write_queue);
#endif

		/*
		 * it's based on tcp.c:skb_add_data().
		 */

#ifdef CONFIG_TCPCP_DEBUG
		/* @@@ for printing payload. (for DEBUG) @@@ */
		pload_data = skb->tail;
#endif

		/* copy data of packet. */
		skb->csum = csum_partial_copy_nocheck(si_skb->data,
						      skb_put(skb, data_len),
						      data_len, 0);
		/* @@@ for DEBUG @@@ */
		TCPCP_DPRINTK("after skb_put.\n");
		TCPCP_DPRINTK("skb->head=0x%p\n",
			skb->head);
		TCPCP_DPRINTK("skb->data=0x%p\n",
			skb->data);
		TCPCP_DPRINTK("skb->tail=0x%p\n",
			skb->tail);
		TCPCP_DPRINTK("skb->len=%u\n", skb->len);
		TCPCP_DPRINTK("skb->csum=%u\n", skb->csum);

#ifdef CONFIG_TCPCP_DEBUG
		/* @@@ print payload. (for DEBUG) @@@ */
		TCPCP_DPRINTK("si_skb->data=\n");
		for (cnt1 = 0; cnt1 < data_len; cnt1 += 4) {
			cnt2++;
			TCPCP_DPRINTK("  %02d: %02X%02X%02X%02X\n", cnt2,
				      si_skb->data[cnt1],
				      si_skb->data[cnt1 + 1],
				      si_skb->data[cnt1 + 2],
				      si_skb->data[cnt1 + 3]);
		}
		cnt2 = 0;

		TCPCP_DPRINTK("pload_data=\n");
		for (cnt1 = 0; cnt1 < data_len; cnt1 += 4) {
			cnt2++;
			TCPCP_DPRINTK("  %02d: %02X%02X%02X%02X\n", cnt2,
				      pload_data[cnt1],
				      pload_data[cnt1 + 1],
				      pload_data[cnt1 + 2],
				      pload_data[cnt1 + 3]);
		}
		cnt2 = 0;
#endif

		/* @@@ for DEBUG @@@ */
		TCPCP_DPRINTK("tp->snd_nxt   =%u\n",
			      tp->snd_nxt);
		TCPCP_DPRINTK("seq           =%u\n",
			      seq);
		TCPCP_DPRINTK("seq+data_len-1=%u\n",
			      seq + data_len - 1);
		
		/* .... */
		if (between(tp->snd_nxt, seq, seq + data_len - 1)) {
			sk->sk_send_head = skb;

			/* @@@ for DEBUG @@@ */
			TCPCP_DPRINTK("skb             =0x%p\n",
				      skb);
			TCPCP_DPRINTK("sk->sk_send_head=0x%p\n",
				      sk->sk_send_head);
		}

		/* @@@ for DEBUG @@@ */
		TCPCP_DPRINTK("skb_shinfo(skb)->gso_segs=%u\n",
			      skb_shinfo(skb)->gso_segs);
		TCPCP_DPRINTK("tp->packets_out=%u\n",
			      tp->packets_out);

		if (!sk->sk_send_head) {
			/* @@@ for DEBUG @@@ */
			TCPCP_DPRINTK("sk->sk_send_head=0x%p\n",
				sk->sk_send_head);

			tp->packets_out += tcp_skb_pcount(skb);
		}
		/* @@@ for DEBUG @@@ */
		TCPCP_DPRINTK("tp->packets_out=%u\n",
			tp->packets_out);

		/* move write_seq and snd_una. */
		tp->write_seq = TCP_SKB_CB(skb)->end_seq;
		/* @@@ for DEBUG @@@ */
		TCPCP_DPRINTK("tp->write_seq=%u\n",
			tp->write_seq);

		if (cnt == 0)
			tp->snd_una = seq;
		/* @@@ for DEBUG  @@@ */
		TCPCP_DPRINTK("cnt=%u\n",
			cnt);
		TCPCP_DPRINTK("seq        =%u\n",
			seq);
		TCPCP_DPRINTK("tp->snd_una=%u\n",
			tp->snd_una);

		/* @@@ for DEBUG @@@ */
		TCPCP_DPRINTK("skb->truesize     =%u\n",
			skb->truesize);
		TCPCP_DPRINTK("sk->sk_sndbuf     =%u\n",
			sk->sk_sndbuf);
		TCPCP_DPRINTK("sk->sk_wmem_queued=%u\n",
			sk->sk_wmem_queued);

		si_skb = (struct tcpcp_si_skb *)
			 (si_skb->data + ((data_len + 3) & ~3));
	}

	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("tp->snd_nxt  =%u\n",
		tp->snd_nxt);
	TCPCP_DPRINTK("tp->snd_una  =%u\n",
		tp->snd_una);
	TCPCP_DPRINTK("tp->write_seq=%u\n",
		tp->write_seq);

	/* check snd_nxt, snd_una, write_seq. */
	if (!between(tp->snd_nxt, tp->snd_una, tp->write_seq))
		return -EINVAL;

	/* end. */
	TCPCP_DPRINTK("end.\n");
	return 0;
}

/**
 * do_set_si - set TCP-SI for TCP socket.
 *
 * This function is based on TCPCP-9:do_setici().
 * Caller must lock socket.
 *
 * [Args]
 *   (IN) *sk     : TCP socket
 *   (IN) *si_buf : TCP-SI buffer
 *
 * [Return]
 *   0             : OK.
 *  -ENOPROTOOPT   : NG. (tcpcp_check_compatibility)
 *  -EINVAL        : NG. (pre_v6bind/pre_v4bind/pre_v6connect/pre_v4connect)
 *  -EACCES        : NG. (pre_v6bind/pre_v4bind)
 *  -EADDRNOTAVAIL : NG. (pre_v6bind/pre_v4bind)
 *  -ENETUNREACH   : NG. (pre_v6connect)
 *  -EBADFD        : NG.
 **/
static int do_set_si(struct sock *sk, struct tcpcp_si_buf *si_buf)
{
	struct tcpcp_si_hdr *si_hdr = &(si_buf->si_hdr);
	struct tcpcp_si_skb_buf *snd_buf = &(si_buf->snd_buf);
	struct tcpcp_si_skb_buf *rcv_buf = &(si_buf->rcv_buf);
	struct tcp_sock *tp = tcp_sk(sk);
	int error;

	/* plug-in. */
	plugin_dummy();

	/* start. */
	TCPCP_DPRINTK("start. si_buf=0x%p\n", si_buf);

	/* Step 1 : check compatibility. */
	error = tcpcp_check_compatibility(si_hdr);
	if (error != 0) {
		/* only -ENOPROTOOPT. */
		TCPCP_DPRINTK("[Error] TCP-SI is not compatible on this"
			" system. error=%d\n", error);
		return error;
	}

	/* Step 1.5 : check socket family. */
	switch (si_hdr->id.ip_version) {
	case 4:
		if (sk->sk_family != PF_INET) {
			TCPCP_DPRINTK("[Error] ip_version is IPv4, but"
				" sk_family is not PF_INET.\n");
			return -EBADFD;
		}
		break;
	case 6:
		if (sk->sk_family != PF_INET6) {
			TCPCP_DPRINTK("[Error] ip_version is IPv6, but"
				" sk_family is not PF_IENT6.\n");
			return -EBADFD;
		}
		break;
	default:
		TCPCP_DPRINTK("[Error] ip_version is invalid.\n");
		return -EINVAL;
	}

	/* Step 2 : set socket options. (before bind) */
 	tcpcp_setopt_before_bind(sk, si_hdr);

	/* Step 3 : pre-bind. */
#if defined(CONFIG_IPV6) || defined(CONFIG_IPV6_MODULE)
	if (sk->sk_family == PF_INET6)
		error = tcpcp_pre_v6bind(sk, si_hdr);
	else
#endif
		error = tcpcp_pre_v4bind(sk, si_hdr);
	if (error != 0) {
		TCPCP_DPRINTK("[Error] can't pre_bind. error=%d\n",
			error);
		return error;
	}

	/* Step 4 : set socket options. (after bind) */
 	tcpcp_setopt_after_bind(sk, si_hdr);

	/* Step 5 : pre-connect. */
#if defined(CONFIG_IPV6) || defined(CONFIG_IPV6_MODULE)
	if (sk->sk_family == PF_INET6)
		error = tcpcp_pre_v6connect(sk, si_hdr);
	else
#endif
		error = tcpcp_pre_v4connect(sk, si_hdr);
	if (error != 0) {
		TCPCP_DPRINTK("[Error] can't pre_connect. error=%d\n",
			error);
		goto err;
	}

	/* Step 6 : recover TCP-SI header. */
	tcpcp_recover_si_hdr(sk, si_hdr);

	/* Step 7 : recover write_queue. */
	error = tcpcp_recover_sndbuf(sk, snd_buf);
	if (error != 0) {
		TCPCP_DPRINTK("[Error] can't recover write_queue."
			" error=%d\n", error);
		goto err;
	}

	/* Step 8 : recover receive_queue. */
	error = tcpcp_recover_rcvbuf(sk, rcv_buf);
	if (error != 0) {
		TCPCP_DPRINTK("[Error] can't recover receive_queue."
			" error=%d\n", error);
		goto err;
	}

	/* Step 9 : check socket error. */
	error = sock_error(sk);
	if (error != 0) {
		TCPCP_DPRINTK("[Error] sock_error failure. error=%d\n",
			error);
		error = -EBADFD;
		goto err;
	}

	/* Step 10 */
	tp->set_flag = TCPCP_SET_FLAG_ON;
	tcpcp_change_xmit(sk, TCPCP_FLAG_STOP);
	/* @@@ for DEBUG @@@*/
	TCPCP_DPRINTK("tp->set_flag=%d\n", tp->set_flag);

	/* end. */
	TCPCP_DPRINTK("end.\n");
	return 0;

err:
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("sk->sk_flags=%lu\n", sk->sk_flags);

	/* in case of an error, don't make the poor socket linger ... */
	sock_set_flag(sk, SOCK_LINGER);
	sk->sk_lingertime = 0;
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("sk->sk_flags=%lu\n", sk->sk_flags);
	TCPCP_DPRINTK("sk->sk_lingertime=%lu\n",
		sk->sk_lingertime);

	/* end. */
	TCPCP_DPRINTK("end. error=%d\n", error);
	return error;
}

/**
 * tcpcp_close_socket
 *   - close TCP sockets. (for error operation of do_tcpcp_set_si)
 *
 * [Args]
 *   (IN) *f_entry : first FDR Info. Entry
 **/
static void tcpcp_close_socket(struct tcpcp_com_entry *f_entry)
{
	struct fdr *fdrs = NULL;
	unsigned int fdr_cnt;
	unsigned int cnt;
	int fd;

	/* plug-in. */
	plugin_dummy();

	/* start. */
	TCPCP_DPRINTK("start. f_entry=0x%p\n", f_entry);

#ifdef CONFIG_TCPCP_DEBUG
	/* @@@ .... (for DEBUG) @@@ */
	check_fd();
#endif

	while (f_entry != NULL) {
		/* @@@ for DEBUG @@@ */
		TCPCP_DPRINTK("f_entry=0x%p\n", f_entry);

		/* calculate number of FDR Info. */
		fdr_cnt = f_entry->data_len / sizeof(struct fdr);
		/* @@@ for DEBUG @@@ */
		TCPCP_DPRINTK("f_entry->data_len =%u\n",
			f_entry->data_len);
		TCPCP_DPRINTK("fdr_cnt           =%u\n",
			fdr_cnt);

		fdrs = (struct fdr *)(f_entry->data);
		/* @@@ for DEBUG @@@ */
		TCPCP_DPRINTK("fdrs=0x%p\n", fdrs);

		for (cnt = 0; cnt < fdr_cnt; cnt++) {
			/* get new FD. */
			fd = fdrs->result;
			if (fd >= 0) {
				/* @@@ for DEBUG @@@ */
				TCPCP_DPRINTK("fd=%d\n",
					fd);

				/* close TCP socket. */
				do_close(fd);
			}
			fdrs++;
			/* @@@ for DEBUG @@@ */
			TCPCP_DPRINTK("fdrs=0x%p\n",
				fdrs);
		}

		f_entry = f_entry->next;
		/* @@@ for DEBUG @@@ */
		TCPCP_DPRINTK("f_entry=0x%p\n",
			f_entry);
	}

#ifdef CONFIG_TCPCP_DEBUG
	/* @@@ .... (for DEBUG) @@@ */
	check_fd();
#endif

	/* end. */
	TCPCP_DPRINTK("end.\n");
	return;
}

/**
 * do_tcpcp_set_si - set TCP-SIs for TCP sockets.
 *
 * [Args]
 *   (IN) *sk      : TCP socket
 *   (IN) *r_entry : Request Entry
 *   (IN) *user_si : TCP-SI data from user (const)
 *
 * [Return]
 *   0            : OK.
 *  -EINVAL       : NG. (sys_socket/...)
 *  -ENOMEM       : NG.
 *  -EFAULT       : NG.
 *  -EAFNOSUPPORT : NG. (sys_socket)
 *  -ENFILE       : NG. (sys_socket)
 *   other        : NG. (sys_socket)
 **/
static int do_tcpcp_set_si(struct sock *sk,
			  const struct tcpcp_total_si_hdr *user_si,
			  struct tcpcp_req_entry *r_entry)
{
	struct task_struct *tsk = current;
	struct files_struct *files = tsk->files;
	struct file *filp = NULL;
	struct sock *sk_i = NULL;
	struct tcpcp_si_buf si_buf;
	struct tcpcp_com_entry *fdr_entry;
	unsigned int total_si_hdr_len;
	unsigned int total_ucopy;
	int fd = 0;
	int error = 0;
	int si_cnt = 0;
	unsigned int fd_all = 0;
	unsigned int fd_ng = 0;
	int ret = 0;
	int new_fd = 0;
	int old_fd = 0;
	char *r_ptr;
	struct fdtable *fdt;
#ifdef CONFIG_TCPCP_DEBUG
	/* @@@ for DEBUG @@@ */
	unsigned int loop_cnt = 0;
#endif

	/* plug-in. */
	plugin_dummy();

	/* start. */
	TCPCP_DPRINTK("start. sk=0x%p, r_entry=0x%p, user_si=0x%p\n",
		sk, r_entry, user_si);

	/* set read-point. */
	r_ptr = (char *)user_si;
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("r_ptr=0x%p\n", r_ptr);

	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("user_si->major=%u\n",
		user_si->major);

	/* check TCP-SIs version. */
	if (user_si->major != TCPCP_CURRENT_SI_MAJOR)
		return -EINVAL;

	/* get count of TCP-SIs. */
	si_cnt = ntohs(user_si->si_cnt);
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("user_si->si_cnt=%u\n",
		user_si->si_cnt);
	TCPCP_DPRINTK("si_cnt         =%u\n",
		si_cnt);

	/* check si_cnt for ONE processing. */
	if (!(r_entry->flag & TCPCP_FLAG_ALL)) {
		if (si_cnt != 1) {
			TCPCP_DPRINTK("In spite of ONE processing," \
					" TCP-SIs for ALL is specified.\n");
			return -EINVAL;
		}
	}

	/* move read-point. */
	total_si_hdr_len = sizeof(struct tcpcp_total_si_hdr);
	r_ptr += total_si_hdr_len;
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("total_si_hdr_len=%u\n",
		total_si_hdr_len);
	TCPCP_DPRINTK("r_ptr=0x%p\n", r_ptr);

	/* get length of TCP-SI. (exclude TCP-SI header) */
	total_ucopy = ntohl(user_si->total_si_length) * 4 - total_si_hdr_len;
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("user_si->total_si_length=%u\n",
		user_si->total_si_length);
	TCPCP_DPRINTK("total_ucopy             =%u\n",
		total_ucopy);
	fdt = files_fdtable(files);
	for (fd = 0; fd < fdt->max_fds; fd++) {
		/* @@@ for DEBUG @@@ */
		TCPCP_DPRINTK("fd=%u\n", fd);

		sk_i = tcpcp_get_sk_i(sk, fd, r_entry);
		if (sk_i == NULL)
			continue;
		/* @@@ for DEBUG @@@ */
		TCPCP_DPRINTK("nothing continue after tcpcp_get_sk_i.\n");

		if (r_entry->flag & TCPCP_FLAG_ALL)
			/* close FD. (only "ALL") */
			/* lock_sock and release_sock are unnecessary.
			 * because spin_lock to files is in do_close. */
			do_close(fd);

		/*
		 *  if one, fd value is set in the Request entry
		 *  by tcpcp_get_sk_i function.
		 */
	}

	/* make TCP-SI buffer for setting TCP-SIs to kernel. */
	error = tcpcp_alloc_si_buf(&si_buf);
	if (error != 0) {
		/* only -ENOMEM. */
		TCPCP_DPRINTK("[Error] can't allocate si_buf."
			" error=%d\n", error);
		return error;
	}

	if (r_entry->flag & TCPCP_FLAG_ALL) {
		/* add FDR Info. Entry. (only "ALL")*/
		error = tcpcp_add_com_entry(r_entry->flag, &(r_entry->f_entry));
		if (error != 0)
			return -ENOMEM;
	}
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("r_entry->f_entry=0x%p\n",
		r_entry->f_entry);

	/* initialize now FDR Info. Entry. */
	fdr_entry = r_entry->f_entry;

	while ((total_ucopy != 0) || (si_cnt != 0)) {
		unsigned int si_len;

#ifdef CONFIG_TCPCP_DEBUG
		/* @@@ for DEBUG @@@ */
		loop_cnt++;
		TCPCP_DPRINTK("loop_cnt=%u\n", loop_cnt);
#endif

		/* copy TCP-SI from userland to Kernel. */
		error = tcpcp_copy_si_from_user((struct tcpcp_si_hdr *)r_ptr,
						&si_buf, total_ucopy);
		if (error != 0) {
			TCPCP_DPRINTK("[Error] can't copy TCP-SIs from"
				" user. error=%d\n", error);
			goto err;
		}

		si_cnt--;
		/* @@@ for DEBUG @@@ */
		TCPCP_DPRINTK("si_cnt=%u\n", si_cnt);

		/* move read-point. */
		si_len = ntohl(si_buf.si_hdr.si_length) * 4;
		r_ptr += si_len; 
		total_ucopy -= si_len;
		/* @@@ for DEBUG @@@ */
		TCPCP_DPRINTK("si_len=%u\n", si_len);
		TCPCP_DPRINTK("r_ptr=0x%p\n", r_ptr);
		TCPCP_DPRINTK("total_ucopy=%u\n",
			total_ucopy);

		if (r_entry->flag & TCPCP_FLAG_ALL) {
			/* create socket. (only "ALL")*/
#if defined(CONFIG_IPV6) || defined(CONFIG_IPV6_MODULE)
			/* @@@ for DEBUG @@@ */
			TCPCP_DPRINTK("si_buf.si_hdr.id.ip_version=%u\n",
				si_buf.si_hdr.id.ip_version);

			if (si_buf.si_hdr.id.ip_version == 6)
				new_fd = tcpcp_sys_socket(PF_INET6, SOCK_STREAM,
						IPPROTO_TCP);
			else
#endif
				new_fd = tcpcp_sys_socket(PF_INET, SOCK_STREAM,
						IPPROTO_TCP);
			/* @@@ for DEBUG @@@ */
			TCPCP_DPRINTK("new_fd=%d\n",
				new_fd);

			if (new_fd < 0) {
				/* set return value */
				error = new_fd;
				goto err;
			}

			/* get SOCKET_I() from new FD. */
			filp = fcheck_files(files, new_fd);
			sk_i = SOCKET_I(filp->f_dentry->d_inode)->sk;
			if (sk_i == NULL)
				continue;

			/* @@@ for DEBUG @@@ */
			TCPCP_DPRINTK("nothing continue after"
				" SOCKET_I\n");
		} else 
			sk_i = sk;
		/* @@@ for DEBUG @@@ */
		TCPCP_DPRINTK("sk  =0x%p.\n",sk);
		TCPCP_DPRINTK("sk_i=0x%p.\n",sk_i);

		/* set TCP-SI. */
		lock_sock(sk_i);
		ret = do_set_si(sk_i, &si_buf);
		release_sock(sk_i);
		/* @@@ for DEBUG @@@ */
		TCPCP_DPRINTK("after do_set_si. ret=%d\n",
			ret);

		fd_all++;
		/* @@@ for DEBUG @@@ */
		TCPCP_DPRINTK("fd_all=%u\n", fd_all);

		if (!(r_entry->flag & TCPCP_FLAG_ALL)) {
			/* @@@ for DEBUG @@@ */
			TCPCP_DPRINTK("Break for loop.\n");

			if (ret != 0) {
				error = ret;
				goto err;
			}
			break;
		}

		/* Only ALL, the following processings are performed. */
		if (ret != 0) {
			/* In case of NG (count fd_ng and close new_fd.) */
			fd_ng++;
			error = do_close(new_fd);
		} else
			/* In case of OK */
			ret = new_fd;
		/* @@@ for DEBUG @@@ */
		TCPCP_DPRINTK("fd_ng=%u\n", fd_ng);
		TCPCP_DPRINTK("ret=%d\n", ret);

		/* get old FD. */
		old_fd = ntohl(si_buf.si_hdr.fd);
		/* @@@ for DEBUG @@@ */
		TCPCP_DPRINTK("si_buf.si_hdr.fd=%d\n",
			si_buf.si_hdr.fd);
		TCPCP_DPRINTK("old_fd          =%d\n",
			old_fd);

		/* make FDR Info. (only "ALL") */
		error = tcpcp_make_res(old_fd, ret, &fdr_entry);
		if (error != 0) {
			/* only -ENOMEM. */
			TCPCP_DPRINTK("[Error] can't make FDR Info."
				" error=%d\n", error);
			/* Caller delete Request Entry with common entries. */
			goto err;
		}

		/* clear TCP-SI buffer (set si_buf all zero) */ 
		tcpcp_clear_si_buf(&si_buf);
	}

	/* free TCP-SI buffer. */
	tcpcp_free_si_buf(&si_buf);

	/* set Processing Result Info. */
	r_entry->res.fd_all = fd_all;
	r_entry->res.fd_ng = fd_ng;
	r_entry->res.buf_size = 0;
	r_entry->res.fdrs_size = sizeof(struct fdr) * fd_all;
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("r_entry->res.fd_all   =%u\n",
		r_entry->res.fd_all);
	TCPCP_DPRINTK("r_entry->res.fd_ng    =%u\n",
		r_entry->res.fd_ng);
	TCPCP_DPRINTK("r_entry->res.buf_size =%u\n",
		r_entry->res.buf_size);
	TCPCP_DPRINTK("r_entry->res.fdrs_size=%u\n",
		r_entry->res.fdrs_size);

	/* end. */
	TCPCP_DPRINTK("end.\n");
	return 0;

err:
	/* Only in the case of ALL, tcpcp_close_socket is performed. */
	if (r_entry->flag & TCPCP_FLAG_ALL)
		/* close socket made in this function */
		tcpcp_close_socket(r_entry->f_entry);

	/* free memory for send/recv buffer */
	tcpcp_free_si_buf(&si_buf);

	/* end. */
	TCPCP_DPRINTK("end. error=%d\n", error);
	return error;
}


/* ----- start TCP sockets ------------------------------------------------- */

#if defined(CONFIG_IPV6) || defined(CONFIG_IPV6_MODULE)
/**
 * tcpcp_v6bind - Step A : TCP/IPv6 bind.
 *
 * This function is based on af_inet6.c:inet6_bind().
 * Caller must lock socket.
 * 
 * [Args]
 *   (IN) *sk : TCP socket
 * 
 * [Return]
 *   0             : OK.
 *  -ENODEV        : NG.
 *  -EADDRNOTAVAIL : NG.
 *  -EADDRINUSE    : NG.
 **/
static int tcpcp_v6bind(struct sock *sk)
{
	struct inet_sock *inet = inet_sk(sk);
	struct ipv6_pinfo *inet6 = inet6_sk(sk);
	int addr_type;
	unsigned short snum;
	int err_val;

	/* plug-in. */
	plugin_dummy();

	/* start. */
	TCPCP_DPRINTK("start. sk=0x%p\n", sk);

	/* get type of IPv6 source address. */
	addr_type = ipv6_addr_type(&(inet6->saddr));
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("addr_type=%u\n", addr_type);

	/* check type of IPv6 source address. */
	if ((addr_type == IPV6_ADDR_ANY) || (addr_type & IPV6_ADDR_MULTICAST)) {
		/* TCPCP not support IPV6_ADDR_ANY and multicast. */
		TCPCP_DPRINTK("[Error] type of saddr is ANY or"
			" multicast.\n");
		return -EADDRNOTAVAIL;
	}

	if (addr_type != IPV6_ADDR_MAPPED) {
		struct net_device *dev = NULL;

		if (addr_type & IPV6_ADDR_LINKLOCAL) {
			dev = dev_get_by_index(&init_net, sk->sk_bound_dev_if);
			/* @@@ for DEBUG @@@ */
			TCPCP_DPRINTK("sk->sk_bound_dev_if=%u\n",
				sk->sk_bound_dev_if);

			if (!dev) {
				TCPCP_DPRINTK("[Error] can't get"
					" device.\n");
				return -ENODEV;
			}
		}

		if (ipv6_chk_addr(&init_net, &(inet6->saddr), dev, 0) == 0) {
			if (dev) {
				dev_put(dev);
				/* @@@ for DEBUG @@@ */
				TCPCP_DPRINTK("dev->refcnt.counter=%d\n",
					dev->refcnt.counter);
			}

			TCPCP_DPRINTK("chk_addr failed for %x:%x:%x:%x:%x:%x:%x:%x\n",
				      ntohs(inet6->saddr.s6_addr16[0]),
				      ntohs(inet6->saddr.s6_addr16[1]),
				      ntohs(inet6->saddr.s6_addr16[2]),
				      ntohs(inet6->saddr.s6_addr16[3]),
				      ntohs(inet6->saddr.s6_addr16[4]),
				      ntohs(inet6->saddr.s6_addr16[5]),
				      ntohs(inet6->saddr.s6_addr16[6]),
				      ntohs(inet6->saddr.s6_addr16[7]));

			return -EADDRNOTAVAIL;
		}

		if (dev) {
			dev_put(dev);
			/* @@@ for DEBUG @@@ */
			TCPCP_DPRINTK("dev->refcnt.counter=%d\n",
				dev->refcnt.counter);
		}
	}

	/* make sure we are allowed to bind here. */
	snum = ntohs(inet->sport);
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("inet->sport=%u\n", inet->sport);
	TCPCP_DPRINTK("snum       =%u\n", snum);

	/* bind TCP port. */
	sk->sk_reuse = 2;
	err_val = sk->sk_prot->get_port(sk, snum);
	if (err_val != 0) {
		TCPCP_DPRINTK("get_port() returned =%d\n",
			      err_val);
		TCPCP_DPRINTK("inet6->saddr       =%x:%x:%x:%x:%x:%x:%x:%x\n",
			      ntohs(inet6_sk(sk)->saddr.s6_addr16[0]),
			      ntohs(inet6_sk(sk)->saddr.s6_addr16[1]),
			      ntohs(inet6_sk(sk)->saddr.s6_addr16[2]),
			      ntohs(inet6_sk(sk)->saddr.s6_addr16[3]),
			      ntohs(inet6_sk(sk)->saddr.s6_addr16[4]),
			      ntohs(inet6_sk(sk)->saddr.s6_addr16[5]),
			      ntohs(inet6_sk(sk)->saddr.s6_addr16[6]),
			      ntohs(inet6_sk(sk)->saddr.s6_addr16[7]));
		TCPCP_DPRINTK("inet->sport num    =%u\n",
			      ntohs(inet_sk(sk)->sport));
		TCPCP_DPRINTK("sk->sk_bound_dev_if=%u\n",
			      sk->sk_bound_dev_if);
		TCPCP_DPRINTK("inet6->rcv_saddr   =%x:%x:%x:%x:%x:%x:%x:%x\n",
			      ntohs(inet6_sk(sk)->rcv_saddr.s6_addr[0]),
			      ntohs(inet6_sk(sk)->rcv_saddr.s6_addr[1]),
			      ntohs(inet6_sk(sk)->rcv_saddr.s6_addr[2]),
			      ntohs(inet6_sk(sk)->rcv_saddr.s6_addr[3]),
			      ntohs(inet6_sk(sk)->rcv_saddr.s6_addr[4]),
			      ntohs(inet6_sk(sk)->rcv_saddr.s6_addr[5]),
			      ntohs(inet6_sk(sk)->rcv_saddr.s6_addr[6]),
			      ntohs(inet6_sk(sk)->rcv_saddr.s6_addr[7]));
		TCPCP_DPRINTK("sk->sk_state       =%u\n",
			      sk->sk_state);

		TCPCP_DPRINTK("[Error] can't bind TCP port.\n");

		inet_reset_saddr(sk);
		/* @@@ for DEBUG @@@ */
		TCPCP_DPRINTK("inet6->saddr    =%x:%x:%x:%x:%x:%x:%x:%x\n",
			      ntohs(inet6_sk(sk)->saddr.s6_addr16[0]),
			      ntohs(inet6_sk(sk)->saddr.s6_addr16[1]),
			      ntohs(inet6_sk(sk)->saddr.s6_addr16[2]),
			      ntohs(inet6_sk(sk)->saddr.s6_addr16[3]),
			      ntohs(inet6_sk(sk)->saddr.s6_addr16[4]),
			      ntohs(inet6_sk(sk)->saddr.s6_addr16[5]),
			      ntohs(inet6_sk(sk)->saddr.s6_addr16[6]),
			      ntohs(inet6_sk(sk)->saddr.s6_addr16[7]));
		TCPCP_DPRINTK("inet->rcv_saddr=0x%x\n",
			inet_sk(sk)->rcv_saddr);

		return -EADDRINUSE;
	}

	/* end. */
	TCPCP_DPRINTK("end.\n");
	return 0;
}
#endif /* defined(CONFIG_IPV6) || defined(CONFIG_IPV6_MODULE) */

/**
 * tcpcp_v4bind - Step A : TCP/IPv4 bind.
 *
 * This function is based on af_inet.c:inet_bind().
 * Caller must lock socket.
 * 
 * [Args]
 *   (IN) *sk : TCP socket
 * 
 * [Return]
 *   0          : OK.
 *  -EADDRINUSE : NG.
 **/
static int tcpcp_v4bind(struct sock *sk)
{
	struct inet_sock *inet = inet_sk(sk);
	unsigned short snum;

	/* plug-in. */
	plugin_dummy();

	/* start. */
	TCPCP_DPRINTK("start. sk=0x%p\n", sk);

	/* make sure we are allowed to bind here. */
	snum = ntohs(inet->sport);
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("inet->sport=%u\n", inet->sport);
	TCPCP_DPRINTK("snum       =%u\n", snum);

	/* bind TCP port. */
	sk->sk_reuse = 2;
	if (sk->sk_prot->get_port(sk, snum) != 0) {
		TCPCP_DPRINTK("[Error] can't bind TCP port.\n");

		inet->saddr = inet->rcv_saddr = 0;
		/* @@@ for DEBUG @@@ */
		TCPCP_DPRINTK("inet->saddr    =0x%x\n",
			inet->saddr);
		TCPCP_DPRINTK("inet->rcv_saddr=0x%x\n",
			inet->rcv_saddr);

		return -EADDRINUSE;
	}

	/* end. */
	TCPCP_DPRINTK("end.\n");
	return 0;
}

#if defined(CONFIG_IPV6) || defined(CONFIG_IPV6_MODULE)
/**
 * tcpcp_v6connect - Step B : TCP/IPv6 connect.
 *
 * This function is based on tcp_ipv6.c:tcp_v6_connect().
 * It walks TCP through to the SYN_SENT state.
 * Caller must lock socket.
 * 
 * [Args]
 *   (IN) *sk : TCP socket
 * 
 * [Return]
 *   0             : OK.
 *  -EINVAL        : NG.
 *  -EADDRNOTAVAIL : NG.
 *  -EPERM         : NG.
 *  -EAGAIN        : NG.
 *  -ERESTART      : NG.
 *   other         : NG. (xfrm_lookup)
 **/
static int tcpcp_v6connect(struct sock *sk)
{
	struct inet_sock *inet = inet_sk(sk);
	struct ipv6_pinfo *inet6 = inet6_sk(sk);
	struct tcp_sock *tp = tcp_sk(sk);
	struct in6_addr *final_p = NULL, final;
	struct flowi fl;
	struct dst_entry *dst;
	int error;

	/* plug-in. */
	plugin_dummy();

	/* start. */
	TCPCP_DPRINTK("start. sk=0x%p\n", sk);

	/* set flow label. */
	memset(&fl, 0, sizeof(fl));

#if 0
	if (inet6->sndflow) {
	  fl.fl6_flowlabel = usin->sin6_flowinfo & IPV6_FLOWINFO_MASK;
	  IP6_ECN_flow_init(fl.fl6_flowlabel);
	  if (fl.fl6_flowlabel&IPV6_FLOWLABEL_MASK) {
	    struct ip6_flowlabel *flowlabel;
	    flowlabel = fl6_sock_lookup(sk, fl.fl6_flowlabel);
	    if (flowlabel == NULL)
	      return -EINVAL;
	    ipv6_addr_copy(&usin->sin6_addr, &flowlabel->dst);
	    fl6_sock_release(flowlabel);
	  }
	}
#endif

	fl.proto = IPPROTO_TCP;
	ipv6_addr_copy(&(fl.fl6_dst), &(inet6->daddr));
	ipv6_addr_copy(&(fl.fl6_src), &(inet6->saddr));
	fl.oif = sk->sk_bound_dev_if;
	fl.fl_ip_dport = inet->dport;
	fl.fl_ip_sport = inet->sport;
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("fl.proto=%u\n", fl.proto);
	TCPCP_DPRINTK("inet6->daddr=%04x:%04x:%04x:%04x:"
		"%04x:%04x:%04x:%04x\n",
		ntohs(inet6->daddr.s6_addr16[0]),
		ntohs(inet6->daddr.s6_addr16[1]),
		ntohs(inet6->daddr.s6_addr16[2]),
		ntohs(inet6->daddr.s6_addr16[3]),
		ntohs(inet6->daddr.s6_addr16[4]),
		ntohs(inet6->daddr.s6_addr16[5]),
		ntohs(inet6->daddr.s6_addr16[6]),
		ntohs(inet6->daddr.s6_addr16[7]));
	TCPCP_DPRINTK("fl.fl6_dst  =%04x:%04x:%04x:%04x:"
		"%04x:%04x:%04x:%04x\n",
		ntohs(fl.fl6_dst.s6_addr16[0]),
		ntohs(fl.fl6_dst.s6_addr16[1]),
		ntohs(fl.fl6_dst.s6_addr16[2]),
		ntohs(fl.fl6_dst.s6_addr16[3]),
		ntohs(fl.fl6_dst.s6_addr16[4]),
		ntohs(fl.fl6_dst.s6_addr16[5]),
		ntohs(fl.fl6_dst.s6_addr16[6]),
		ntohs(fl.fl6_dst.s6_addr16[7]));
	TCPCP_DPRINTK("inet6->saddr=%04x:%04x:%04x:%04x:"
		"%04x:%04x:%04x:%04x\n",
		ntohs(inet6->saddr.s6_addr16[0]),
		ntohs(inet6->saddr.s6_addr16[1]),
		ntohs(inet6->saddr.s6_addr16[2]),
		ntohs(inet6->saddr.s6_addr16[3]),
		ntohs(inet6->saddr.s6_addr16[4]),
		ntohs(inet6->saddr.s6_addr16[5]),
		ntohs(inet6->saddr.s6_addr16[6]),
		ntohs(inet6->saddr.s6_addr16[7]));
	TCPCP_DPRINTK("fl.fl6_src  =%04x:%04x:%04x:%04x:"
		"%04x:%04x:%04x:%04x\n",
		ntohs(fl.fl6_src.s6_addr16[0]),
		ntohs(fl.fl6_src.s6_addr16[1]),
		ntohs(fl.fl6_src.s6_addr16[2]),
		ntohs(fl.fl6_src.s6_addr16[3]),
		ntohs(fl.fl6_src.s6_addr16[4]),
		ntohs(fl.fl6_src.s6_addr16[5]),
		ntohs(fl.fl6_src.s6_addr16[6]),
		ntohs(fl.fl6_src.s6_addr16[7]));
	TCPCP_DPRINTK("sk->sk_bound_dev_if=%u\n",
		sk->sk_bound_dev_if);
	TCPCP_DPRINTK("fl.oif             =%u\n",
		fl.oif);
	TCPCP_DPRINTK("inet->dport   =%u\n",
		inet->dport);
	TCPCP_DPRINTK("fl.fl_ip_dport=%u\n",
		fl.fl_ip_dport);
	TCPCP_DPRINTK("inet->sport   =%u\n",
		inet->sport);
	TCPCP_DPRINTK("fl.fl_ip_sport=%u\n",
		fl.fl_ip_sport);

	/* @@@ for DEBUG @@@ */
	if (inet6->opt != 0)
		TCPCP_DPRINTK("inet6->opt->srcrt=0x%p\n",
			inet6->opt->srcrt);

	/* .... */
	if ((inet6->opt != 0) && (inet6->opt->srcrt != 0)) {
		struct rt0_hdr *rt0 = (struct rt0_hdr *)inet6->opt->srcrt;
		ipv6_addr_copy(&final, &(fl.fl6_dst));
		ipv6_addr_copy(&(fl.fl6_dst), rt0->addr);
		final_p = &final;
		/* @@@ for DEBUG @@@  */
		TCPCP_DPRINTK("rt0=0x%p\n", rt0);
		TCPCP_DPRINTK("final     =%04x:%04x:%04x:%04x:"
			"%04x:%04x:%04x:%04x\n",
			ntohs(final.s6_addr16[0]),
			ntohs(final.s6_addr16[1]),
			ntohs(final.s6_addr16[2]),
			ntohs(final.s6_addr16[3]),
			ntohs(final.s6_addr16[4]),
			ntohs(final.s6_addr16[5]),
			ntohs(final.s6_addr16[6]),
			ntohs(final.s6_addr16[7]));
		TCPCP_DPRINTK("rt0->addr =%04x:%04x:%04x:%04x:"
			"%04x:%04x:%04x:%04x\n",
			ntohs(rt0->addr[0].s6_addr16[0]),
			ntohs(rt0->addr[0].s6_addr16[1]),
			ntohs(rt0->addr[0].s6_addr16[2]),
			ntohs(rt0->addr[0].s6_addr16[3]),
			ntohs(rt0->addr[0].s6_addr16[4]),
			ntohs(rt0->addr[0].s6_addr16[5]),
			ntohs(rt0->addr[0].s6_addr16[6]),
			ntohs(rt0->addr[0].s6_addr16[7]));
		TCPCP_DPRINTK("fl.fl6_dst=%04x:%04x:%04x:%04x:"
			"%04x:%04x:%04x:%04x\n",
			ntohs(fl.fl6_dst.s6_addr16[0]),
			ntohs(fl.fl6_dst.s6_addr16[1]),
			ntohs(fl.fl6_dst.s6_addr16[2]),
			ntohs(fl.fl6_dst.s6_addr16[3]),
			ntohs(fl.fl6_dst.s6_addr16[4]),
			ntohs(fl.fl6_dst.s6_addr16[5]),
			ntohs(fl.fl6_dst.s6_addr16[6]),
			ntohs(fl.fl6_dst.s6_addr16[7]));
		TCPCP_DPRINTK("&final =0x%p\n", &final);
		TCPCP_DPRINTK("final_p=0x%p\n", final_p);
	}

	/* get Routing Table. */
	error = tcpcp_ip6_dst_lookup(sk, &dst, &fl);
	if (error != 0) {
		TCPCP_DPRINTK("[Error] tcp_ip6_dst_lookup failure."
			" error=%d\n", error);
		goto err;
	}
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("after tcpcp_ip6_dst_lookup.\n");
	TCPCP_DPRINTK("dst->__refcnt.counter=%d\n",
		      dst->__refcnt.counter);

	/* set daddr from Routing Table ? */
	if (final_p != NULL) {
		ipv6_addr_copy(&fl.fl6_dst, final_p);
		/* @@@ for DEBUG @@@ */
		TCPCP_DPRINTK("final_p   =%04x:%04x:%04x:%04x:"
			"%04x:%04x:%04x:%04x\n",
			ntohs(final_p->s6_addr16[0]),
			ntohs(final_p->s6_addr16[1]),
			ntohs(final_p->s6_addr16[2]),
			ntohs(final_p->s6_addr16[3]),
			ntohs(final_p->s6_addr16[4]),
			ntohs(final_p->s6_addr16[5]),
			ntohs(final_p->s6_addr16[6]),
			ntohs(final_p->s6_addr16[7]));
		TCPCP_DPRINTK("fl.fl6_dst=%04x:%04x:%04x:%04x:"
			"%04x:%04x:%04x:%04x\n",
			ntohs(fl.fl6_dst.s6_addr16[0]),
			ntohs(fl.fl6_dst.s6_addr16[1]),
			ntohs(fl.fl6_dst.s6_addr16[2]),
			ntohs(fl.fl6_dst.s6_addr16[3]),
			ntohs(fl.fl6_dst.s6_addr16[4]),
			ntohs(fl.fl6_dst.s6_addr16[5]),
			ntohs(fl.fl6_dst.s6_addr16[6]),
			ntohs(fl.fl6_dst.s6_addr16[7]));
	}

	/* .... */
	error = __xfrm_lookup(&dst, &fl, sk, XFRM_LOOKUP_WAIT);
	if (error < 0) {
		TCPCP_DPRINTK("[Error] xfrm_lookup failure. error=%d\n",
			error);

		/* @@@ for DEBUG @@@ */
		TCPCP_DPRINTK("dst->__refcnt.counter=%d\n",
			dst->__refcnt.counter);

		dst_release(dst);
		/* @@@ for DEBUG @@@ */
		TCPCP_DPRINTK("dst->__refcnt.counter=%d\n",
			dst->__refcnt.counter);

		goto err;
	}
	/* @@@ for DEBUG @@@ */
        TCPCP_DPRINTK("after xfrm_lookup.\n");

	/* change state from TCP_CLOSE to TCP_SYN_SENT.*/
	tcp_set_state(sk, TCP_SYN_SENT);
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("sk->sk_state=%u\n", sk->sk_state);

	/* bind a port for a connect operation and hash it. */
	error = tcpcp_inet6_hash_connect(&tcp_death_row, sk);
	if (error != 0) {
		/* only -EADDRNOTAVAIL. not return 1. */
		TCPCP_DPRINTK("[Error] tcpcp_inet_hash_connect failure."
			" error=%d\n", error);

		/* @@@ for DEBUG @@@ */
		TCPCP_DPRINTK("dst->__refcnt.counter =%d\n",
			dst->__refcnt.counter);

		tcp_set_state(sk, TCP_CLOSE);
		dst_release(dst);
		/* @@@ for DEBUG @@@ */
		TCPCP_DPRINTK("sk->sk_state=%u\n",
			sk->sk_state);
		TCPCP_DPRINTK("dst->__refcnt.counter=%d\n",
			dst->__refcnt.counter);

		goto err;
	}
	/* @@@ for DEBUG @@@ */
        TCPCP_DPRINTK("after tcpcp_inet_hash_connect.\n");

	/* set Routing Table. */
        __ip6_dst_store(sk, dst, NULL, NULL);
	sk->sk_route_caps
	    = dst->dev->features & ~(NETIF_F_IP_CSUM | NETIF_F_TSO);
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("dst->dev->features=%lu\n",
		dst->dev->features);
	TCPCP_DPRINTK("sk->sk_route_caps =%d\n",
		sk->sk_route_caps);



	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("dst->header_len    =%d\n",
		dst->header_len);
	TCPCP_DPRINTK("inet_csk(sk)->icsk_ext_hdr_len=%u\n",
		inet_csk(sk)->icsk_ext_hdr_len);

	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("tp->mss_cache    =%u\n",
		tp->mss_cache);
	TCPCP_DPRINTK("inet_csk(sk)->icsk_pmtu_cookie  =%u\n",
          inet_csk(sk)->icsk_pmtu_cookie);

	/* .... */
	tcp_sync_mss(sk, dst_mtu(dst));
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("tp->mss_cache    =%u\n",
                tp->mss_cache);
  TCPCP_DPRINTK("inet_csk(sk)->icsk_pmtu_cookie  =%u\n",
          inet_csk(sk)->icsk_pmtu_cookie);

	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("tp->window_clamp=%u\n",
		tp->window_clamp);

	/* set window_clamp. */
	if (tp->window_clamp == 0)
		tp->window_clamp = dst_metric(dst, RTAX_WINDOW);
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("tp->window_clamp=%u\n",
		tp->window_clamp);

	/* end. */
	TCPCP_DPRINTK("end.\n");
	return 0;

err:
	inet->dport = 0;
	sk->sk_route_caps = 0;
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("inet->dport=%u\n",
		inet->dport);
	TCPCP_DPRINTK("sk->sk_route_caps=%d\n",
		sk->sk_route_caps);

	/* end. */
	TCPCP_DPRINTK("end. error=%d\n", error);
	return error;
}
#endif /* defined(CONFIG_IPV6) || defined(CONFIG_IPV6_MODULE) */

/**
 * tcpcp_v4connect - Step B : TCP/IPv4 connect.
 *
 * This function is based on tcp_ipv4.c:tcp_v4_connect().
 * It walks TCP through to the SYN_SENT state.
 * Caller must lock socket.
 *
 * [Args]
 *   (IN) *sk : TCP socket
 * 
 * [Return]
 *   0             : OK.
 *  -EINVAL        : NG.
 *  -ENETUNREACH   : NG.
 *  -ENOBUFS       : NG.
 *  -EPERM         : NG.
 *  -EAGAIN        : NG.
 *  -ERESTART      : NG.
 *  -EADDRNOTAVAIL : NG.
 *   other         : NG. (xfrm_lookup)
 **/
static int tcpcp_v4connect(struct sock *sk)
{
	struct inet_sock *inet = inet_sk(sk);
	struct tcp_sock *tp = tcp_sk(sk);
  struct net *net = sock_net(sk);
	struct rtable *rt;
	u32 daddr, nexthop;
	struct flowi fl;
	int error;

	/* plug-in. */
	plugin_dummy();

	/* start. */
	TCPCP_DPRINTK("start. sk=0x%p\n", sk);

	/* set nexthop. */
	nexthop = daddr = inet->daddr;
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("inet->daddr=0x%x\n", inet->daddr);
	TCPCP_DPRINTK("nexthop    =0x%x\n", nexthop);

	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("inet->opt=0x%p\n", inet->opt);

	if ((inet->opt != NULL) && (inet->opt->srr != 0)) {
    TCPCP_DPRINTK("inet->opt->srr=%u\n",
            inet->opt->srr);

		if (daddr == 0) {
			TCPCP_DPRINTK("[Error] daddr is 0.0.0.0.\n");
			return -EINVAL;
		}
		nexthop = inet->opt->faddr;
		/* @@@ for DEBUG @@@ */
		TCPCP_DPRINTK("inet->opt->faddr=0x%x\n",
			inet->opt->faddr);
	}
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("nexthop=0x%x\n", nexthop);

	/*
	 * This code is based on include/net/route.h:ip_route_connect().
	 */
	fl.oif = sk->sk_bound_dev_if;
	fl.fl4_dst = nexthop;
	fl.fl4_src = inet->saddr;
	fl.fl4_tos = RT_CONN_FLAGS(sk);
	fl.proto = IPPROTO_TCP;
	fl.fl_ip_sport = inet->sport;
	fl.fl_ip_dport = inet->dport;

	/* get Routing Table. */
	if ((nexthop == 0) || (inet->saddr == 0)) {
		error = __ip_route_output_key(net, &rt, &fl);
		if (error != 0) {
			TCPCP_DPRINTK("[Error] __ip_route_output_key"
				" failure. error=%d\n", error);
			return error;
		}

		/* @@@ for DEBUG @@@ */
		TCPCP_DPRINTK("rt->u.dst.__refcnt.counter =%d\n",
			rt->u.dst.__refcnt.counter);

		fl.fl4_dst = rt->rt_dst;
		fl.fl4_src = rt->rt_src;
		ip_rt_put(rt);
		/* @@@ for DEBUG @@@ */
		TCPCP_DPRINTK("rt->rt_dst=0x%x\n",
			rt->rt_dst);
		TCPCP_DPRINTK("fl.fl4_dst=0x%x\n",
			fl.fl4_dst);
                TCPCP_DPRINTK("rt->rt_src=0x%x\n",
                        rt->rt_src);
		TCPCP_DPRINTK("fl.fl4_src=0x%x\n",
			fl.fl4_src);
                TCPCP_DPRINTK("rt->u.dst.__refcnt.counter=%d\n",
                        rt->u.dst.__refcnt.counter);

		rt = NULL;
	}

  security_sk_classify_flow(sk, &fl);
	error =  tcpcp_ip_route_output_flow(net, &rt, &fl, sk, 0);
	if (error < 0) {
		TCPCP_DPRINTK("[Error] tcpcp_ip_route_output_flow failure"
			" (1). error=%d\n", error);
		if (error == -ENETUNREACH)
			IP_INC_STATS_BH(net, IPSTATS_MIB_OUTNOROUTES);
		return error;
	}

	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("rt->rt_flags=%u\n", rt->rt_flags);

	/* check flags of Routing Table. */
	if (rt->rt_flags & (RTCF_MULTICAST | RTCF_BROADCAST)) {
		TCPCP_DPRINTK("[Error] rt_flags is multi/broadcast.\n");

		ip_rt_put(rt);
		/* @@@ DEBUG @@@ */
		TCPCP_DPRINTK("rt->u.dst.__refcnt.counter=%d\n",
			rt->u.dst.__refcnt.counter);

		return -ENETUNREACH;
	}

	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("inet->opt     =0x%p\n", inet->opt);
  if( NULL != inet->opt ) {
    TCPCP_DPRINTK("inet->opt->srr=%u\n",
            inet->opt->srr);
  }
  
	/* set daddr from Routing Table. */
	if ((inet->opt == NULL) || (inet->opt->srr == 0)) {
	 	daddr = rt->rt_dst;
		/* @@@ for DEBUG @@@ */
		TCPCP_DPRINTK("daddr=0x%x\n", daddr);
	}

	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("tp->rx_opt.ts_recent_stamp=%lu\n",	
		tp->rx_opt.ts_recent_stamp);
	TCPCP_DPRINTK("rt->rt_dst=0%x\n", rt->rt_dst);
	TCPCP_DPRINTK("daddr     =0x%x\n", daddr);

	if ((tcp_death_row.sysctl_tw_recycle != 0) && (tp->rx_opt.ts_recent_stamp == 0)
					 && (rt->rt_dst == daddr)) {
		/*
		 * This code is based on include/net/route.h:rt_get_peer().
		 */
		struct inet_peer *peer;

		if (rt->peer != 0) {
			peer = rt->peer;
			/* @@@ for DEBUG @@@ */
			TCPCP_DPRINTK("rt->peer=0x%p\n",
				rt->peer);
			TCPCP_DPRINTK("peer    =0x%p\n",
				peer);
		} else {
			/* @@@ for DEBUG @@@ */
			TCPCP_DPRINTK("before tcpcp_rt_bind_peer.\n");

			tcpcp_rt_bind_peer(rt, 0);
			peer = rt->peer;

			/* @@@ for DEBUG @@@ */
			TCPCP_DPRINTK("after tcpcp_rt_bind_peer.\n");
			TCPCP_DPRINTK("rt->peer=0x%p\n",
				rt->peer);
			TCPCP_DPRINTK("peer    =0x%p\n",
				peer);
		}

		/* VJ's idea. 
		 * we save last timestamp seen from the destination in peer
		 * table, when entering state TIME-WAIT and initialize
		 * ts_recent from it, when trying new connection. */

		/* @@@ for DEBUG @@@ */
		TCPCP_DPRINTK("peer=0x%p\n", peer);
		if (peer != 0) {
			TCPCP_DPRINTK("peer->tcp_ts_stamp+TCP_PAWS_MSL"
				"=%lu\n",
				peer->tcp_ts_stamp + TCP_PAWS_MSL);
			TCPCP_DPRINTK("get_seconds()                  "
              "=%ld\n", get_seconds());
		}

		if ((peer != 0)
		    && (peer->tcp_ts_stamp + TCP_PAWS_MSL >= get_seconds())) {
			tp->rx_opt.ts_recent_stamp = peer->tcp_ts_stamp;
			tp->rx_opt.ts_recent = peer->tcp_ts;
			/* @@@ for DEBUG @@@ */
			TCPCP_DPRINTK("peer->tcp_ts_stamp =%lu\n",
				peer->tcp_ts_stamp);
			TCPCP_DPRINTK("tp->rx_opt.ts_recent_stamp=%lu\n",
				tp->rx_opt.ts_recent_stamp);
			TCPCP_DPRINTK("peer->tcp_ts =%u\n",
				peer->tcp_ts);
			TCPCP_DPRINTK("tp->rx_opt.ts_recent=%u\n",
				tp->rx_opt.ts_recent);
		}
	}

	/* change state from TCP_CLOSE to TCP_SYN_SENT. */
	tcp_set_state(sk, TCP_SYN_SENT);
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("sk->sk_state=%d\n", sk->sk_state);

	/* bind a port for a connect operation and hash it. */
	error = tcpcp_inet_hash_connect(&tcp_death_row, sk);
	if (error != 0) {
		/* only -EADDRNOTAVAIL. */
		TCPCP_DPRINTK("[Error] tcpcp_inet_hash_connect failure."
			" error=%d\n", error);
		goto err;
	}

	/*
	 * This code is based on include/net/route.h:ip_route_newports().
	 */

	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("inet->sport       =%u\n",
		 inet->sport);
	TCPCP_DPRINTK("rt->fl.fl_ip_sport=%u\n",
		rt->fl.fl_ip_sport);
	TCPCP_DPRINTK("inet->dport       =%u\n",
		inet->dport);
	TCPCP_DPRINTK("rt->fl.fl_ip_dport=%u\n",
		rt->fl.fl_ip_dport);

	/* .... */
	if ((inet->sport != rt->fl.fl_ip_sport) ||
	    (inet->dport != rt->fl.fl_ip_dport)) {

		memcpy(&fl, &(rt)->fl, sizeof(fl));
		fl.fl_ip_sport = inet->sport;
		fl.fl_ip_dport = inet->dport;
		ip_rt_put(rt);
		rt = NULL;

		/* @@@ for DEBUG @@@ */
		TCPCP_DPRINTK("fl.fl_ip_sport=%u\n",
			fl.fl_ip_sport);
		TCPCP_DPRINTK("fl.fl_ip_dport=%u\n",
			fl.fl_ip_dport);

    security_sk_classify_flow(sk, &fl);
		error = tcpcp_ip_route_output_flow(net, &rt, &fl, sk, 0);
		if (error != 0) {
			TCPCP_DPRINTK("[Error] tcpcp_ip_route_output_flow"
				" failure (2). error=%d\n",
				error);	
			goto err;
		}
	}

	/* @@@ for DEBUG */
	TCPCP_DPRINTK("&(rt->u.dst)=0x%p\n",
		&(rt->u.dst));
	TCPCP_DPRINTK("sk->sk_dst_cache=0x%p\n",
		sk->sk_dst_cache);
	TCPCP_DPRINTK("sk->sk_route_caps=%u\n",
		sk->sk_route_caps);

	/* OK, now commit destination to socket.  */
	sk->sk_gso_type = SKB_GSO_TCPV4;
	sk_setup_caps(sk, &(rt->u.dst));

	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("sk->sk_dst_cache=0x%p\n",
		sk->sk_dst_cache);
	TCPCP_DPRINTK("sk->sk_route_caps=%u\n",
		sk->sk_route_caps);



	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("rt->u.dst.header_len=%u\n",
		rt->u.dst.header_len);
	TCPCP_DPRINTK("inet_csk(sk)->icsk_ext_hdr_len =%u\n",
		inet_csk(sk)->icsk_ext_hdr_len);

	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("tp->mss_cache    =%u\n",
		tp->mss_cache);
	TCPCP_DPRINTK("inet_csk(sk)->icsk_pmtu_cookie  =%u\n",
          inet_csk(sk)->icsk_pmtu_cookie);

	/* .... */
	tcp_sync_mss(sk, dst_mtu(&(rt->u.dst)));
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("tp->mss_cache    =%u\n",
		tp->mss_cache);
	TCPCP_DPRINTK("inet_csk(sk)->icsk_pmtu_cookie  =%u\n",
		inet_csk(sk)->icsk_pmtu_cookie);

	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("tp->window_clamp=%u\n",
		tp->window_clamp);

	/* set window_clamp. */
	if (tp->window_clamp == 0)
		tp->window_clamp = dst_metric(&(rt->u.dst), RTAX_WINDOW);
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("tp->window_clamp=%u\n",
		tp->window_clamp);

	rt = NULL;

	/* end. */
	TCPCP_DPRINTK("end.\n");
	return 0;

err:
	/* This unhashes the socket and releases the local port,
	 * if necessary. */

	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("@@@ ERROR OPERATION! @@@\n");
	TCPCP_DPRINTK("sk->sk_state=%u\n",
		sk->sk_state);
	if (rt != NULL)
		TCPCP_DPRINTK("rt->u.dst.__refcnt.counter =%d\n",
			rt->u.dst.__refcnt.counter);

	tcp_set_state(sk, TCP_CLOSE);
	ip_rt_put(rt);
	sk->sk_route_caps = 0;
	inet->dport = 0;
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("sk->sk_state=%u\n", sk->sk_state);
	if (rt != NULL)
		TCPCP_DPRINTK("rt->u.dst.__refcnt.counter =%d\n",
			rt->u.dst.__refcnt.counter);
	TCPCP_DPRINTK("sk->sk_route_caps=%d\n",
		sk->sk_route_caps);
	TCPCP_DPRINTK("inet->dport=%u\n", inet->dport);

	/* end. */
	TCPCP_DPRINTK("end. error=%d\n", error);
	return error;
}

/**
 * do_start - start TCP socket.
 *
 * Caller must lock socket.
 *
 * [Args]
 *   (IN) *sk : TCP socket
 *
 * [Return]
 *   0      : OK
 *  -EINVAL        : NG. (v6connect/v4connect)
 *  -EADDRNOTAVAIL : NG. (v6connect/v4connect)
 *  -EPERM         : NG. (v6connect/v4connect)
 *  -EAGAIN        : NG. (v6connect/v4connect)
 *  -ERESTART      : NG. (v6connect/v4connect)
 *   1             : NG. (v6connect)
 *  -ENETUNREACH   : NG. (v4connect)
 *  -ENOBUFS       : NG. (v4connect)
 *  -EADDRINUSE    : NG. (v6bind/v4bind)
 *  -ENODEV        : NG. (v6bind)
 *  -EADDRNOTAVAIL : NG. (v6bind)
 *   other         : NG. (v6connet/v4connet)
 **/
static int do_start(struct sock *sk)
{
	struct tcp_sock *tp = tcp_sk(sk);
	struct sk_buff *skb;
	int error;

	/* plug-in. */
	plugin_dummy();

	/* start. */
	TCPCP_DPRINTK("start. sk=0x%p\n", sk);

	if (tp->inet_conn.icsk_af_ops->queue_xmit != tcpcp_discard_xmit) {
		TCPCP_DPRINTK("TCPCP socket stop proccess is not"
			" performed.\n");
		return 0;
	}

	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("tp->set_flag=%u\n", tp->set_flag);

	if (tp->set_flag == TCPCP_SET_FLAG_ON) {
		TCPCP_DPRINTK("TCPCP set SIs proccess is performed.\n");
#if defined(CONFIG_IPV6) || defined(CONFIG_IPV6_MODULE)
		if (sk->sk_family == PF_INET6) {
			/* Step A : bind. */
			error = tcpcp_v6bind(sk);
			if (error != 0) {
				TCPCP_DPRINTK("[Error] can't bind."
					" error=%d\n", error);
				return error;
			}

			/* Step B : connect. */
			error = tcpcp_v6connect(sk);
		} else 
#endif
		{
			/* Step A : bind. */
			error = tcpcp_v4bind(sk);
			if (error != 0) {
				TCPCP_DPRINTK("[Error] can't bind."
					" error=%d\n", error);
				return error;
			}

			/* Step B : connect. */
			error = tcpcp_v4connect(sk);
		}
		if (error != 0) {
			TCPCP_DPRINTK("[Error] can't connect."
				" error=%d\n", error);
			return error;
		}

		/* .... */
		sk->sk_socket->state = SS_CONNECTED;
		tcp_set_state(sk, TCP_ESTABLISHED);
	}

	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("sk->sk_socket->state=%d\n",
		sk->sk_socket->state);
	TCPCP_DPRINTK("sk->sk_state=%d\n",
		sk->sk_state);

	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("sk->sk_filter=0x%p\n",
		sk->sk_filter);
	TCPCP_DPRINTK("&drop_all    =0x%p\n", &drop_all);

	/* release dummy socket filter. */
	if (sk->sk_filter != NULL) {
		sk_filter_release(sk->sk_filter);
		sk->sk_filter = NULL;
		/* @@@ for DEBUG @@@ */
		TCPCP_DPRINTK("socket filter is released.\n");
		TCPCP_DPRINTK("drop_all.refcnt=%d\n",
			atomic_read(&(drop_all.refcnt)));
	}
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("sk->sk_filter=0x%p\n",
		sk->sk_filter);

	tcpcp_change_xmit(sk, TCPCP_FLAG_START);

	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("tp->snd_nxt   =%lu\n",
		(unsigned long) tp->snd_nxt);
	TCPCP_DPRINTK("tp->rcv_nxt   =%lu\n",
		(unsigned long) tp->rcv_nxt);
	TCPCP_DPRINTK("tp->rcv_wup   =%lu\n",
		(unsigned long) tp->rcv_wup);
	TCPCP_DPRINTK("tp->copied_seq=%lu\n",
		(unsigned long) tp->copied_seq);

	/* .... */
	skb = skb_peek(&sk->sk_write_queue);
	if (skb == NULL) {
		TCPCP_DPRINTK("Write queue is empty.\n");

		if (tp->set_flag == TCPCP_SET_FLAG_ON) {
			TCPCP_DPRINTK("ACK whose payload size of" \
				" a packet is 0 is transmitted.\n");
			tcpcp_send_probe0(sk);
		}

		/* end. */
		TCPCP_DPRINTK("end.\n");
		return 0;
	}

	if (sk->sk_send_head != skb) {
		TCPCP_DPRINTK("Got something to retransmit.\n");

		/* @@@ for DEBUG @@@ */
		TCPCP_DPRINTK("skb=0x%p\n", skb);
		TCPCP_DPRINTK("TCP_SKB_CB(skb)->seq    =%lu\n",
			(unsigned long) TCP_SKB_CB(skb)->seq);
		TCPCP_DPRINTK("TCP_SKB_CB(skb)->end_seq=%lu\n",
			(unsigned long) TCP_SKB_CB(skb)->end_seq);
		TCPCP_DPRINTK("TCP_SKB_CB(skb)->when   =%lu\n",
			      (unsigned long)TCP_SKB_CB(skb)->when);
		TCPCP_DPRINTK("tp->snd_una             =%lu\n",
			(unsigned long) tp->snd_una);
		TCPCP_DPRINTK("sk->sk_wmem_alloc=%lu\n",
			(unsigned long) atomic_read(&sk->sk_wmem_alloc));
		TCPCP_DPRINTK("sk->sk_wmem_queued=%lu\n",
			(unsigned long) sk->sk_wmem_queued);
		TCPCP_DPRINTK("sk->sk_sndbuf     =%lu\n",
			(unsigned long) sk->sk_sndbuf);
		TCPCP_DPRINTK("tp->snd_wnd=%lu\n",
			(unsigned long) tp->snd_wnd);
		TCPCP_DPRINTK("tp->packets_out=%lu\n",
			      (unsigned long) tp->packets_out);

		tcpcp_retransmit_skb(sk, skb_peek(&sk->sk_write_queue));
		inet_csk_reset_xmit_timer(sk, ICSK_TIME_RETRANS, inet_csk(sk)->icsk_rto, TCP_RTO_MAX);
	}

	/* end. */
	TCPCP_DPRINTK("end.\n");
	return 0;
}

/**
 * do_tcpcp_start - start TCP sockets.
 *
 * [Args]
 *   (IN) *sk       : TCP socket
 *   (IN) *r_entry  : Request Entry
 *   (IN) *user_res : Processing Result Info.
 *
 * [Return]
 *   0      : OK.
 *  -ENOMEM : NG. (no memory)
 **/
static int do_tcpcp_start(struct sock *sk, struct tcpcp_req_entry *r_entry,
			  struct result *user_res)
{
	struct task_struct *tsk = current;
	struct files_struct *files = tsk->files;
	struct tcpcp_com_entry *fdr_entry;
	unsigned int fd;
	unsigned int fd_all = 0;
	unsigned int fd_ng = 0;
	int error;
	struct fdtable *fdt;
	/* plug-in. */
	plugin_dummy();

	/* start. */
	TCPCP_DPRINTK("start. sk=0x%p, r_entry=0x%p, user_res=0x%p\n",
		sk, r_entry, user_res);

	if (r_entry->flag & TCPCP_FLAG_ALL) {
 		/* add FDR Info. Entry. (only "ALL")*/
		error = tcpcp_add_com_entry(r_entry->flag, &(r_entry->f_entry));
		if (error != 0)
			return -ENOMEM;
	}
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("r_entry->f_entry=0x%p\n",
		r_entry->f_entry);

	/* initialize now FDR Info. Entry. */
	fdr_entry = r_entry->f_entry;
	fdt = files_fdtable(files);
	/* search FD. */
	for (fd = 0; fd < fdt->max_fds; fd++) {
		struct sock *sk_i;
		int ret;

		/* check FD and get SOCKET_I(). */
		sk_i = tcpcp_get_sk_i(sk, fd, r_entry);
		if (sk_i == NULL) {
			/* @@@ for DEBUG @@@ */
			TCPCP_DPRINTK("sk_i is NULL.\n");

			continue;
		}

		/* start TCP socket. */
		lock_sock(sk_i);
		ret = do_start(sk_i);
		release_sock(sk_i);
		/* @@@ for DEBUG @@@ */
		TCPCP_DPRINTK("ret=%d\n", ret);

		fd_all++;

		if (!(r_entry->flag & TCPCP_FLAG_ALL)) {
			/* @@@ for DEBUG @@@ */
			TCPCP_DPRINTK("Break for loop.\n");

			if (ret != 0)
				return ret;
			break;
		}

		/* Only ALL, the following processings are performed. */
		if (ret != 0)
			fd_ng++;

		/* make FDR Info. (only "ALL") */
		error = tcpcp_make_res(fd, ret, &fdr_entry);
		if (error != 0) {
			/* only -ENOMEM. */
			TCPCP_DPRINTK("[Error] can't make FDR Info."
				" error=%d\n", error);
			/* Caller delete Request Entry with common entries. */
			return error;
		}
	}
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("Loop count=%d\n", fd);

#ifdef CONFIG_TCPCP_DEBUG
	/* @@@ print Process List with lock. (for DEBUG) @@@ */
	print_list();
#endif

	/* set Processing Result Info. */
	user_res->fd_all = fd_all;
	user_res->fd_ng = fd_ng;
	user_res->buf_size = 0;
	user_res->fdrs_size = sizeof(struct fdr) * fd_all;
	memcpy(&(r_entry->res), user_res, sizeof(struct result));
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("r_entry->res->fd_all   =%u\n",
		r_entry->res.fd_all);
	TCPCP_DPRINTK("r_entry->res->fd_ng    =%u\n",
		r_entry->res.fd_ng);
	TCPCP_DPRINTK("r_entry->res->buf_size =%u\n",
		r_entry->res.buf_size);
	TCPCP_DPRINTK("r_entry->res->fdrs_size=%u\n",
		r_entry->res.fdrs_size);

	/* end. */
	TCPCP_DPRINTK("end.\n");
	return 0;
}


/* ----- get TCP Socket Informations (TCP-SIs),
         get FD Result (FDR) Informations ---------------------------------- */
 
/**
 * do_get_si - get TCP-SIs.
 *
 * Caller must reset timeout.
 *
 * [Args]
 *   (IN) *r_entry    : Request Entry
 *   (OUT) *user_si   : TCP-SIs returned to user
 *   (OUT) *user_size : length of TCP-SIs returned to user
 *
 * [Retrun]
 *   0      : OK.
 *  -EFAULT : NG.
 **/
static int do_get_si(struct tcpcp_req_entry *r_entry,
		     struct tcpcp_total_si_hdr *user_si,
		     unsigned int *user_size)
{
	struct tcpcp_com_entry *si_entry;
	char *p_si = NULL;
	unsigned int data_len;
	int error;

	/* plug-in. */
	plugin_dummy();

	/* start. */
	TCPCP_DPRINTK("start. r_entry=0x%p\n", r_entry);

	/* put size of TCP-SIs to userland. */
	error = put_user(r_entry->res.buf_size, user_size);
	if (error != 0)
		return -EFAULT;

	p_si = (char *)user_si;
	si_entry = r_entry->s_entry;

	while (si_entry != NULL) {
		/* @@@ for DEBUG @@@ */
		TCPCP_DPRINTK("si_entry=0x%p.\n",
			si_entry);

		data_len = si_entry->data_len;

		/* copy TCP-SIs from Kernel to userland. */
		error = copy_to_user(p_si, si_entry->data, data_len);
		if (error != 0)
			return -EFAULT;

		p_si += data_len;
		si_entry = si_entry->next;
	}

	/* end. */
	TCPCP_DPRINTK("end.\n");
	return 0;
}

/**
 * do_get_res_size - get result structure to get size of FDRs.
 *
 * [Args]
 *   (IN) *sk       : socket
 *   (IN) *user_res : result structure
 *
 * [Return]
 *   0      : OK.
 *  -EFAULT : NG.
 **/
static int do_get_res_size(struct sock *sk, struct result *user_res)
{
	struct tcpcp_req_entry *r_entry;

	/* plug-in. */
	plugin_dummy();

	/* start. */
	TCPCP_DPRINTK("start. sk=0x%p, user_res=0x%p\n",
		 sk, user_res);

#ifdef CONFIG_TCPCP_DEBUG
	/* @@@ print Process List with lock. (for DEBUG) @@@ */
	print_list();
#endif

	/* lookup Request Entry and unlink from Timeout List. */
	r_entry = tcpcp_reset_timeout(sk);
	if (r_entry == NULL) {
		TCPCP_DPRINTK("[Error] can't lookup Request Entry.\n");
		return -EFAULT;
	}

	if (r_entry->res.fdrs_size == 0) {
		TCPCP_DPRINTK("[Error] fdrs_size is 0.\n");
		/* delete Request Entry with common entries. */
		tcpcp_del_req_entry(r_entry);
		return -EFAULT;
	}

#ifdef CONFIG_TCPCP_DEBUG
	/* @@@ print Process List with lock. (for DEBUG) @@@ */
	print_list();
#endif

	/* copy FDR Info. */
	memcpy(user_res, &(r_entry->res), sizeof(struct result));

	/* link to Timeout List. */
	tcpcp_set_timeout(r_entry);

	/* end. */
	TCPCP_DPRINTK("end.\n");
	return 0;
}

/**
 * do_get_res - get FDR Info.
 *
 * Caller must lock socket.
 *
 * [Args]
 *   (IN) *r_entry    : Request Entry
 *   (OUT) *user_fdrs : each fd processing result informations returned to user
 *   (OUT) *user_size : size that each fd processing result informations
 *                      returned to user
 *
 * [Return]
 *   0      : OK.
 *  -EFAULT : NG.
 **/
static int do_get_res(struct tcpcp_req_entry *r_entry, struct fdr *user_fdrs,
		      unsigned int *user_size)
{
	struct tcpcp_com_entry *fdr_entry;
	char *p_fdrs;
	unsigned int data_len;
	int error;

	/* plug-in. */
	plugin_dummy();

	/* start. */
	TCPCP_DPRINTK("start. r_entry=0x%p\n", r_entry);

	/* put size of FDR Info. to userland. */
	error = put_user(r_entry->res.fdrs_size, user_size);
	if (error != 0)
		return -EFAULT;

	p_fdrs = (char *)user_fdrs;
	fdr_entry = r_entry->f_entry;

	while (fdr_entry != NULL) {
		/* @@@ for DEBUG @@@ */
		TCPCP_DPRINTK("fdr_entry=0x%p\n",
			fdr_entry);

		data_len = fdr_entry->data_len;

		/* copy FDR Info. from Kernel to userland. */
		error = copy_to_user(p_fdrs, fdr_entry->data, data_len);
		if (error != 0)
			return -EFAULT;

		p_fdrs += data_len;
		fdr_entry = fdr_entry->next;
	}

	/* end. */
	TCPCP_DPRINTK("end.\n");
	return 0;
}


/* ----- Substance of a hook ------------------------------------------------ */

/**
 * __tcpcp_com_ope
 *      Substance of tcpcp_make_si_hook, tcpcp_make_si_all_hook,
 *      tcpcp_stop_hook, tcpcp_stop_all_hook, tcpcp_close_all_hook,
 *      tcpcp_start_hook, tcpcp_start_all_hook and tcpcp_get_res_size_hook.
 *
 * [Args]
 *   (IN) *sk		: socket
 *   (IN) optname	: option name of TCPCP operations
 *   (OUT) *user_res	: result info
 *
 * [Return]
 *   0		: OK.
 *   other	: skip
 **/
static int __tcpcp_com_ope(struct sock *sk, int optname,
			   struct result *user_res)
{
	struct tcpcp_req_entry *r_entry = NULL;
	unsigned char flag = 0;
	int error;

	/* plug-in. */
	plugin_dummy();

	/* start. */
	TCPCP_DPRINTK("start. sk=0x%p, optname=%d\n",
		sk, optname);

	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("sk->sk_socket->state=%d\n",
		sk->sk_socket->state);
	TCPCP_DPRINTK("sk->sk_state        =%d\n",
		sk->sk_state);

	/* set TCPCP operation flag. */
	switch (optname) {
	case TCPCP_MAKE_SI:
		flag = TCPCP_FLAG_MAKE_SI;
		break;
	case TCPCP_MAKE_SI_ALL:
		flag = TCPCP_FLAG_MAKE_SI | TCPCP_FLAG_ALL;
		break;
	case TCPCP_STOP:
		flag = TCPCP_FLAG_STOP;
		break;
	case TCPCP_STOP_ALL:
		flag = TCPCP_FLAG_STOP | TCPCP_FLAG_ALL;
		break;
	case TCPCP_CLOSE_ALL:
		flag = TCPCP_FLAG_CLOSE | TCPCP_FLAG_ALL;
		break;
	case TCPCP_START:
		flag = TCPCP_FLAG_START;
		break;
	case TCPCP_START_ALL:
		flag = TCPCP_FLAG_START | TCPCP_FLAG_ALL;
		break;
	case TCPCP_GET_RES_SIZE:
		error = do_get_res_size(sk, user_res);
		/* @@@ for DEBUG @@@ */
		TCPCP_DPRINTK("after do_get_res_size. error=%d\n",
			error);
		TCPCP_DPRINTK("user_res->fdrs_size=%u\n",
			user_res->fdrs_size);

		/* if error, return value is always -EFAULT. */
		return error;
	default:
		return -EINVAL;
	}

	/* check conflict and add Request Entry. */
	error = tcpcp_add_req_entry(sk, flag, &r_entry);
	if (error != 0)
		return error;

#ifdef CONFIG_TCPCP_DEBUG
	/* @@@ print Process List with lock. (for DEBUG) @@@ */
	print_list();
#endif

	/* distribution of the processing to each operation */
	if (flag & TCPCP_FLAG_MAKE_SI)
		/* make TCP-SIs */
		error = do_tcpcp_make_si(sk, r_entry, user_res);
	else if (flag & TCPCP_FLAG_STOP)
		/* stop TCP sockets */
		error = do_tcpcp_stop(sk, r_entry, user_res);
	else if (flag & TCPCP_FLAG_CLOSE)
		/* close TCP sockets */
		error = do_tcpcp_close(sk, r_entry, user_res);
	else if (flag & TCPCP_FLAG_START)
		/* start TCP sockets */
		error = do_tcpcp_start(sk, r_entry, user_res);
	else 
		return -EINVAL;
	if (error != 0) {
		TCPCP_DPRINTK("[Error] do_tcpcp_*** failure. error=%d\n",
			error);

		/* delete Request Entry with common entries. */
		tcpcp_del_req_entry(r_entry);
		return error;
	}
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("after do_tcpcp_***.\n");

#ifdef CONFIG_TCPCP_DEBUG
	/* @@@ print Process List with lock. (for DEBUG) @@@ */
	print_list();
#endif

	/* check TCPCP operation flag. */
	switch (optname) {

	case TCPCP_MAKE_SI:
	case TCPCP_MAKE_SI_ALL:
	case TCPCP_STOP_ALL:
	case TCPCP_CLOSE_ALL:
	case TCPCP_START_ALL:
		/* link to Timeout List. */
		tcpcp_set_timeout(r_entry);
		break;

	case TCPCP_STOP:
	case TCPCP_START:
		/* delete Request Entry with common entries. */
		tcpcp_del_req_entry(r_entry);
	}

#ifdef CONFIG_TCPCP_DEBUG
	/* @@@ print Process List with lock. (for DEBUG) @@@ */
	print_list();
#endif

	/* end. */
	TCPCP_DPRINTK("end.\n");
	return 0;
}

/**
 * __tcpcp_set_si - set SIs from user to Kernel.
 *
 * [Args]
 *   (IN) *sk      : socket
 *   (IN) optname  : option name of TCPCP operations
 *   (IN) *user_si : TCP-SI data from user (const)
 *   (IN) *size    : length of TCP-SI data from user
 *
 * [Return]
 *   0      : OK.
 *  -EINVAL :
 *  -ENOMEM :
 *   other  :
 **/
static int __tcpcp_set_si(struct sock *sk, int optname,
			  const struct tcpcp_total_si_hdr *user_si,
			  unsigned int size)
{
	struct tcpcp_req_entry *r_entry = NULL;
	unsigned char flag;
	int error = 0;

	/* plug-in. */
	plugin_dummy();

	/* start. */
	TCPCP_DPRINTK("start. sk=0x%p, optname=%d, user_si=0x%p,"
		" size=%u\n", sk, optname, user_si, size);

	/* check TCP-SIs size. */
	if (size
	    < sizeof(struct tcpcp_total_si_hdr) + sizeof(struct tcpcp_si_hdr))
		return -EINVAL;
	if (size != ntohl(user_si->total_si_length) * 4)
		return -EINVAL;

	/* set TCPCP operation flag. */
	flag = TCPCP_FLAG_SET_SI;
	if (optname == TCPCP_SET_SI_ALL)
		flag |= TCPCP_FLAG_ALL;
								
	/* release socket (lock_sock() is called at tcp_setsockopt().) */
	release_sock(sk);
	TCPCP_DPRINTK("socket is unlocked.\n");

	/* check conflict and add Request Entry. */
	error = tcpcp_add_req_entry(sk, flag, &r_entry);
	if (error != 0)
		goto out;

#ifdef CONFIG_TCPCP_DEBUG
	/* @@@ print Process List with lock. (for DEBUG) @@@ */
	print_list();
#endif

	/* set TCP-SIs to Kernel. */
	error = do_tcpcp_set_si(sk, user_si, r_entry);
	if (error != 0) {
		/* @@@ for DEBUG @@@ */
		TCPCP_DPRINTK("after do_tcpcp_set_si. error=%d\n",
			error);

		/* delete Request Entry with common entries. */
		tcpcp_del_req_entry(r_entry);
		goto out;
	}
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("after do_tcpcp_set_si.\n");

	if (optname == TCPCP_SET_SI_ALL) {
		/* link to Timeout List. (only "ALL") */
		tcpcp_set_timeout(r_entry);
		/* @@@ for DEBUG @@@ */
		TCPCP_DPRINTK("req_entry->timeout=%u\n",
			r_entry->timeout);
	} else
		/* delete Request Entry with common entries. (only "ONE") */
		tcpcp_del_req_entry(r_entry);

#ifdef CONFIG_TCPCP_DEBUG
	/* @@@ print Process List with lock. (for DEBUG) @@@ */
	print_list();
#endif

out:
	/* lock socket. */
	lock_sock(sk);
	TCPCP_DPRINTK("socket is locked.\n");

	/* end. */
	TCPCP_DPRINTK("end.\n");
	return error;
}

/**
 * __tcpcp_get_si - get TCP-SI from Kernel.
 *
 * [Args]
 *   (IN)  *sk        : socket
 *   (OUT) *user_si   : TCP-SIs to get
 *   (OUT) *user_size : size of user_si
 *
 * [Return]
 *   0      : OK.
 *  -EFAULT : NG. (....)
 **/
static int __tcpcp_get_si(struct sock *sk, struct tcpcp_total_si_hdr *user_si,
			  unsigned int *user_size)
{
	struct tcpcp_req_entry *r_entry;
	int error;

	/* plug-in. */
	plugin_dummy();

	/* start. */
	TCPCP_DPRINTK("start. sk=0x%p\n", sk);

#ifdef CONFIG_TCPCP_DEBUG
	/* @@@ print Process List with lock. (for DEBUG) @@@ */
	print_list();
#endif

	/* lookup Request Entry and unlink from Timeout List. */
	r_entry = tcpcp_reset_timeout(sk);
	if (r_entry == NULL)
		return -EFAULT;
	if (r_entry->s_entry == NULL) {
		error = -EFAULT;
		goto err;
	}

#ifdef CONFIG_TCPCP_DEBUG
	/* @@@ print Process List with lock. (for DEBUG) @@@ */
	print_list();
#endif

	/* get TCP-SIs. */
	error = do_get_si(r_entry, user_si, user_size);
	if (error != 0) {
		/* only -EFAULT. */
		TCPCP_DPRINTK("[Error] can't get TCP-SIs. error=%d\n",
			error);
		goto err;
	}
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("after do_get_si.\n");

	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("user_size        =%u\n",
		*user_size);
	TCPCP_DPRINTK("user_si->total_si=%u\n",
		user_si->total_si_length);
	TCPCP_DPRINTK("user_si->major   =%u\n",
		user_si->major);
	TCPCP_DPRINTK("user_si->minor   =%u\n",
		user_si->minor);
	TCPCP_DPRINTK("user_si->si_cnt  =%u\n",
		user_si->si_cnt);

	if (r_entry->flag & TCPCP_FLAG_ALL) {
		/* link to Timeout List. (only "ALL") */
		tcpcp_set_timeout(r_entry);
		/* @@@ for DEBUG @@@ */
		TCPCP_DPRINTK("r_entry->timeout=%u\n",
			r_entry->timeout);
	} else
		/* delete Request Entry with common entries. (only "ONE") */
		tcpcp_del_req_entry(r_entry);

#ifdef CONFIG_TCPCP_DEBUG
	/* @@@ print Process List with lock. (for DEBUG) @@@ */
	print_list();
#endif

	/* end. */
	TCPCP_DPRINTK("end.\n");
	return 0;

err:
	/* delete Request Entry with common entries. */
	tcpcp_del_req_entry(r_entry);

	/* end. */
	TCPCP_DPRINTK("end.\n");
	return error;
}

/**
 * __tcpcp_get_res - get FDRs from Kernel.
 *
 * [Args]
 *   (IN)  *sk        : TCP socket
 *   (OUT) *user_fdrs : FDRs to get
 *   (OUT) *user_size : size of user_fdrs
 *
 * [Return]
 *   0      : OK.
 *  -EFAULT : NG. (....)
 **/	
static int __tcpcp_get_res(struct sock *sk, struct fdr *user_fdrs,
			   unsigned int *user_size)
{
	struct tcpcp_req_entry *r_entry;
	int error = 0;

	/* plug-in. */
	plugin_dummy();

	/* start. */
	TCPCP_DPRINTK("start. sk=0x%p\n", sk);

#ifdef CONFIG_TCPCP_DEBUG
	/* @@@ print Process List with lock. (for DEBUG) @@@ */
	print_list();
#endif

	/* lookup Request Entry and unlink from Timeout List. */
	r_entry = tcpcp_reset_timeout(sk);
	if (r_entry == NULL)
		return -EFAULT;
	if (r_entry->f_entry == NULL) {
		tcpcp_del_req_entry(r_entry);
		return -EFAULT;
	}

#ifdef CONFIG_TCPCP_DEBUG
	/* @@@ print Process List with lock. (for DEBUG) @@@ */
	print_list();
#endif

	/* get FDR Info. */
	error = do_get_res(r_entry, user_fdrs, user_size);

	/* delete Request Entry with common entries. */
	tcpcp_del_req_entry(r_entry);

#ifdef CONFIG_TCPCP_DEBUG
	/* @@@ print Process List with lock. (for DEBUG) @@@ */
	print_list();
#endif

	/* end. */
	TCPCP_DPRINTK("end. error=%d\n", error);
	return error;
}

/**
 * tcpcp_close - ....
 *
 * [Args]
 *   (IN) *sk : TCP socket
 **/
static void __tcpcp_close(struct sock *sk)
{
	/* plug-in. */
	plugin_dummy();

	/* start. */
	TCPCP_DPRINTK("start. sk=0x%p\n", sk);

	tcpcp_quick_timeout(sk);

	/* end. */
	TCPCP_DPRINTK("end.\n");
	return;
}



/**
 * tcpcp_flush
 *
 */
static int __tcpcp_flush(struct sock *sk, struct timespec __user *user_timeout)
{
  struct timespec timeo;
  int err = 0;
  /* ensure progress by using at least 1 */
  long poll_timeout_max = max(1U, (unsigned)(TCPCP_FLUSH_POLL_TIMEOUT_USECS/(USEC_PER_SEC/HZ)));
  long timeout;
  /* plug-in. */
  plugin_dummy();

  /* Convert timespec to timeout */
  if( copy_from_user(&timeo, user_timeout, sizeof(timeo)) ) {
    return -EFAULT;
  }
  if( (0 > timeo.tv_nsec)
      || (NSEC_PER_SEC <= timeo.tv_nsec) 
      || (0 > timeo.tv_sec) ) {
    return -EINVAL;
  }
  timeout = MAX_SCHEDULE_TIMEOUT;
  if ( (0 == timeo.tv_sec) && (0 == timeo.tv_nsec) ) {
    return 0;
  }
  if ((MAX_SCHEDULE_TIMEOUT/HZ - 1) > timeo.tv_sec ) {
    timeout = timeo.tv_sec*HZ + (timeo.tv_nsec+(NSEC_PER_SEC/HZ-1))/(NSEC_PER_SEC/HZ);
  }

  /* start. */
  TCPCP_DPRINTK("tcpcp_flush. sk=0x%p, user_timeout=0x%p, %ld s, %ld ns, poll_timeout_max=%ld\n",
		sk, user_timeout, timeo.tv_sec, timeo.tv_nsec, poll_timeout_max);

  /* Work on locked sockets! */
  { lock_sock(sk);
    TCP_CHECK_TIMER(sk);

    {
      int done;
      long poll_timeout;
      DEFINE_WAIT(wait);

      /* Sleep/wait while data still exists to send */
      do {
	prepare_to_wait(sk->sk_sleep, &wait, TASK_INTERRUPTIBLE);
	/* set_bit(SOCK_ASYNC_WAITDATA, &sk->sk_socket->flags); */

	/* set poll timeout to minimum. note that timeout > 0. */
	poll_timeout = min( poll_timeout_max, timeout );
	timeout -= poll_timeout;

	/* condition for wait event is when NOT to wait, sk_wait_event
	   returns status of condition */
	done = sk_wait_event(sk, &poll_timeout, tcp_write_queue_empty(sk));
	TCPCP_DPRINTK("tcpcp_flush: loop. done=%d, wmem_queued=%d, poll_timeout=%ld, timeout=%ld, write_queue_empty=%d\n",
		      done, (int)sk->sk_wmem_queued, 
		      (long)poll_timeout, (long)timeout,
		      (int)tcp_write_queue_empty(sk));
	timeout += poll_timeout; // add back remaining time
      
	if( done ) {
	  break;
	}
      } while (!signal_pending(current) && timeout);
      
      /* clear_bit(SOCK_ASYNC_WAITDATA, &sk->sk_socket->flags); */
      finish_wait(sk->sk_sleep, &wait);
    }

    TCP_CHECK_TIMER(sk);
    release_sock(sk); }

  /* Convert timeout back to timespec */
  timeo.tv_sec = timeout / HZ;
  timeo.tv_nsec = ((timeout % HZ) * NSEC_PER_SEC) / HZ;

  if (copy_to_user(user_timeout, &timeo, sizeof(timeo))) {
    return -EFAULT;
  }

  if( sock_error(sk) ) {
    err = -EPIPE;
  }
  // else: write queue is empty or timeout

  /* end. */
  TCPCP_DPRINTK("tcpcp_flush: end. wmem_queued=%d, timeout=%ld, write_queue_empty=%d, sk_err=%d\n",
		(int)sk->sk_wmem_queued, (long)timeout, 
		(int)tcp_write_queue_empty(sk),
		sock_error(sk));
  return err;
}


/* ----- Initialization and (module) exit ---------------------------------- */

/**
 * tcpcp_init - processing tcpcp initilaization.
 **/
static int __init tcpcp_init(void)
{
	/* plug-in. */
   /* plugin_dummy(); */

	/* start. */
	TCPCP_DPRINTK("start.\n");

	/* lock tcpcp_sem. */
	down_write(&tcpcp_sem);
	TCPCP_DPRINTK("tcpcp_sem is locked.\n");

  __sysctl_rmem_max = sysctl_tcp_rmem[1];
  __sysctl_wmem_max = sysctl_tcp_wmem[1];
  
  tcpcp_v4_init();

#if defined(CONFIG_IPV6) || defined(CONFIG_IPV6_MODULE)
  tcpcp_v6_init();
#endif /* defined(CONFIG_IPV6) || defined(CONFIG_IPV6_MODULE) */

	/* initialize Process List */
	memset(&proc_list, 0, sizeof(struct tcpcp_list));
	INIT_LIST_HEAD(&(proc_list.head));
	proc_list.lock = RW_LOCK_UNLOCKED;
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("proc_list.head     =0x%p\n",
		&proc_list.head);
	TCPCP_DPRINTK("proc_list.head.next=0x%p\n",
		proc_list.head.next);
	TCPCP_DPRINTK("proc_list.head.prev=0x%p\n",
		proc_list.head.prev);

	/* initialize Timeout List. */
	memset(&timeout_list, 0, sizeof(struct tcpcp_list));
	INIT_LIST_HEAD(&(timeout_list.head));
	timeout_list.lock = RW_LOCK_UNLOCKED;
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("timeout_list.head     =0x%p\n",
		&timeout_list.head);
	TCPCP_DPRINTK("timeout_list.head.next=0x%p\n",
		timeout_list.head.next);
	TCPCP_DPRINTK("timeout_list.head.prev=0x%p\n",
		timeout_list.head.prev);

   /* set hook. */
	{
    tcpcp_config_t *config =	tcpcp_get_config();
    config->tcpcp_com_ope_hook = __tcpcp_com_ope;
    config->tcpcp_get_si_hook = __tcpcp_get_si;
    config->tcpcp_get_res_hook = __tcpcp_get_res;
    config->tcpcp_set_si_hook = __tcpcp_set_si;
    config->tcpcp_close_hook = __tcpcp_close;
    config->tcpcp_flush_hook = __tcpcp_flush;
     /* @@@ for DEBUG @@@ */
    TCPCP_DPRINTK("__tcpcp_com_ope   =0x%p\n",
            &__tcpcp_com_ope);
    TCPCP_DPRINTK("tcpcp_com_ope_hook=0x%p\n",
            config->tcpcp_com_ope_hook);
    TCPCP_DPRINTK("__tcpcp_get_si    =0x%p\n",
            &__tcpcp_get_si);
    TCPCP_DPRINTK("tcpcp_get_si_hook =0x%p\n",
            config->tcpcp_get_si_hook);
    TCPCP_DPRINTK("__tcpcp_get_res   =0x%p\n",
            &__tcpcp_get_res);
    TCPCP_DPRINTK("tcpcp_get_res_hook=0x%p\n",
            config->tcpcp_get_res_hook);
    TCPCP_DPRINTK("__tcpcp_set_si    =0x%p\n",
            &__tcpcp_set_si);
    TCPCP_DPRINTK("tcpcp_set_si_hook =0x%p\n",
            config->tcpcp_set_si_hook);
    TCPCP_DPRINTK("__tcpcp_close     =0x%p\n",
            &__tcpcp_close);
    TCPCP_DPRINTK("tcpcp_close_hook  =0x%p\n",
            config->tcpcp_close_hook);
    TCPCP_DPRINTK("tcpcp_flush_hook  =0x%p\n",
            config->tcpcp_flush_hook);
	}

	/* unlock tcpcp_sem. */
	up_write(&tcpcp_sem);
	TCPCP_DPRINTK("tcpcp_sem is unlocked.\n");

	/* print message. */
	printk(KERN_INFO "TCPCP : %s : initialized."	\
	       " (SI format version %u.%u)\n", __func__,
	       TCPCP_CURRENT_SI_MAJOR, TCPCP_CURRENT_SI_MINOR);

	/* end. */
	TCPCP_DPRINTK("end.\n");
	return 0;
}

/**
 * tcpcp_exit - processing tcpcp exit.
 **/
static void __exit tcpcp_exit(void)
{
	/* plug-in. */
	plugin_dummy();

	/* start. */
	TCPCP_DPRINTK("start.\n");

	/* lock tcpcp_sem. */
	down_write(&tcpcp_sem);
	TCPCP_DPRINTK("tcpcp_sem is locked.\n");

	/* reset hook. */
  {
    tcpcp_config_t *config = tcpcp_get_config();
    
    config->tcpcp_com_ope_hook = NULL;
    config->tcpcp_get_si_hook = NULL;
    config->tcpcp_get_res_hook = NULL;
    config->tcpcp_set_si_hook = NULL;
    config->tcpcp_close_hook = NULL;
    config->tcpcp_flush_hook = NULL;
     /* @@@ for DEBUG @@@ */
    TCPCP_DPRINTK("tcpcp_com_ope_hook=0x%p\n",
            config->tcpcp_com_ope_hook);
    TCPCP_DPRINTK("tcpcp_get_si_hook =0x%p\n",
            config->tcpcp_get_si_hook);
    TCPCP_DPRINTK("tcpcp_get_res_hook=0x%p\n",
            config->tcpcp_get_res_hook);
    TCPCP_DPRINTK("tcpcp_set_si_hook =0x%p\n",
            config->tcpcp_set_si_hook);
    TCPCP_DPRINTK("tcpcp_close_hook  =0x%p\n",
            config->tcpcp_close_hook);
    TCPCP_DPRINTK("tcpcp_flush_hook  =0x%p\n",
            config->tcpcp_flush_hook);
  }

	/* delete and free all entry. */
	tcpcp_del_all_entry();

	if (timer_pending(&tcpcp_timer) != 0) {
		/* sleep Timeout-watcher. */
		del_timer_sync(&tcpcp_timer);
		TCPCP_DPRINTK("TCPCP timeout watcher is deleted.\n");
	}
	/* @@@ for DEBUG @@@ */
	TCPCP_DPRINTK("timer_pending=%d\n",
		timer_pending(&tcpcp_timer));

	/* unlock tcpcp_sem. */
	up_write(&tcpcp_sem);
	TCPCP_DPRINTK("tcpcp_sem is unlocked.\n");

#ifdef CONFIG_TCPCP_DEBUG
	/* @@@ print Process List with lock. (for DEBUG) @@@ */
	print_list();
#endif

	/* print messasge. */
	printk(KERN_INFO "TCPCP : %s : exit.\n", __func__);

	/* end. */
	TCPCP_DPRINTK("end.\n");
	return;
}


/* ----- for DEBUG --------------------------------------------------------- */

#ifdef CONFIG_TCPCP_DEBUG
/**
 * __print_list - print Process List without lock. (for DEBUG)
 **/
static void __print_list(void)
{
	struct tcpcp_proc_entry *p_entry = NULL;
	struct tcpcp_req_entry *r_entry = NULL;
	struct tcpcp_com_entry *c_entry, *next_c_entry = NULL;

	/* start. */
	TCPCP_DPRINTK("@@@@@@@@@@ print list @@@@@@@@@@\n");

	/* print Process List head. */
	TCPCP_DPRINTK("[ PH ]\n");

	/* search Process List. */
	list_for_each_entry(p_entry, &(proc_list.head), node) {
		/* print Process Entry. */
		TCPCP_DPRINTK("[ PE ] tgid=%d, flag=%u\n",
			p_entry->tgid, p_entry->flag);

		/* search Request Entry. */
		list_for_each_entry(r_entry, &(p_entry->r_list_head), node) {
			/* print Request Entry. */
			TCPCP_DPRINTK("  ->[ RE ] sk=0x%p, pid=%d, flag=%u, fd=%d\n",	
				r_entry->sk, r_entry->pid, r_entry->flag,
				r_entry->fd);

			/* print TCP-SIs Entries. */
			c_entry = r_entry->s_entry;
			while ((c_entry) != NULL) {
				next_c_entry = c_entry->next;
				TCPCP_DPRINTK("    ->[SOCK] data_len=%u\n",
					c_entry->data_len);
				c_entry = next_c_entry;
			}

			/* print FDR Info Entries. */
			c_entry = r_entry->f_entry;
			while ((c_entry) != NULL) {
				next_c_entry = c_entry->next;
				TCPCP_DPRINTK("    ->[REST] data_len=%u\n",
					c_entry->data_len);
				c_entry = next_c_entry;
			}
		}
	}

	/* print Timer List head. */
	TCPCP_DPRINTK("[ TH ]\n");

	/* search Timeout Request List. */
	list_for_each_entry(r_entry, &(timeout_list.head), timer_node) {
		/* print Request Entry. */
		TCPCP_DPRINTK("  ->[ RE ] sk=0x%p, pid=%d, flag=%u, fd=%d,"
			" timeout=%u\n",
			r_entry->sk, r_entry->pid, r_entry->flag,
			r_entry->fd, r_entry->timeout);
	}

	/* end. */
	TCPCP_DPRINTK("@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@\n");
	return;
}

/**
 * print_list - print Process List with lock. (for DEBUG)
 **/
static void print_list(void)
{
	/* lock Process List (and Timeout List). */
	write_lock(&(proc_list.lock));

	__print_list();

	/* unlock Process List (and Timeout List). */
	write_unlock(&(proc_list.lock));

	return;
}

/**
 * skb_queue_check - print writr_queue / receive_queue. (for DEBUG)
 *
 * [Args]
 *   (IN) *queue : write_queue / receive_queue of TCP socket
 **/
static void skb_queue_check(struct sk_buff_head *queue)
{
	struct sk_buff *skb;
	int cnt = 0;

	TCPCP_DPRINTK("queue->qlen=%u\n", queue->qlen);
	TCPCP_DPRINTK("queue      =0x%p\n", queue);
	TCPCP_DPRINTK("queue->next=0x%p\n", queue->next);
	TCPCP_DPRINTK("queue->prev=0x%p\n", queue->prev);

	skb_queue_walk(queue, skb) {
		cnt++;
		TCPCP_DPRINTK("skb[%d]:0x%p\n", cnt, skb);
	}

	return;
}

/**
 * check_fd - .... (for DEBUG)
 **/
static void check_fd(void)
{
  struct fdtable *fdt = files_fdtable(current->files);
	struct file * filp;
	struct sock * sk_i;
	unsigned int fd;
	unsigned int file = 0;
	unsigned int socket_ng = 0;
	unsigned int socket = 0;
	unsigned int sockbuf[fdt->max_fds];
	unsigned int cnt = 0;

	TCPCP_DPRINTK("*******************\n");

	for (fd = 0; fd < fdt->max_fds; fd++) {
		filp = fdt->fd[fd];
		if (filp == NULL)
			continue;

		if ( ! S_ISSOCK(filp->f_dentry->d_inode->i_mode) ) {
			file ++;
			continue;
		}

		sk_i = SOCKET_I(filp->f_dentry->d_inode)->sk;
		if (sk_i == NULL) {
			socket_ng ++;
			continue;
		}

		socket ++;
		sockbuf[cnt] = fd;
		cnt ++;
	}

	TCPCP_DPRINTK("file      : %u\n", file);
	TCPCP_DPRINTK("socket_ng : %u\n", socket_ng);
	TCPCP_DPRINTK("socket    : %u\n", socket);

	if (cnt == 0) {
		TCPCP_DPRINTK("*******************\n");
		return;
	}

	for (fd = 0; fd < cnt; fd ++) {
		TCPCP_DPRINTK("sockbuf[%u] : %u\n",
			      fd, sockbuf[fd]);
	}

	TCPCP_DPRINTK("*******************\n");
	return;
}
#endif /* CONFIG_TCPCP_DEBUG */


module_init(tcpcp_init);
module_exit(tcpcp_exit);
